/*******************************************************************************
 * Software in the Sustain Ecosystem are Released Under Terms of Apache Software License 
 * 
 * This research has been supported by funding from the US National Science Foundation's CSSI program through awards 1931363, 1931324, 1931335, and 1931283. The project is a joint effort involving Colorado State University, Arizona State University, the University of California-Irvine, and the University of Maryland - Baltimore County. All redistributions of the software must also include this information. 
 * 
 * TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 * 
 * 
 * 1. Definitions.
 * 
 * "License" shall mean the terms and conditions for use, reproduction, and distribution as defined by Sections 1 through 9 of this document.
 * 
 * "Licensor" shall mean the copyright owner or entity authorized by the copyright owner that is granting the License.
 * 
 * "Legal Entity" shall mean the union of the acting entity and all other entities that control, are controlled by, or are under common control with that entity. For the purposes of this definition, "control" means (i) the power, direct or indirect, to cause the direction or management of such entity, whether by contract or otherwise, or (ii) ownership of fifty percent (50%) or more of the outstanding shares, or (iii) beneficial ownership of such entity.
 * 
 * "You" (or "Your") shall mean an individual or Legal Entity exercising permissions granted by this License.
 * 
 * "Source" form shall mean the preferred form for making modifications, including but not limited to software source code, documentation source, and configuration files.
 * 
 * "Object" form shall mean any form resulting from mechanical transformation or translation of a Source form, including but not limited to compiled object code, generated documentation, and conversions to other media types.
 * 
 * "Work" shall mean the work of authorship, whether in Source or Object form, made available under the License, as indicated by a copyright notice that is included in or attached to the work (an example is provided in the Appendix below).
 * 
 * "Derivative Works" shall mean any work, whether in Source or Object form, that is based on (or derived from) the Work and for which the editorial revisions, annotations, elaborations, or other modifications represent, as a whole, an original work of authorship. For the purposes of this License, Derivative Works shall not include works that remain separable from, or merely link (or bind by name) to the interfaces of, the Work and Derivative Works thereof.
 * 
 * "Contribution" shall mean any work of authorship, including the original version of the Work and any modifications or additions to that Work or Derivative Works thereof, that is intentionally submitted to Licensor for inclusion in the Work by the copyright owner or by an individual or Legal Entity authorized to submit on behalf of the copyright owner. For the purposes of this definition, "submitted" means any form of electronic, verbal, or written communication sent to the Licensor or its representatives, including but not limited to communication on electronic mailing lists, source code control systems, and issue tracking systems that are managed by, or on behalf of, the Licensor for the purpose of discussing and improving the Work, but excluding communication that is conspicuously marked or otherwise designated in writing by the copyright owner as "Not a Contribution."
 * 
 * "Contributor" shall mean Licensor and any individual or Legal Entity on behalf of whom a Contribution has been received by Licensor and subsequently incorporated within the Work.
 * 
 * 2. Grant of Copyright License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable copyright license to reproduce, prepare Derivative Works of, publicly display, publicly perform, sublicense, and distribute the Work and such Derivative Works in Source or Object form.
 * 
 * 3. Grant of Patent License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable (except as stated in this section) patent license to make, have made, use, offer to sell, sell, import, and otherwise transfer the Work, where such license applies only to those patent claims licensable by such Contributor that are necessarily infringed by their Contribution(s) alone or by combination of their Contribution(s) with the Work to which such Contribution(s) was submitted. If You institute patent litigation against any entity (including a cross-claim or counterclaim in a lawsuit) alleging that the Work or a Contribution incorporated within the Work constitutes direct or contributory patent infringement, then any patent licenses granted to You under this License for that Work shall terminate as of the date such litigation is filed.
 * 
 * 4. Redistribution. You may reproduce and distribute copies of the Work or Derivative Works thereof in any medium, with or without modifications, and in Source or Object form, provided that You meet the following conditions:
 * 
 * You must give any other recipients of the Work or Derivative Works a copy of this License; and
 * You must cause any modified files to carry prominent notices stating that You changed the files; and
 * You must retain, in the Source form of any Derivative Works that You distribute, all copyright, patent, trademark, and attribution notices from the Source form of the Work, excluding those notices that do not pertain to any part of the Derivative Works; and
 * If the Work includes a "NOTICE" text file as part of its distribution, then any Derivative Works that You distribute must include a readable copy of the attribution notices contained within such NOTICE file, excluding those notices that do not pertain to any part of the Derivative Works, in at least one of the following places: within a NOTICE text file distributed as part of the Derivative Works; within the Source form or documentation, if provided along with the Derivative Works; or, within a display generated by the Derivative Works, if and wherever such third-party notices normally appear. The contents of the NOTICE file are for informational purposes only and do not modify the License. You may add Your own attribution notices within Derivative Works that You distribute, alongside or as an addendum to the NOTICE text from the Work, provided that such additional attribution notices cannot be construed as modifying the License. 
 * 
 * You may add Your own copyright statement to Your modifications and may provide additional or different license terms and conditions for use, reproduction, or distribution of Your modifications, or for any such Derivative Works as a whole, provided Your use, reproduction, and distribution of the Work otherwise complies with the conditions stated in this License.
 * 5. Submission of Contributions. Unless You explicitly state otherwise, any Contribution intentionally submitted for inclusion in the Work by You to the Licensor shall be under the terms and conditions of this License, without any additional terms or conditions. Notwithstanding the above, nothing herein shall supersede or modify the terms of any separate license agreement you may have executed with Licensor regarding such Contributions.
 * 
 * 6. Trademarks. This License does not grant permission to use the trade names, trademarks, service marks, or product names of the Licensor, except as required for reasonable and customary use in describing the origin of the Work and reproducing the content of the NOTICE file.
 * 
 * 7. Disclaimer of Warranty. Unless required by applicable law or agreed to in writing, Licensor provides the Work (and each Contributor provides its Contributions) on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied, including, without limitation, any warranties or conditions of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A PARTICULAR PURPOSE. You are solely responsible for determining the appropriateness of using or redistributing the Work and assume any risks associated with Your exercise of permissions under this License.
 * 
 * 8. Limitation of Liability. In no event and under no legal theory, whether in tort (including negligence), contract, or otherwise, unless required by applicable law (such as deliberate and grossly negligent acts) or agreed to in writing, shall any Contributor be liable to You for damages, including any direct, indirect, special, incidental, or consequential damages of any character arising as a result of this License or out of the use or inability to use the Work (including but not limited to damages for loss of goodwill, work stoppage, computer failure or malfunction, or any and all other commercial damages or losses), even if such Contributor has been advised of the possibility of such damages.
 * 
 * 9. Accepting Warranty or Additional Liability. While redistributing the Work or Derivative Works thereof, You may choose to offer, and charge a fee for, acceptance of support, warranty, indemnity, or other liability obligations and/or rights consistent with this License. However, in accepting such obligations, You may act only on Your own behalf and on Your sole responsibility, not on behalf of any other Contributor, and only if You agree to indemnify, defend, and hold each Contributor harmless for any liability incurred by, or claims asserted against, such Contributor by reason of your accepting any such warranty or additional liability. 
 * 
 * END OF TERMS AND CONDITIONS
 ******************************************************************************/


package galileo.bmp;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.image.BufferedImage;
import java.awt.image.DataBufferByte;
import java.io.File;
import java.io.IOException;
import java.net.InetAddress;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.imageio.ImageIO;

import galileo.dataset.Coordinates;
import galileo.dataset.Point;
import galileo.dataset.SpatialRange;
import galileo.util.GeoHash;

/**
 * Implements a bitmap-based spatial index that can be used to determine whether
 * or not data is stored within specified spatial regions.
 *
 * @author malensek
 */
public class GeoavailabilityGrid {

	private static final Logger logger = Logger.getLogger("galileo");

	private int width, height;

	public Bitmap bmp = new Bitmap();
	private SortedSet<Integer> pendingUpdates = new TreeSet<>();

	private SpatialRange baseRange;
	private float xDegreesPerPixel;
	private float yDegreesPerPixel;
	private String baseHash;
	
	public static void main(String arg[]) {
		
		GeoavailabilityGrid gg = new GeoavailabilityGrid("9xjqb",6);
		gg.addPoint(new Coordinates(40.5964f,-105.1007f));
	}

	public GeoavailabilityGrid(String baseGeohash, int precision) {
		this.baseRange = GeoHash.decodeHash(baseGeohash);
		this.baseHash = baseGeohash;
		/*
		 * height, width calculated like so: width = 2^(floor(precision / 2))
		 * height = 2^(ceil(precision / 2))
		 */
		int w = precision / 2;
		int h = precision / 2;
		if (precision % 2 != 0) {
			h += 1;
		}

		this.width = (1 << w); /* = 2^w */
		this.height = (1 << h); /* = 2^h */

		/*
		 * Determine the number of degrees in the x and y directions for the
		 * base spatial range this geoavailability grid represents
		 */
		float xDegrees = baseRange.getUpperBoundForLongitude() - baseRange.getLowerBoundForLongitude();
		float yDegrees = baseRange.getUpperBoundForLatitude() - baseRange.getLowerBoundForLatitude();

		/* Determine the number of degrees represented by each grid pixel */
		xDegreesPerPixel = xDegrees / (float) this.width;
		yDegreesPerPixel = yDegrees / (float) this.height;

		if (logger.isLoggable(Level.INFO)) {
			/*logger.log(Level.INFO,
					"Created geoavailability grid: " + "geohash={0}, precision={1}, "
							+ "width={2}, height={3}, baseRange={6}, " + "xDegreesPerPixel={4}, yDegreesPerPixel={5}",
					new Object[] { baseGeohash, precision, width, height, xDegreesPerPixel, yDegreesPerPixel,
							baseRange });*/
		}
	}
	
	public GeoavailabilityGrid(GeoavailabilityGrid grid){
		this.baseRange = grid.baseRange;
		this.baseHash = grid.baseHash;
		this.width = grid.width;
		this.height = grid.height;
		this.xDegreesPerPixel = grid.xDegreesPerPixel;
		this.yDegreesPerPixel = grid.yDegreesPerPixel;
	}

	/**
	 * Adds a new point to this GeoavailabilityGrid.
	 *
	 * @param coords
	 *            The location (coordinates in lat, lon) to add.
	 *
	 * @return true if the point could be added to grid, false otherwise (for
	 *         example, if the point falls outside the purview of the grid)
	 */
	public boolean addPoint(Coordinates coords) {
		Point<Integer> gridPoint = coordinatesToXY(coords);
		return addPoint(gridPoint.X(), gridPoint.Y());
	}

	public boolean addPoint(int x, int y) {
		if (x < 0 || x >= this.width || y < 0 || y >= this.height)
			return false;
		int index = XYtoIndex(x, y);
		if (this.bmp.set(index) == false) {
			/* Could not set the bits now; add to pending updates */
			pendingUpdates.add(index);
		}
		return true;
	}

	/**
	 * Converts a coordinate pair (defined with latitude, longitude in decimal
	 * degrees) to an x, y location in the grid.
	 *
	 * @param coords
	 *            the Coordinates to convert.
	 *
	 * @return Corresponding x, y location in the grid.
	 */
	public Point<Integer> coordinatesToXY(Coordinates coords) {

		/*
		 * Assuming (x, y) coordinates for the geoavailability grids, latitude
		 * will decrease as y increases, and longitude will increase as x
		 * increases. This is reflected in how we compute the differences
		 * between the base points and the coordinates in question.
		 */
		float xDiff = coords.getLongitude() - baseRange.getLowerBoundForLongitude();

		float yDiff = baseRange.getUpperBoundForLatitude() - coords.getLatitude();
		int x = (int) (xDiff / xDegreesPerPixel);
		int y = (int) (yDiff / yDegreesPerPixel);

		return new Point<>(x, y);
	}

	/**
	 * Converts X, Y coordinates to a particular index within the underlying
	 * bitmap implementation. Essentially this converts a 2D index to a 1D
	 * index.
	 *
	 * @param x
	 *            The x coordinate to convert
	 * @param y
	 *            The y coorddinate to convert
	 *
	 * @return A single integer representing the bitmap location of the X, Y
	 *         coordinates.
	 */
	public int XYtoIndex(int x, int y) {
		return y * this.width + x;
	}

	/**
	 * Converts a bitmap index to X, Y coordinates in the grid.
	 */
	public Point<Integer> indexToXY(int index) {
		int x = index % this.width;
		int y = index / this.width;
		return new Point<>(x, y);
	}

	/**
	 * Converts an X, Y grid point to the corresponding SpatialRange that the
	 * grid point spans. (0, 0) indicates (90, -180)
	 */
	public SpatialRange XYtoSpatialRange(int x, int y) {
		if (x < 0 || x >= this.width || y < 0 || y >= this.height) {
			throw new IllegalArgumentException("Out-of-bounds grid coordinates specified");
		}

		float baseLon = baseRange.getLowerBoundForLongitude();
		float baseLat = baseRange.getUpperBoundForLatitude();

		float lowerLon = baseLon + (x * xDegreesPerPixel);
		float upperLon = lowerLon + xDegreesPerPixel;
		float upperLat = baseLat - (y * yDegreesPerPixel);
		float lowerLat = upperLat - yDegreesPerPixel;

		return new SpatialRange(lowerLat, upperLat, lowerLon, upperLon);
	}

	/**
	 * Converts a bitmap index location to a corresponding SpatialRange that the
	 * indexed grid point spans.
	 */
	public SpatialRange indexToSpatialRange(int index) {
		Point<Integer> gridLocation = indexToXY(index);
		return XYtoSpatialRange(gridLocation.X(), gridLocation.Y());
	}

	/**
	 * Applies pending updates that have not yet been integrated into the
	 * GeoavailabilityGrid instance.
	 */
	private void applyUpdates() {
		Bitmap updateBitmap = new Bitmap();
		for (int i : pendingUpdates) {
			if (updateBitmap.set(i) == false) {
				logger.warning("Could not set update bit");
			}
		}
		pendingUpdates.clear();

		this.bmp = this.bmp.or(updateBitmap);
	}

	/**
	 * Reports whether or not the supplied {@link GeoavailabilityQuery} instance
	 * intersects with the bits set in this geoavailability grid. This operation
	 * can be much faster than performing a full inspection of what bits are
	 * actually set.
	 *
	 * @param query
	 *            The query geometry to test for intersection.
	 *
	 * @return true if the supplied {@link GeoavailabilityQuery} intersects with
	 *         the data in the geoavailability grid.
	 */
	public boolean intersects(GeoavailabilityQuery query) throws BitmapException {
		applyUpdates();
		Bitmap queryBitmap = QueryTransform.queryToGridBitmap(query, this);
		return this.bmp.intersects(queryBitmap);
	}

	/**
	 * Queries the geoavailability grid, which involves performing a logical AND
	 * operation and reporting the resulting Bitmap.
	 *
	 * @param query
	 *            The query geometry to evaluate against the geoavailability
	 *            grid.
	 *
	 * @return An array of bitmap indices that matched the query.
	 */
	public int[] query(GeoavailabilityQuery query) throws BitmapException {
		applyUpdates();
		Bitmap queryBitmap = QueryTransform.queryToGridBitmap(query, this);
		if(logger.isLoggable(Level.FINEST)){
			BufferedImage b = null;
			try {
				String hostname = InetAddress.getLocalHost().getHostName();
				File queryImg = new File(getBaseHash() + "-" + hostname + "-query.png");
				if (!queryImg.exists()) {
					b = BitmapVisualization.drawBitmap(queryBitmap, getWidth(), getHeight(), Color.RED);
					BitmapVisualization.imageToFile(b, getBaseHash() + "-" + hostname + "-query.png");
				}
				File dataImg = new File(getBaseHash() + "-" + hostname + "-data.png");
				Bitmap thisBmp = this.bmp;
				if (dataImg.exists()) {
					BufferedImage in = ImageIO.read(dataImg);
					BufferedImage newImage = new BufferedImage(in.getWidth(), in.getHeight(),
							BufferedImage.TYPE_BYTE_INDEXED);

					Graphics2D g = newImage.createGraphics();
					g.drawImage(in, 0, 0, null);
					g.dispose();
					DataBufferByte buffer = ((DataBufferByte) newImage.getData().getDataBuffer());
					byte[] data = buffer.getData();
					Bitmap dataBitmap = Bitmap.fromBytes(data, 0, 0, in.getWidth(), in.getHeight(), getWidth(),
							getHeight());
					thisBmp = thisBmp.or(dataBitmap);
				}
				b = BitmapVisualization.drawIterableMap(thisBmp.iterator(), getWidth(), getHeight(), Color.BLUE);
				BitmapVisualization.imageToFile(b, getBaseHash() + "-" + hostname + "-data.png");
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
		return this.bmp.and(queryBitmap).toArray();
	}
	
	public int[] query(Bitmap queryBitmap){
		applyUpdates();
		return this.bmp.and(queryBitmap).toArray();
	}

	/**
	 * Retrieves the underlying Bitmap instance backing this
	 * GeoavailabilityGrid.
	 */
	protected Bitmap getBitmap() {
		applyUpdates();
		return bmp;
	}

	public String getBaseHash() {
		return this.baseHash;
	}

	/**
	 * Retrieves the width of this GeoavailabilityGrid, in grid cells.
	 */
	public int getWidth() {
		return width;
	}

	/**
	 * Retrieves the height of this GeoavailabilityGrid, in grid cells.
	 */
	public int getHeight() {
		return height;
	}

	/**
	 * Retrieves the base SpatialRange that this GeoavailabilityGrid is
	 * responsible for; the base range defines the geographic scope of this
	 * GeoavailabilityGrid instance.
	 *
	 * @return {@link SpatialRange} representing this GeoavailabilityGrid's
	 *         scope.
	 */
	public SpatialRange getBaseRange() {
		return new SpatialRange(baseRange);
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((baseRange == null) ? 0 : baseRange.hashCode());
		result = prime * result + ((bmp == null) ? 0 : bmp.hashCode());
		result = prime * result + height;
		result = prime * result + width;
		result = prime * result + Float.floatToIntBits(xDegreesPerPixel);
		result = prime * result + Float.floatToIntBits(yDegreesPerPixel);
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (getClass() != obj.getClass()) {
			return false;
		}
		GeoavailabilityGrid other = (GeoavailabilityGrid) obj;
		if (baseRange == null) {
			if (other.baseRange != null) {
				return false;
			}
		} else if (!baseRange.equals(other.baseRange)) {
			return false;
		}
		if (bmp == null) {
			if (other.bmp != null) {
				return false;
			}
		} else if (!bmp.equals(other.bmp)) {
			return false;
		}
		if (height != other.height) {
			return false;
		}
		if (width != other.width) {
			return false;
		}
		if (Float.floatToIntBits(xDegreesPerPixel) != Float.floatToIntBits(other.xDegreesPerPixel)) {
			return false;
		}
		if (Float.floatToIntBits(yDegreesPerPixel) != Float.floatToIntBits(other.yDegreesPerPixel)) {
			return false;
		}
		return true;
	}
}
