/*******************************************************************************
 * Software in the Sustain Ecosystem are Released Under Terms of Apache Software License 
 * 
 * This research has been supported by funding from the US National Science Foundation's CSSI program through awards 1931363, 1931324, 1931335, and 1931283. The project is a joint effort involving Colorado State University, Arizona State University, the University of California-Irvine, and the University of Maryland - Baltimore County. All redistributions of the software must also include this information. 
 * 
 * TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 * 
 * 
 * 1. Definitions.
 * 
 * "License" shall mean the terms and conditions for use, reproduction, and distribution as defined by Sections 1 through 9 of this document.
 * 
 * "Licensor" shall mean the copyright owner or entity authorized by the copyright owner that is granting the License.
 * 
 * "Legal Entity" shall mean the union of the acting entity and all other entities that control, are controlled by, or are under common control with that entity. For the purposes of this definition, "control" means (i) the power, direct or indirect, to cause the direction or management of such entity, whether by contract or otherwise, or (ii) ownership of fifty percent (50%) or more of the outstanding shares, or (iii) beneficial ownership of such entity.
 * 
 * "You" (or "Your") shall mean an individual or Legal Entity exercising permissions granted by this License.
 * 
 * "Source" form shall mean the preferred form for making modifications, including but not limited to software source code, documentation source, and configuration files.
 * 
 * "Object" form shall mean any form resulting from mechanical transformation or translation of a Source form, including but not limited to compiled object code, generated documentation, and conversions to other media types.
 * 
 * "Work" shall mean the work of authorship, whether in Source or Object form, made available under the License, as indicated by a copyright notice that is included in or attached to the work (an example is provided in the Appendix below).
 * 
 * "Derivative Works" shall mean any work, whether in Source or Object form, that is based on (or derived from) the Work and for which the editorial revisions, annotations, elaborations, or other modifications represent, as a whole, an original work of authorship. For the purposes of this License, Derivative Works shall not include works that remain separable from, or merely link (or bind by name) to the interfaces of, the Work and Derivative Works thereof.
 * 
 * "Contribution" shall mean any work of authorship, including the original version of the Work and any modifications or additions to that Work or Derivative Works thereof, that is intentionally submitted to Licensor for inclusion in the Work by the copyright owner or by an individual or Legal Entity authorized to submit on behalf of the copyright owner. For the purposes of this definition, "submitted" means any form of electronic, verbal, or written communication sent to the Licensor or its representatives, including but not limited to communication on electronic mailing lists, source code control systems, and issue tracking systems that are managed by, or on behalf of, the Licensor for the purpose of discussing and improving the Work, but excluding communication that is conspicuously marked or otherwise designated in writing by the copyright owner as "Not a Contribution."
 * 
 * "Contributor" shall mean Licensor and any individual or Legal Entity on behalf of whom a Contribution has been received by Licensor and subsequently incorporated within the Work.
 * 
 * 2. Grant of Copyright License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable copyright license to reproduce, prepare Derivative Works of, publicly display, publicly perform, sublicense, and distribute the Work and such Derivative Works in Source or Object form.
 * 
 * 3. Grant of Patent License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable (except as stated in this section) patent license to make, have made, use, offer to sell, sell, import, and otherwise transfer the Work, where such license applies only to those patent claims licensable by such Contributor that are necessarily infringed by their Contribution(s) alone or by combination of their Contribution(s) with the Work to which such Contribution(s) was submitted. If You institute patent litigation against any entity (including a cross-claim or counterclaim in a lawsuit) alleging that the Work or a Contribution incorporated within the Work constitutes direct or contributory patent infringement, then any patent licenses granted to You under this License for that Work shall terminate as of the date such litigation is filed.
 * 
 * 4. Redistribution. You may reproduce and distribute copies of the Work or Derivative Works thereof in any medium, with or without modifications, and in Source or Object form, provided that You meet the following conditions:
 * 
 * You must give any other recipients of the Work or Derivative Works a copy of this License; and
 * You must cause any modified files to carry prominent notices stating that You changed the files; and
 * You must retain, in the Source form of any Derivative Works that You distribute, all copyright, patent, trademark, and attribution notices from the Source form of the Work, excluding those notices that do not pertain to any part of the Derivative Works; and
 * If the Work includes a "NOTICE" text file as part of its distribution, then any Derivative Works that You distribute must include a readable copy of the attribution notices contained within such NOTICE file, excluding those notices that do not pertain to any part of the Derivative Works, in at least one of the following places: within a NOTICE text file distributed as part of the Derivative Works; within the Source form or documentation, if provided along with the Derivative Works; or, within a display generated by the Derivative Works, if and wherever such third-party notices normally appear. The contents of the NOTICE file are for informational purposes only and do not modify the License. You may add Your own attribution notices within Derivative Works that You distribute, alongside or as an addendum to the NOTICE text from the Work, provided that such additional attribution notices cannot be construed as modifying the License. 
 * 
 * You may add Your own copyright statement to Your modifications and may provide additional or different license terms and conditions for use, reproduction, or distribution of Your modifications, or for any such Derivative Works as a whole, provided Your use, reproduction, and distribution of the Work otherwise complies with the conditions stated in this License.
 * 5. Submission of Contributions. Unless You explicitly state otherwise, any Contribution intentionally submitted for inclusion in the Work by You to the Licensor shall be under the terms and conditions of this License, without any additional terms or conditions. Notwithstanding the above, nothing herein shall supersede or modify the terms of any separate license agreement you may have executed with Licensor regarding such Contributions.
 * 
 * 6. Trademarks. This License does not grant permission to use the trade names, trademarks, service marks, or product names of the Licensor, except as required for reasonable and customary use in describing the origin of the Work and reproducing the content of the NOTICE file.
 * 
 * 7. Disclaimer of Warranty. Unless required by applicable law or agreed to in writing, Licensor provides the Work (and each Contributor provides its Contributions) on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied, including, without limitation, any warranties or conditions of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A PARTICULAR PURPOSE. You are solely responsible for determining the appropriateness of using or redistributing the Work and assume any risks associated with Your exercise of permissions under this License.
 * 
 * 8. Limitation of Liability. In no event and under no legal theory, whether in tort (including negligence), contract, or otherwise, unless required by applicable law (such as deliberate and grossly negligent acts) or agreed to in writing, shall any Contributor be liable to You for damages, including any direct, indirect, special, incidental, or consequential damages of any character arising as a result of this License or out of the use or inability to use the Work (including but not limited to damages for loss of goodwill, work stoppage, computer failure or malfunction, or any and all other commercial damages or losses), even if such Contributor has been advised of the possibility of such damages.
 * 
 * 9. Accepting Warranty or Additional Liability. While redistributing the Work or Derivative Works thereof, You may choose to offer, and charge a fee for, acceptance of support, warranty, indemnity, or other liability obligations and/or rights consistent with this License. However, in accepting such obligations, You may act only on Your own behalf and on Your sole responsibility, not on behalf of any other Contributor, and only if You agree to indemnify, defend, and hold each Contributor harmless for any liability incurred by, or claims asserted against, such Contributor by reason of your accepting any such warranty or additional liability. 
 * 
 * END OF TERMS AND CONDITIONS
 ******************************************************************************/
package galileo.comm;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import galileo.dataset.SpatialHint;
import galileo.dataset.feature.FeatureType;
import galileo.event.Event;
import galileo.serialization.SerializationException;
import galileo.serialization.SerializationInputStream;
import galileo.serialization.SerializationOutputStream;
import galileo.util.GeoHash;
import galileo.util.Pair;

/**
 * For use by clients to create or delete file systems in galileo
 * 
 * @author kachikaran
 *
 */
public class FilesystemRequest implements Event {
	public static final int MAX_PRECISION = GeoHash.MAX_PRECISION/5;
	/**
	 * Name of the filesystem. 5-50 characters. Can include alphabets, numbers
	 * and hyphens
	 */
	private String name;

	/**
	 * FileSystemAction enum. Can be CREATE, DELETE or PERSIST
	 */
	private FilesystemAction action;

	/**
	 * The geohash precision that should be used to store the region-wise data.
	 * Range [2-FilesystemRequest.MAX_PRECISION]. Default is 4.
	 */
	private int precision;

	/**
	 * Temporal Type to use for the Network organization. Default is
	 * DAY_OF_MONTH. Can be YEAR, MONTH, DAY_OF_MONTH or HOUR_OF_DAY
	 */
	private TemporalType temporalType;

	/**
	 * The number of nodes in one group. The last group may not contain equal
	 * number as the rest depending on the number of nodes available. Default is
	 * 0 indicating the underlying network organization
	 */
	private int nodesPerGroup;

	/**
	 * The list featureName, featureType pairs such as Pair
	 * &lt;String,FeatureType&gt;("ch4", FeatureType.FLOAT)
	 */
	private List<Pair<String, FeatureType>> featureList;

	/**
	 * Should include a latitude and longitude hint which must be present in the
	 * feature list and their type should be of type FeatureType.FLOAT
	 */
	private SpatialHint spatialHint;
	
	
	private String temporalHint;
	
	/**
	 *  The uncertainty in join */
	
	private int spatialResolution;
	private int temporalResolution;
	private int spatialPartitioningType = 0;
	private List<String> summaryHints;
	
	
	
	
	
	private boolean isRasterized;
	
	

	/**
	 * @param name:
	 *            Name of the filesystem. 5-50 characters. Can include
	 *            alphabets, numbers and hyphens
	 * @param action:
	 *            FileSystemAction enum. Can be CREATE, DELETE or PERSIST
	 * @param featureList:
	 *            The list featureName, featureType pairs such as
	 *            Pair&lt;String,FeatureType&gt;("ch4", FeatureType.FLOAT)
	 * @param spatialHint:
	 *            Should include a latitude and longitude hint which must be
	 *            present in the feature list and their type should be of type
	 *            FeatureType.FLOAT
	 */
	public FilesystemRequest(String name, FilesystemAction action, List<Pair<String, FeatureType>> featureList,
			SpatialHint spatialHint) {
		if (name == null || name.trim().length() == 0 || !name.matches("[a-z0-9-]{5,50}"))
			throw new IllegalArgumentException(
					"name is required and must be lowercase having length at least 5 and at most 50 characters. "
							+ "alphabets, numbers and hyphens are allowed.");
		if (action == null)
			throw new IllegalArgumentException(
					"action cannot be null. must be one of the actions specified by galileo.comm.FileSystemAction");
		if (featureList != null && spatialHint == null)
			throw new IllegalArgumentException("Spatial hint is needed when feature list is provided");
		if (this.featureList != null && this.spatialHint != null) {
			boolean latOK = false;
			boolean lngOK = false;
			for (Pair<String, FeatureType> pair : this.featureList) {
				if (pair.a.equals(this.spatialHint.getLatitudeHint()) && pair.b == FeatureType.FLOAT)
					latOK = true;
				else if (pair.a.equals(this.spatialHint.getLongitudeHint()) && pair.b == FeatureType.FLOAT)
					lngOK = true;
			}
			if (!latOK)
				throw new IllegalArgumentException(
						"latitude hint must be one of the features in feature list and its type must be FeatureType.FLOAT");
			if (!lngOK)
				throw new IllegalArgumentException(
						"longitude hint must be one of the features in feature list and its type must be FeatureType.FLOAT");
		}
		this.name = name;
		this.precision = 4;
		this.nodesPerGroup = 0;
		this.temporalType = TemporalType.DAY_OF_MONTH;
		this.action = action;
		this.featureList = featureList;
		this.spatialHint = spatialHint;
	}
	
	
	public FilesystemRequest(String name, FilesystemAction action, List<Pair<String, FeatureType>> featureList,
			SpatialHint spatialHint, int spatialResolution, int temporalResolution, boolean isRasterized, String temporalHint) {
		if (name == null || name.trim().length() == 0 || !name.matches("[a-z0-9-]{5,50}"))
			throw new IllegalArgumentException(
					"name is required and must be lowercase having length at least 5 and at most 50 characters. "
							+ "alphabets, numbers and hyphens are allowed.");
		if (action == null)
			throw new IllegalArgumentException(
					"action cannot be null. must be one of the actions specified by galileo.comm.FileSystemAction");
		if (featureList != null && spatialHint == null)
			throw new IllegalArgumentException("Spatial hint is needed when feature list is provided");
		if (this.featureList != null && this.spatialHint != null) {
			boolean latOK = false;
			boolean lngOK = false;
			for (Pair<String, FeatureType> pair : this.featureList) {
				if (pair.a.equals(this.spatialHint.getLatitudeHint()) && pair.b == FeatureType.FLOAT)
					latOK = true;
				else if (pair.a.equals(this.spatialHint.getLongitudeHint()) && pair.b == FeatureType.FLOAT)
					lngOK = true;
			}
			if (!latOK)
				throw new IllegalArgumentException(
						"latitude hint must be one of the features in feature list and its type must be FeatureType.FLOAT");
			if (!lngOK)
				throw new IllegalArgumentException(
						"longitude hint must be one of the features in feature list and its type must be FeatureType.FLOAT");
		}
		this.name = name;
		this.precision = 4;
		this.nodesPerGroup = 0;
		this.temporalType = TemporalType.DAY_OF_MONTH;
		this.action = action;
		this.featureList = featureList;
		this.spatialHint = spatialHint;
		this.temporalHint = temporalHint;
		this.spatialResolution = spatialResolution;
		this.temporalResolution = temporalResolution;
		this.isRasterized = isRasterized;
	}

	public String getFeatures() {
		if (this.featureList == null)
			return null;
		StringBuffer sb = new StringBuffer();
		for (Pair<String, FeatureType> pair : this.featureList) {
			sb.append(pair.a + ":" + pair.b.toInt() + ",");
		}
		sb.setLength(sb.length() - 1);
		return sb.toString();
	}

	private boolean hasFeatures() {
		return this.featureList != null;
	}

	private boolean hasSpatialHint() {
		return this.spatialHint != null;
	}

	public List<Pair<String, FeatureType>> getFeatureList() {
		return this.featureList;
	}

	private List<Pair<String, FeatureType>> getFeatureList(String features) {
		if (features == null)
			return null;
		String[] pairs = features.split(",");
		this.featureList = new ArrayList<>();
		for (String pair : pairs) {
			String[] pairSplit = pair.split(":");
			this.featureList.add(
					new Pair<String, FeatureType>(pairSplit[0], FeatureType.fromInt(Integer.parseInt(pairSplit[1]))));
		}
		return this.featureList;
	}

	public SpatialHint getSpatialHint() {
		return this.spatialHint;
	}

	public void setPrecision(int precision) {
		if (precision >=2 && precision <= MAX_PRECISION)
			this.precision = precision;
	}

	public void setTemporalType(TemporalType temporalType) {
		if (temporalType != null) {
			this.temporalType = temporalType;
		}
	}

	public void setNodesPerGroup(int numNodes) {
		if (numNodes > 0)
			this.nodesPerGroup = numNodes;
	}

	public int getNodesPerGroup() {
		return this.nodesPerGroup;
	}

	public int getPrecision() {
		return this.precision;
	}

	public String getName() {
		return this.name;
	}

	public String getTemporalString() {
		return this.temporalType.name();
	}

	public int getTemporalValue() {
		return this.temporalType.getType();
	}

	public TemporalType getTemporalType() {
		return this.temporalType;
	}

	public FilesystemAction getAction() {
		return this.action;
	}

	@Deserialize
	public FilesystemRequest(SerializationInputStream in) throws IOException, SerializationException {
		this.name = in.readString();
		this.precision = in.readInt();
		this.action = FilesystemAction.fromAction(in.readString());
		this.temporalType = TemporalType.fromType(in.readInt());
		this.nodesPerGroup = in.readInt();
		if (in.readBoolean())
			this.featureList = getFeatureList(in.readString());
		if (in.readBoolean())
			this.spatialHint = new SpatialHint(in);
		this.temporalHint = in.readString();
		this.spatialResolution = in.readInt();
		this.temporalResolution = in.readInt();
		this.spatialPartitioningType = in.readInt();
		this.isRasterized = in.readBoolean();
		
		boolean hasSummaryHints = in.readBoolean();
		
		
		this.summaryHints = new ArrayList<String>();
		if(hasSummaryHints)
			in.readStringCollection(this.summaryHints);
		
	}

	@Override
	public void serialize(SerializationOutputStream out) throws IOException {
		out.writeString(this.name);
		out.writeInt(this.precision);
		out.writeString(this.action.getAction());
		out.writeInt(this.temporalType.getType());
		out.writeInt(this.nodesPerGroup);
		out.writeBoolean(hasFeatures());
		if (hasFeatures())
			out.writeString(getFeatures());
		out.writeBoolean(hasSpatialHint());
		if (hasSpatialHint())
			this.spatialHint.serialize(out);
		out.writeString(this.temporalHint);
		out.writeInt(this.spatialResolution);
		out.writeInt(this.temporalResolution);
		out.writeInt(this.spatialPartitioningType);
		out.writeBoolean(this.isRasterized);
		
		boolean hasSummaryHints = false;
		if(summaryHints != null && summaryHints.size() > 0)
			hasSummaryHints = true;
		out.writeBoolean(hasSummaryHints);
		if(hasSummaryHints)
			out.writeStringCollection(summaryHints);

	}

	public int getSpatialResolution() {
		return spatialResolution;
	}

	public void setSpatialResolution(int uncertaintyPrecision) {
		this.spatialResolution = uncertaintyPrecision;
	}

	public boolean isRasterized() {
		return isRasterized;
	}

	public void setRasterized(boolean isRasterized) {
		this.isRasterized = isRasterized;
	}


	public int getTemporalResolution() {
		return temporalResolution;
	}


	public void setTemporalResolution(int temporalResolution) {
		this.temporalResolution = temporalResolution;
	}


	public String getTemporalHint() {
		return temporalHint;
	}


	public void setTemporalHint(String temporalHint) {
		this.temporalHint = temporalHint;
	}


	public int getSpatialPartitioningType() {
		return spatialPartitioningType;
	}


	public void setSpatialPartitioningType(int spatialPartitioningType) {
		this.spatialPartitioningType = spatialPartitioningType;
	}


	public List<String> getSummaryHints() {
		return summaryHints;
	}


	public void setSummaryHints(List<String> summaryHints) {
		this.summaryHints = summaryHints;
	}


	public void setName(String name) {
		this.name = name;
	}


	public void setFeatureList(List<Pair<String, FeatureType>> featureList) {
		this.featureList = featureList;
	}


	public void setSpatialHint(SpatialHint spatialHint) {
		this.spatialHint = spatialHint;
	}

}
