/*******************************************************************************
 * Software in the Sustain Ecosystem are Released Under Terms of Apache Software License 
 * 
 * This research has been supported by funding from the US National Science Foundation's CSSI program through awards 1931363, 1931324, 1931335, and 1931283. The project is a joint effort involving Colorado State University, Arizona State University, the University of California-Irvine, and the University of Maryland - Baltimore County. All redistributions of the software must also include this information. 
 * 
 * TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 * 
 * 
 * 1. Definitions.
 * 
 * "License" shall mean the terms and conditions for use, reproduction, and distribution as defined by Sections 1 through 9 of this document.
 * 
 * "Licensor" shall mean the copyright owner or entity authorized by the copyright owner that is granting the License.
 * 
 * "Legal Entity" shall mean the union of the acting entity and all other entities that control, are controlled by, or are under common control with that entity. For the purposes of this definition, "control" means (i) the power, direct or indirect, to cause the direction or management of such entity, whether by contract or otherwise, or (ii) ownership of fifty percent (50%) or more of the outstanding shares, or (iii) beneficial ownership of such entity.
 * 
 * "You" (or "Your") shall mean an individual or Legal Entity exercising permissions granted by this License.
 * 
 * "Source" form shall mean the preferred form for making modifications, including but not limited to software source code, documentation source, and configuration files.
 * 
 * "Object" form shall mean any form resulting from mechanical transformation or translation of a Source form, including but not limited to compiled object code, generated documentation, and conversions to other media types.
 * 
 * "Work" shall mean the work of authorship, whether in Source or Object form, made available under the License, as indicated by a copyright notice that is included in or attached to the work (an example is provided in the Appendix below).
 * 
 * "Derivative Works" shall mean any work, whether in Source or Object form, that is based on (or derived from) the Work and for which the editorial revisions, annotations, elaborations, or other modifications represent, as a whole, an original work of authorship. For the purposes of this License, Derivative Works shall not include works that remain separable from, or merely link (or bind by name) to the interfaces of, the Work and Derivative Works thereof.
 * 
 * "Contribution" shall mean any work of authorship, including the original version of the Work and any modifications or additions to that Work or Derivative Works thereof, that is intentionally submitted to Licensor for inclusion in the Work by the copyright owner or by an individual or Legal Entity authorized to submit on behalf of the copyright owner. For the purposes of this definition, "submitted" means any form of electronic, verbal, or written communication sent to the Licensor or its representatives, including but not limited to communication on electronic mailing lists, source code control systems, and issue tracking systems that are managed by, or on behalf of, the Licensor for the purpose of discussing and improving the Work, but excluding communication that is conspicuously marked or otherwise designated in writing by the copyright owner as "Not a Contribution."
 * 
 * "Contributor" shall mean Licensor and any individual or Legal Entity on behalf of whom a Contribution has been received by Licensor and subsequently incorporated within the Work.
 * 
 * 2. Grant of Copyright License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable copyright license to reproduce, prepare Derivative Works of, publicly display, publicly perform, sublicense, and distribute the Work and such Derivative Works in Source or Object form.
 * 
 * 3. Grant of Patent License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable (except as stated in this section) patent license to make, have made, use, offer to sell, sell, import, and otherwise transfer the Work, where such license applies only to those patent claims licensable by such Contributor that are necessarily infringed by their Contribution(s) alone or by combination of their Contribution(s) with the Work to which such Contribution(s) was submitted. If You institute patent litigation against any entity (including a cross-claim or counterclaim in a lawsuit) alleging that the Work or a Contribution incorporated within the Work constitutes direct or contributory patent infringement, then any patent licenses granted to You under this License for that Work shall terminate as of the date such litigation is filed.
 * 
 * 4. Redistribution. You may reproduce and distribute copies of the Work or Derivative Works thereof in any medium, with or without modifications, and in Source or Object form, provided that You meet the following conditions:
 * 
 * You must give any other recipients of the Work or Derivative Works a copy of this License; and
 * You must cause any modified files to carry prominent notices stating that You changed the files; and
 * You must retain, in the Source form of any Derivative Works that You distribute, all copyright, patent, trademark, and attribution notices from the Source form of the Work, excluding those notices that do not pertain to any part of the Derivative Works; and
 * If the Work includes a "NOTICE" text file as part of its distribution, then any Derivative Works that You distribute must include a readable copy of the attribution notices contained within such NOTICE file, excluding those notices that do not pertain to any part of the Derivative Works, in at least one of the following places: within a NOTICE text file distributed as part of the Derivative Works; within the Source form or documentation, if provided along with the Derivative Works; or, within a display generated by the Derivative Works, if and wherever such third-party notices normally appear. The contents of the NOTICE file are for informational purposes only and do not modify the License. You may add Your own attribution notices within Derivative Works that You distribute, alongside or as an addendum to the NOTICE text from the Work, provided that such additional attribution notices cannot be construed as modifying the License. 
 * 
 * You may add Your own copyright statement to Your modifications and may provide additional or different license terms and conditions for use, reproduction, or distribution of Your modifications, or for any such Derivative Works as a whole, provided Your use, reproduction, and distribution of the Work otherwise complies with the conditions stated in this License.
 * 5. Submission of Contributions. Unless You explicitly state otherwise, any Contribution intentionally submitted for inclusion in the Work by You to the Licensor shall be under the terms and conditions of this License, without any additional terms or conditions. Notwithstanding the above, nothing herein shall supersede or modify the terms of any separate license agreement you may have executed with Licensor regarding such Contributions.
 * 
 * 6. Trademarks. This License does not grant permission to use the trade names, trademarks, service marks, or product names of the Licensor, except as required for reasonable and customary use in describing the origin of the Work and reproducing the content of the NOTICE file.
 * 
 * 7. Disclaimer of Warranty. Unless required by applicable law or agreed to in writing, Licensor provides the Work (and each Contributor provides its Contributions) on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied, including, without limitation, any warranties or conditions of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A PARTICULAR PURPOSE. You are solely responsible for determining the appropriateness of using or redistributing the Work and assume any risks associated with Your exercise of permissions under this License.
 * 
 * 8. Limitation of Liability. In no event and under no legal theory, whether in tort (including negligence), contract, or otherwise, unless required by applicable law (such as deliberate and grossly negligent acts) or agreed to in writing, shall any Contributor be liable to You for damages, including any direct, indirect, special, incidental, or consequential damages of any character arising as a result of this License or out of the use or inability to use the Work (including but not limited to damages for loss of goodwill, work stoppage, computer failure or malfunction, or any and all other commercial damages or losses), even if such Contributor has been advised of the possibility of such damages.
 * 
 * 9. Accepting Warranty or Additional Liability. While redistributing the Work or Derivative Works thereof, You may choose to offer, and charge a fee for, acceptance of support, warranty, indemnity, or other liability obligations and/or rights consistent with this License. However, in accepting such obligations, You may act only on Your own behalf and on Your sole responsibility, not on behalf of any other Contributor, and only if You agree to indemnify, defend, and hold each Contributor harmless for any liability incurred by, or claims asserted against, such Contributor by reason of your accepting any such warranty or additional liability. 
 * 
 * END OF TERMS AND CONDITIONS
 ******************************************************************************/
package galileo.dht;

import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.logging.Level;
import java.util.logging.Logger;

import galileo.comm.DataIntegrationFinalResponse;
import galileo.comm.DataIntegrationResponse;
import galileo.comm.GalileoEventMap;
import galileo.comm.SurveyEventResponse;
import galileo.comm.SurveyResponse;
import galileo.comm.TrainingDataEvent;
import galileo.comm.TrainingDataResponse;
import galileo.event.BasicEventWrapper;
import galileo.event.Event;
import galileo.event.EventContext;
import galileo.net.ClientMessageRouter;
import galileo.net.GalileoMessage;
import galileo.net.MessageListener;
import galileo.net.NetworkDestination;
import galileo.net.RequestListener;
import galileo.serialization.SerializationException;
import galileo.util.MyPorter;

/**
 * This class will collect the responses from all the nodes of galileo and then
 * transfers the result to the listener. Used by the {@link StorageNode} class.
 * 
 * @author sapmitra
 */
public class SurveyRequestHandler implements MessageListener {

	private static final Logger logger = Logger.getLogger("galileo");
	private GalileoEventMap eventMap;
	private BasicEventWrapper eventWrapper;
	private ClientMessageRouter router;
	private AtomicInteger expectedResponses;
	private Collection<NetworkDestination> nodes;
	private EventContext clientContext;
	private List<GalileoMessage> responses;
	private RequestListener requestListener;
	private Event response;
	private long elapsedTime;
	private int numTrainingPoints;
	private AtomicInteger expectedTrainingResponses;
	private String allTrainingData;
	private String fsName;
	private String featureName;
	private double latEps,lonEps,timeEps;
	private String storagePath;
	private String currentNode;
	private boolean hasModel;
	private String model;
	
	

	public SurveyRequestHandler(Collection<NetworkDestination> nodes, EventContext clientContext,
			int numTrainingPoints, String fsName, String featureName, double d, double e, double f, 
			String trainingResultsDir, String nodeString, RequestListener listener, boolean hasModel, String model) throws IOException {
		this.nodes = nodes;
		this.clientContext = clientContext;
		this.requestListener = listener;

		this.router = new ClientMessageRouter(true);
		this.router.addListener(this);
		this.responses = new ArrayList<GalileoMessage>();
		this.eventMap = new GalileoEventMap();
		this.eventWrapper = new BasicEventWrapper(this.eventMap);
		this.expectedResponses = new AtomicInteger(this.nodes.size());
		this.numTrainingPoints = numTrainingPoints;
		this.expectedTrainingResponses = new AtomicInteger(0);
		this.fsName = fsName;
		this.featureName = featureName;
		this.latEps = d;
		this.lonEps = e;
		this.timeEps = f;
		this.allTrainingData = "";
		String eventId = String.valueOf(System.currentTimeMillis());
		this.storagePath = trainingResultsDir + "/trainingData" + eventId;
		this.currentNode = nodeString;
		
		this.hasModel = hasModel;
		
		this.model = model;
		
	}

	public void closeRequest() {
		silentClose(); // closing the router to make sure that no new responses
						// are added.
		
		try {
			if(this.response instanceof SurveyResponse) {
				
				SurveyResponse actualResponse = (SurveyResponse) this.response;
				
				actualResponse.setOutputPath(this.storagePath+".blk");
				actualResponse.setNodeString(currentNode);
			}
				
		} catch (Exception e) {
			logger.log(Level.SEVERE,
					"An unknown exception occurred while processing the response message. Details follow:"
							+ e.getMessage(), e);
		}
		
		this.requestListener.onRequestCompleted(this.response, clientContext, this);
	}

	@Override
	public void onMessage(GalileoMessage message) {
		try {
			
			Event event = this.eventWrapper.unwrap(message);
			
			// SURVEY RESPONSE CAME IN
			if(event instanceof SurveyEventResponse) {
				
				logger.log(Level.INFO, "RIKI: SURVEY LEVEL 1 RESPONSE RECEIVED");
				if (null != message)
					this.responses.add(message);
				int awaitedResponses = this.expectedResponses.decrementAndGet();
				logger.log(Level.INFO, "Awaiting " + awaitedResponses + " more message(s)");
				
				if (awaitedResponses <= 0) {
					this.elapsedTime = System.currentTimeMillis() - this.elapsedTime;
					
					List<String> pathInfos = new ArrayList<String>();
					List<String> blocks = new ArrayList<String>();
					List<Long> recordCounts = new ArrayList<Long>();
					
					performStartifiedSamplingAndRequest(pathInfos, blocks, recordCounts);
				}
				
			} else if (event instanceof TrainingDataResponse) {
				
				int awaitedResponses = this.expectedTrainingResponses.decrementAndGet();
				TrainingDataResponse rsp = (TrainingDataResponse)event;
				
				logger.log(Level.INFO, "RIKI: TRAINING DATA RESPONSE RECEIVED");
				if(rsp.getDataPoints() != null && rsp.getDataPoints().length() > 0)
					allTrainingData+= rsp.getDataPoints();
				
				/* The close will happen when the second set of requests have been replied to */
				
				if (awaitedResponses <= 0) {
					
					FileOutputStream fos = null;
					try {
						fos = new FileOutputStream(this.storagePath+".blk");
						fos.write("lat,long,epoch,beta,err,actval\n".getBytes());
						fos.write(allTrainingData.getBytes("UTF-8"));
						fos.close();
						
					} catch (IOException e) {
						logger.log(Level.SEVERE, "Something went wrong while storing training data to the filesystem.", e);
						this.storagePath = null;
					}
					
					this.elapsedTime = System.currentTimeMillis() - this.elapsedTime;
					logger.log(Level.INFO, "Closing the survey request and sending back the response.");
					new Thread() {
						public void run() {
							SurveyRequestHandler.this.closeRequest();
						}
					}.start();
				}
			}
		} catch (IOException | SerializationException e) {
			logger.log(Level.SEVERE, "An exception occurred while processing the response message. Details follow:"
					+ e.getMessage(), e);
		} catch (Exception e) {
			logger.log(Level.SEVERE,
					"An unknown exception occurred while processing the response message. Details follow:"
							+ e.getMessage(), e);
		}
	}

	/**
	 * Perform stratified sampling based on number of records and send request for training data
	 * @author sapmitra
	 * @param pathInfos
	 * @param blocks
	 * @param recordCounts
	 */
	private void performStartifiedSamplingAndRequest(List<String> pathInfos, List<String> blocks, List<Long> recordCounts) {
		
		long totalRecordCount = 0;
		int expectedTrainingMsgs = 0;
		
		// PathInfos and blocks has same number of entries
		for (GalileoMessage gresponse : this.responses) {
			Event event;
			try {
				event = this.eventWrapper.unwrap(gresponse);
				if(event instanceof SurveyEventResponse) {
					
					SurveyEventResponse eventResponse = (SurveyEventResponse) event;
					
					if(eventResponse.getPathInfos() != null && eventResponse.getPathInfos().size() > 0) {
						pathInfos.addAll(eventResponse.getPathInfos());
						blocks.addAll(eventResponse.getBlocks());
						recordCounts.addAll(eventResponse.getRecordCounts());
						
						for(long recC : eventResponse.getRecordCounts())
							totalRecordCount+= recC;
					}
					
				}
			} catch (IOException | SerializationException e) {
				logger.log(Level.SEVERE, "An exception occurred while processing the response message. Details follow:"
						+ e.getMessage(), e);
			} catch (Exception e) {
				logger.log(Level.SEVERE,
						"An unknown exception occurred while processing the response message. Details follow:"
								+ e.getMessage(), e);
			}
		}
		
		// BUILDING THE NODEWISE REQUEST
		Map<NetworkDestination, TrainingRequirements> nodeWiseRequest = new HashMap<NetworkDestination, TrainingRequirements>();
		
		// STRATIFIED SAMPLING AND REQUEST CREATED IN ONE GO
		// for each block in each path in each node
		for(int i=0; i < pathInfos.size(); i++) {
			
			String pathInfo = pathInfos.get(i);
			System.out.println("RIKI:PATH RECV "+pathInfo);
			NetworkDestination nd = extractNodeInfo(pathInfo);
			
			// The number of training points to be extracted from this block
			double fraction = (double)recordCounts.get(i)/(double)totalRecordCount;
			
			int numPoints = (int)(Math.floor(fraction*numTrainingPoints));
			logger.info("FRACTION: "+ fraction+" "+numPoints);
			
			String block = blocks.get(i);
			
			if(numPoints > 1) {
				logger.info("NUMPOINTS "+ block+" "+numPoints);
				TrainingRequirements tr;
				if(nodeWiseRequest.get(nd) == null) {
					expectedTrainingMsgs++;
					tr = new TrainingRequirements(new ArrayList<String>(), new ArrayList<Integer>(), new ArrayList<String>());
					nodeWiseRequest.put(nd, tr);
				} else {
					tr = nodeWiseRequest.get(nd);
				}
				tr.addPathInfo(pathInfo);
				tr.addBlockPath(block);
				tr.addNumPoints(numPoints);
				
			}
			
		}
		this.expectedTrainingResponses = new AtomicInteger(expectedTrainingMsgs);
		
		// SEND OUT TRAINING DATA REQUEST
		for(NetworkDestination nd : nodeWiseRequest.keySet()) {
			
			TrainingRequirements tr = nodeWiseRequest.get(nd);
			TrainingDataEvent tde = new TrainingDataEvent(tr, fsName, featureName, latEps, lonEps, timeEps, hasModel, model);
			GalileoMessage mrequest;
			
			try {
				
				mrequest = this.eventWrapper.wrap(tde);
				this.router.sendMessage(nd, mrequest);
				logger.info("RIKI: TRAINING DATA REQUEST SENT TO " + nd.toString());
				
			} catch (IOException e) {
				logger.severe("ERROR OCCURED WHILE REQUESTING FOR TRAINING DATA: "+e.toString());
			}
			
		}
		
	}
	
	private NetworkDestination extractNodeInfo(String pathInfo) {
		String nodePort = pathInfo.split("\\$\\$")[0];
		String[] tokens = nodePort.split(":");
		String nodeName = tokens[0];
		int port = Integer.valueOf(tokens[1]);
		
		for(NetworkDestination n : nodes) {
			
			if(n.getHostname().equals(nodeName) && n.getPort() == port) {
				return n;
			}
		}
		return null;
		
	}

	public static void main(String arg[]) {
		String s = "hello:world".split(":")[0];
		System.out.println(s);
		List<String> sa = new ArrayList<String>();
		sa.addAll(null);
		
	}
	/**
	 * Handles the client request on behalf of the node that received the
	 * request
	 * 
	 * @param request
	 *            - This must be a server side event: Generic Event or
	 *            QueryEvent
	 * @param response
	 */
	public void handleRequest(Event request, Event response) {
		try {
			this.response = response;
			GalileoMessage mrequest = this.eventWrapper.wrap(request);
			for (NetworkDestination node : nodes) {
				this.router.sendMessage(node, mrequest);
				logger.info("Request sent to " + node.toString());
			}
			this.elapsedTime = System.currentTimeMillis();
		} catch (IOException e) {
			logger.log(Level.INFO,
					"Failed to send request to other nodes in the network. Details follow: " + e.getMessage());
		}
	}

	public void silentClose() {
		try {
			this.router.forceShutdown();
		} catch (Exception e) {
			logger.log(Level.INFO, "Failed to shutdown the completed client request handler: ", e);
		}
	}

	@Override
	public void onConnect(NetworkDestination endpoint) {

	}

	@Override
	public void onDisconnect(NetworkDestination endpoint) {

	}
}
