/*******************************************************************************
 * Software in the Sustain Ecosystem are Released Under Terms of Apache Software License 
 * 
 * This research has been supported by funding from the US National Science Foundation's CSSI program through awards 1931363, 1931324, 1931335, and 1931283. The project is a joint effort involving Colorado State University, Arizona State University, the University of California-Irvine, and the University of Maryland - Baltimore County. All redistributions of the software must also include this information. 
 * 
 * TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 * 
 * 
 * 1. Definitions.
 * 
 * "License" shall mean the terms and conditions for use, reproduction, and distribution as defined by Sections 1 through 9 of this document.
 * 
 * "Licensor" shall mean the copyright owner or entity authorized by the copyright owner that is granting the License.
 * 
 * "Legal Entity" shall mean the union of the acting entity and all other entities that control, are controlled by, or are under common control with that entity. For the purposes of this definition, "control" means (i) the power, direct or indirect, to cause the direction or management of such entity, whether by contract or otherwise, or (ii) ownership of fifty percent (50%) or more of the outstanding shares, or (iii) beneficial ownership of such entity.
 * 
 * "You" (or "Your") shall mean an individual or Legal Entity exercising permissions granted by this License.
 * 
 * "Source" form shall mean the preferred form for making modifications, including but not limited to software source code, documentation source, and configuration files.
 * 
 * "Object" form shall mean any form resulting from mechanical transformation or translation of a Source form, including but not limited to compiled object code, generated documentation, and conversions to other media types.
 * 
 * "Work" shall mean the work of authorship, whether in Source or Object form, made available under the License, as indicated by a copyright notice that is included in or attached to the work (an example is provided in the Appendix below).
 * 
 * "Derivative Works" shall mean any work, whether in Source or Object form, that is based on (or derived from) the Work and for which the editorial revisions, annotations, elaborations, or other modifications represent, as a whole, an original work of authorship. For the purposes of this License, Derivative Works shall not include works that remain separable from, or merely link (or bind by name) to the interfaces of, the Work and Derivative Works thereof.
 * 
 * "Contribution" shall mean any work of authorship, including the original version of the Work and any modifications or additions to that Work or Derivative Works thereof, that is intentionally submitted to Licensor for inclusion in the Work by the copyright owner or by an individual or Legal Entity authorized to submit on behalf of the copyright owner. For the purposes of this definition, "submitted" means any form of electronic, verbal, or written communication sent to the Licensor or its representatives, including but not limited to communication on electronic mailing lists, source code control systems, and issue tracking systems that are managed by, or on behalf of, the Licensor for the purpose of discussing and improving the Work, but excluding communication that is conspicuously marked or otherwise designated in writing by the copyright owner as "Not a Contribution."
 * 
 * "Contributor" shall mean Licensor and any individual or Legal Entity on behalf of whom a Contribution has been received by Licensor and subsequently incorporated within the Work.
 * 
 * 2. Grant of Copyright License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable copyright license to reproduce, prepare Derivative Works of, publicly display, publicly perform, sublicense, and distribute the Work and such Derivative Works in Source or Object form.
 * 
 * 3. Grant of Patent License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable (except as stated in this section) patent license to make, have made, use, offer to sell, sell, import, and otherwise transfer the Work, where such license applies only to those patent claims licensable by such Contributor that are necessarily infringed by their Contribution(s) alone or by combination of their Contribution(s) with the Work to which such Contribution(s) was submitted. If You institute patent litigation against any entity (including a cross-claim or counterclaim in a lawsuit) alleging that the Work or a Contribution incorporated within the Work constitutes direct or contributory patent infringement, then any patent licenses granted to You under this License for that Work shall terminate as of the date such litigation is filed.
 * 
 * 4. Redistribution. You may reproduce and distribute copies of the Work or Derivative Works thereof in any medium, with or without modifications, and in Source or Object form, provided that You meet the following conditions:
 * 
 * You must give any other recipients of the Work or Derivative Works a copy of this License; and
 * You must cause any modified files to carry prominent notices stating that You changed the files; and
 * You must retain, in the Source form of any Derivative Works that You distribute, all copyright, patent, trademark, and attribution notices from the Source form of the Work, excluding those notices that do not pertain to any part of the Derivative Works; and
 * If the Work includes a "NOTICE" text file as part of its distribution, then any Derivative Works that You distribute must include a readable copy of the attribution notices contained within such NOTICE file, excluding those notices that do not pertain to any part of the Derivative Works, in at least one of the following places: within a NOTICE text file distributed as part of the Derivative Works; within the Source form or documentation, if provided along with the Derivative Works; or, within a display generated by the Derivative Works, if and wherever such third-party notices normally appear. The contents of the NOTICE file are for informational purposes only and do not modify the License. You may add Your own attribution notices within Derivative Works that You distribute, alongside or as an addendum to the NOTICE text from the Work, provided that such additional attribution notices cannot be construed as modifying the License. 
 * 
 * You may add Your own copyright statement to Your modifications and may provide additional or different license terms and conditions for use, reproduction, or distribution of Your modifications, or for any such Derivative Works as a whole, provided Your use, reproduction, and distribution of the Work otherwise complies with the conditions stated in this License.
 * 5. Submission of Contributions. Unless You explicitly state otherwise, any Contribution intentionally submitted for inclusion in the Work by You to the Licensor shall be under the terms and conditions of this License, without any additional terms or conditions. Notwithstanding the above, nothing herein shall supersede or modify the terms of any separate license agreement you may have executed with Licensor regarding such Contributions.
 * 
 * 6. Trademarks. This License does not grant permission to use the trade names, trademarks, service marks, or product names of the Licensor, except as required for reasonable and customary use in describing the origin of the Work and reproducing the content of the NOTICE file.
 * 
 * 7. Disclaimer of Warranty. Unless required by applicable law or agreed to in writing, Licensor provides the Work (and each Contributor provides its Contributions) on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied, including, without limitation, any warranties or conditions of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A PARTICULAR PURPOSE. You are solely responsible for determining the appropriateness of using or redistributing the Work and assume any risks associated with Your exercise of permissions under this License.
 * 
 * 8. Limitation of Liability. In no event and under no legal theory, whether in tort (including negligence), contract, or otherwise, unless required by applicable law (such as deliberate and grossly negligent acts) or agreed to in writing, shall any Contributor be liable to You for damages, including any direct, indirect, special, incidental, or consequential damages of any character arising as a result of this License or out of the use or inability to use the Work (including but not limited to damages for loss of goodwill, work stoppage, computer failure or malfunction, or any and all other commercial damages or losses), even if such Contributor has been advised of the possibility of such damages.
 * 
 * 9. Accepting Warranty or Additional Liability. While redistributing the Work or Derivative Works thereof, You may choose to offer, and charge a fee for, acceptance of support, warranty, indemnity, or other liability obligations and/or rights consistent with this License. However, in accepting such obligations, You may act only on Your own behalf and on Your sole responsibility, not on behalf of any other Contributor, and only if You agree to indemnify, defend, and hold each Contributor harmless for any liability incurred by, or claims asserted against, such Contributor by reason of your accepting any such warranty or additional liability. 
 * 
 * END OF TERMS AND CONDITIONS
 ******************************************************************************/
package galileo.dht;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.ThreadLocalRandom;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.logging.Level;
import java.util.logging.Logger;

import galileo.comm.GalileoEventMap;
import galileo.comm.HeartbeatRequest;
import galileo.comm.HeartbeatResponse;
import galileo.dataset.Metadata;
import galileo.dataset.SpatialProperties;
import galileo.dataset.SpatialRange;
import galileo.dht.hash.HashException;
import galileo.event.BasicEventWrapper;
import galileo.event.Event;
import galileo.fs.GeospatialFileSystem;
import galileo.graph.CliqueContainer;
import galileo.graph.TopCliqueFinder;
import galileo.net.ClientMessageRouter;
import galileo.net.GalileoMessage;
import galileo.net.MessageListener;
import galileo.net.NetworkDestination;
import galileo.serialization.SerializationException;
import galileo.util.GeoHash;

/**
 * This class will collect the responses from all the nodes of galileo and then
 * transfers the result to the listener. Used by the {@link StorageNode} class.
 * 
 * @author sapmitra
 */
public class TileHandoffHandler implements MessageListener {

	private static final Logger logger = Logger.getLogger("galileo");
	
	private long WAIT_TIME = 3000l;
	
	// WE ARE GOING TO RETRY FOR NEW HELPER NODES THIS MANY TIMES
	private int RETRY_COUNT = 10;
	
	private boolean waitTimeCheck = true;
	
	private GalileoEventMap eventMap;
	private BasicEventWrapper eventWrapper;
	private ClientMessageRouter router;
	private AtomicInteger expectedResponses;
	private Collection<NetworkDestination> nodes;
	private Map<String, NetworkDestination> nodeStringToNodeMap;
	private List<GalileoMessage> responses;
	private long elapsedTime;
	
	private Map<String, CliqueContainer> topKCliques;
	
	private GeospatialFileSystem fs;
	
	private Partitioner<Metadata> partitioner;
	
	private boolean somethingGotReplicated = false;
	
	// IF THIS IS THE FIRST TIME REQUESTING FOR HELPER NODES
	private boolean firstTry = true;
	
	
	private StorageNode currentNode;
	

	public TileHandoffHandler(List<NetworkDestination> allOtherNodes, GeospatialFileSystem fs, NetworkDestination currentNode, long waitTime, StorageNode thisNode) throws IOException {
		
		this.nodes = allOtherNodes;

		nodeStringToNodeMap = new HashMap<String, NetworkDestination>();
		
		for(NetworkDestination nd : allOtherNodes) {
			
			nodeStringToNodeMap.put(nd.stringRepresentation(), nd);
			
		}
		
		this.router = new ClientMessageRouter(true);
		this.router.addListener(this);
		this.responses = new ArrayList<GalileoMessage>();
		this.eventMap = new GalileoEventMap();
		this.eventWrapper = new BasicEventWrapper(this.eventMap);
		
		this.expectedResponses = new AtomicInteger(0);
		this.currentNode = thisNode;
		this.WAIT_TIME = waitTime;
		
		// FIND TOP N CLIQUES
		// CLIQUE IS THE UNIT OF DATA TRANSFER IN THIS SYSTEM
		topKCliques = TopCliqueFinder.getTopKCliques(fs.getStCache(), fs.getSpatialPartitioningType());
		
		logger.info("TOP CLIQUES: "+ topKCliques.keySet()+" SIZE: "+topKCliques.values().size());
		
		// CALCULATING THE BITMAP COVERED BY THE TILES IN EACH PARTICULAR CLIQUE
		for(String k : topKCliques.keySet()) {
			
			CliqueContainer clique = topKCliques.get(k);
			
			// POPULATES A BITMAP INSIDE THE CLIQUE CONTAINER
			// STRICTLY MEANT TO BE HOUSED IN THE DISTRESSED NODE
			// NOT MEANT FOR THE HELPING NODE
			clique.calculateBitmap(fs);
			
			
		}
		
		partitioner = fs.getPartitioner();
		
		this.fs = fs;
		
	}

	
	/**
	 * HERE WE HANDLE :
	 * 1) INTERPRETING THE HEARTBEAT MESSAGES
	 * 2) PICKING TOP CLIQUES
	 * 3) SENDING OF CLIQUES TO RESPECTIVE NODES
	 * 4) CREATING ROUTING TABLES
	 * @author sapmitra
	 */
	public void closeRequest() {
		
		int responseCount = 0;
		
		for (GalileoMessage gresponse : this.responses) {
			
			responseCount++;
			
			Event event;
			
			try {
				event = this.eventWrapper.unwrap(gresponse);
				
				if (event instanceof HeartbeatResponse) {
					
					HeartbeatResponse eventResponse = (HeartbeatResponse) event;
					for (int i = 0; i< eventResponse.getDirection().size(); i++) {
						
						logger.info("RIKI: DISTRESS RESPONSE SAYS: "+eventResponse.getResultFlag());
						if(eventResponse.getResultFlag().get(i)) {
							// THIS CLIQUE HAS BEEN REPLICATED SUCCESSFULLY
							// REMOVE THIS CLIQUE FROM TOP CLIQUE
							// ADD THIS ENTRY TO THE ROUTING TABLE
							
							String cliqueKey = eventResponse.getGeohashOfClique().get(i);
							
							String nodeKey = eventResponse.getNodeString();
							
							NetworkDestination nd = nodeStringToNodeMap.get(nodeKey);
							topKCliques.get(cliqueKey).setReplicatedNode(nd);
							
							somethingGotReplicated = true;
							
						} else {
							// FAILED TO REPLICATE THIS, TRY ANOTHER NODE
							// KEEP THIS ENTRY IN TOP CLIQUES
							// CALCULATE THE NEXT ANTIPODE GEOHASH
							
							// THIS WILL NOT BE CALLED IF THERE WAS A FAILURE
							
						}
						
					}
					
				}
			} catch (IOException | SerializationException e) {
				logger.log(Level.SEVERE, "An exception occurred while processing the response message. Details follow:"
						+ e.getMessage(), e);
			} catch (Exception e) {
				logger.log(Level.SEVERE, "An unknown exception occurred while processing the response message. Details follow:"
								+ e.getMessage(), e);
			}
		}
		
		int totalHomesFound = 0;
		
		for(CliqueContainer clique : topKCliques.values()) {

			if(clique.getReplicatedNode() != null)
				totalHomesFound++;
		}
		
		// IF NOT ALL CLIQUES HAVE FOUND A HOME, REDO THE PROCESS
		if(topKCliques.size() != totalHomesFound && RETRY_COUNT >= 0) {
			
			logger.info("RIKI: NOT ALL CLIQUES FOUND HOME....RETRYING FOR REMAINING CLIQUES");
			RETRY_COUNT--;
			handleRequest();
			
		} else {
			
			logger.info("RIKI: DISTRESS RESPONSES GETTING LOADED IN ROUTING TABLE....");
			silentClose(); // closing the router to make sure that no new responses are added.
			
			// POPULATE ROUTING TABLE WITH CLIQUES THAT GOT REPLICATED
			fs.populateRoutingTable(topKCliques);
			
			// IF AT ALL SOMETHING GOT REPLICATED
			if(somethingGotReplicated)
				currentNode.setHotspotHasBeenHandledTime(System.currentTimeMillis());
			
			logger.info("RIKI: HEARTBEAT COMPILED WITH "+responseCount+" MESSAGES");
			
			currentNode.setHotspotBeingHandled(false);
			
		}
		
		
		
		/*try {
			afterHeartbeatCheck();
		} catch (HashException | PartitionException e) {
			// TODO Auto-generated catch block
			logger.severe("RIKI: ERROR AFTER SUCCESSFUL HEARTBEAT");
		}*/
		
	}


	@Override
	public void onMessage(GalileoMessage message) {
		
		logger.info("RIKI: HEARTBEAT RESPONSE RECEIVED");
		
		if (null != message)
			this.responses.add(message);
		
		int awaitedResponses = this.expectedResponses.decrementAndGet();
		logger.log(Level.INFO, "Awaiting " + awaitedResponses + " more message(s)");
		
		
		if (awaitedResponses <= 0) {
			// PREVENT TIMEOUTCHECKER FROM MEDDLING
			this.waitTimeCheck = false;
			this.elapsedTime = System.currentTimeMillis() - this.elapsedTime;
			
			closeRequest();
				
		}
	}
	
	/**
	 * Given a geohash, find the node that houses it
	 * @author sapmitra
	 * @param geoHashAntipode
	 * @return
	 */
	public NodeInfo getNodeForGeoHash(String geoHashAntipode) {
		
		// FINDING THE NODE THAT HOUSES THE ANTIPODE GEOHASH
		SpatialRange spatialRange = GeoHash.decodeHash(geoHashAntipode);
		
		SpatialProperties spatialProperties = new SpatialProperties(spatialRange);
		Metadata metadata = new Metadata();
		metadata.setName(geoHashAntipode);
		metadata.setSpatialProperties(spatialProperties);
		
		NodeInfo targetNode = null;
		try {
			targetNode = partitioner.locateData(metadata);
		} catch (HashException | PartitionException e) {
			logger.severe("RIKI: CANNOT FIND DESTINATION FOR GEOHASH: "+ geoHashAntipode);
		}
	
		//String nodeString = targetNode.stringRepresentation();
		
		return targetNode;
		
	}

	/**
	 * Handles the client request on behalf of the node that received the
	 * request
	 * 
	 * @param request
	 *            - This must be a server side event: Generic Event or
	 *            QueryEvent
	 * @param nr_current 
	 * @param hostString 
	 * @param response
	 */
	public void handleRequest() {
		
		try {
			
			/**
			 * STEP 1: FIND TOP K CLIQUES IN THE NODE
			 */
			Map<String, List<RoutingEntry>> routingTable = fs.getRoutingTable();
			
			// MAP OF WHICH CLIQUE GOES TO WHICH NODE
			Map<String, List<CliqueContainer>> nodeToCliquesMap = new HashMap<String, List<CliqueContainer>>();
			
			if(topKCliques.size() <= 0) {
				
				currentNode.setHotspotBeingHandled(false);
				return;
			}
			
			// FOR EACH CLIQUE, FIND A SUITABLE NODE TO SEND IT TO
			// THE NODE HAS TO BE THE ANTIPODE OF THE GEOHASH IN QUESTION
			for(Entry<String, CliqueContainer> entry : topKCliques.entrySet()) {
				
				boolean keepLooping = true;
				
				// THE TARGET HELPER NODE NODE
				String nodeString = null;
				
				while(keepLooping) {
					
					// EAST OR WEST
					int randDirection = ThreadLocalRandom.current().nextInt(0,8);
					
					String geohashKey = entry.getKey();
					CliqueContainer clique = entry.getValue();
					
					// IN CASE OF RETRY, AVOID CLIQUES THAT HAVE ALREADY FOUND A HOME
					if(clique.getReplicatedNode() != null)
						continue;
					
					String geoHashAntipode = GeoHash.getAntipodeGeohash(geohashKey);
					
					NodeInfo targetNode = null;
					
					// IF THIS IS A RETRY TO ANOTHER NODE
					if(!firstTry) {
						
						// GET THE GEOHASH ANTIPODE
						// GET ITS NEIGHBOR
						// LOOP TILL THE NODE IS DIFFERENT FROM THE OLD ANTIPODE
						
						randDirection = clique.getDirection();
						
						geoHashAntipode = clique.getGeohashAntipode();
						
						String newGeohashAntipode = geoHashAntipode;
						
						NodeInfo tempTargetNode = getNodeForGeoHash(geoHashAntipode);
						
						String oldNode = tempTargetNode.stringRepresentation();
						
						String newNode = oldNode;
						
						// LOOKING FOR THE NEXT ALTERNATE ANTIPODE
						while(newNode.equals(oldNode)) {
							// KEEP SHIFTING TILL A NEW CANDIDATE NODE HAS BEEN FOUND
							newGeohashAntipode = GeoHash.getNeighbours(newGeohashAntipode)[randDirection];
							
							tempTargetNode = getNodeForGeoHash(newGeohashAntipode);
							
							// IN CASE WE HIT THE EDGE OF THE WORLD, RESET DIRECTION
							if(tempTargetNode == null) {
								
								// RESETTING THE SEARCH FOR ANTIPODE
								newGeohashAntipode = geoHashAntipode;
								randDirection = ThreadLocalRandom.current().nextInt(0,8);
								
								clique.setDirection(randDirection);
								newNode = oldNode;
								continue;
							}
							
							newNode = tempTargetNode.stringRepresentation();
						}
						
						targetNode = tempTargetNode;
						geoHashAntipode = newGeohashAntipode;
						
					} else {
						
						// TILL A SUITABLE NODE HAS BEEN FOUND
						
						logger.info("RIKI: GEOHASH ANTIPODE: "+geoHashAntipode);
						// FINDING THE NODE THAT HOUSES THE ANTIPODE GEOHASH
						SpatialRange spatialRange = GeoHash.decodeHash(geoHashAntipode);
						
						SpatialProperties spatialProperties = new SpatialProperties(spatialRange);
						Metadata metadata = new Metadata();
						metadata.setName(geoHashAntipode);
						
						metadata.setSpatialProperties(spatialProperties);
						
						try {
							targetNode = partitioner.locateData(metadata);
						} catch (HashException | PartitionException e) {
							logger.severe("RIKI: CANNOT FIND ANTIPODE DESTINATION");
						}
						clique.setDirection(randDirection);
					}
					
					nodeString = targetNode.stringRepresentation();
					
					// KEEPS TRACK OF WHICH ANTIPODE IS CURRENTLY BEING DEALT WITH
					clique.setGeohashAntipode(geoHashAntipode);
					
					
					logger.info("RIKI: CLIQUE: "+geohashKey +" HAS AN ANTIPODE "+geoHashAntipode+" AND CORRESPONDING HELPER NODE IS "+ nodeString);
					
					// IF THIS NODE IS ALREADY GOT A COPY OF THIS GEOHASH
					keepLooping = checkIfEntryExistsInRoutingTable(routingTable, nodeString, geohashKey);
					
					if(keepLooping) {
						
						logger.info("RIKI: WE NEED TO KEEP LOOKING FOR A NEW HELPER NODE FOR CLIQUE: "+geohashKey);
						continue;
						
					} else {
						if(nodeToCliquesMap.get(nodeString) == null) {
							
							List<CliqueContainer> cliques = new ArrayList<CliqueContainer>();
							
							cliques.add(clique);
							nodeToCliquesMap.put(nodeString, cliques);
						} else {
							
							List<CliqueContainer> cliques = nodeToCliquesMap.get(nodeString);
							cliques.add(clique);
						}
						break;
					}
				}
				
				
			}
			
			
			expectedResponses = new AtomicInteger(nodeToCliquesMap.size());
			
			
			firstTry = false;
			
			
			// USE nodeToCliquesMap TO DIRECT CLIQUES TO RESPECTIVE NODES
			// IF THE NODES DONT SEND BACK POSITIVE ACK, TAKE THE REMAINING CLIQUES AND PERFORM THIS SAME SEQUENCE AGAIN
			
			
			
			for(Entry<String, List<CliqueContainer>> entry : nodeToCliquesMap.entrySet()) {
				
				long eventTime = System.currentTimeMillis();
				String nodeKey = entry.getKey();
				List<CliqueContainer> cliquesToSend = entry.getValue();
				
				NetworkDestination nodeToSendTo = nodeStringToNodeMap.get(nodeKey);
				
				HeartbeatRequest hr = new HeartbeatRequest(cliquesToSend, currentNode.getNodeString(), eventTime);
				
				GalileoMessage mrequest = this.eventWrapper.wrap(hr);
				
				this.router.sendMessage(nodeToSendTo, mrequest);
				logger.info("RIKI: DISTRESS REQUEST SENT TO " + nodeToSendTo.toString());
				
				
				
			}
			
			// CHECKS IF THERE IS A TIMEOUT IN RESPONSE COMING BACK FROM THE HELPER NODES
			TimeoutChecker tc = new TimeoutChecker(this, WAIT_TIME);
			Thread internalThread = new Thread(tc);
			
			internalThread.start();
			
		} catch (IOException e) {
			logger.log(Level.INFO,
					"Failed to send request to other nodes in the network. Details follow: " + e.getMessage());
		}
		
	}

	
	/**
	 * CHECKS IF A PARTICULAR NODE ALREADY HOUSES A PARTICULAR CLIQUE ALREADY
	 * IF SO, FIND ANOTHER NODE TO REPLICATE IN....TO MAKE MULTIPLE COPIES OF THE REPLICA
	 * 
	 * @author sapmitra
	 * @param routingTable
	 * @param nodeString
	 * @param geohashKey
	 * @return
	 */
	private boolean checkIfEntryExistsInRoutingTable(Map<String, List<RoutingEntry>> routingTable, String nodeString,
			String geohashKey) {
		
		
		if(routingTable.containsKey(geohashKey)) {
			
			List<RoutingEntry> relist = routingTable.get(geohashKey);
			if(relist != null && relist.size()>0) {
				for(RoutingEntry re : relist) {
					
					if(re.getHelperNode().stringRepresentation().equals(nodeString)) {
						return true;
					}
				}
			}
			
		}
		return false;
	}


	public void silentClose() {
		try {
			this.router.forceShutdown();
		} catch (Exception e) {
			logger.log(Level.INFO, "Failed to shutdown the completed client request handler: ", e);
		}
	}

	@Override
	public void onConnect(NetworkDestination endpoint) {

	}

	@Override
	public void onDisconnect(NetworkDestination endpoint) {

	}

	public boolean isWaitTimeCheck() {
		return waitTimeCheck;
	}

	public void setWaitTimeCheck(boolean waitTimeCheck) {
		this.waitTimeCheck = waitTimeCheck;
	}
	
	

}
