/*******************************************************************************
 * Software in the Sustain Ecosystem are Released Under Terms of Apache Software License 
 * 
 * This research has been supported by funding from the US National Science Foundation's CSSI program through awards 1931363, 1931324, 1931335, and 1931283. The project is a joint effort involving Colorado State University, Arizona State University, the University of California-Irvine, and the University of Maryland - Baltimore County. All redistributions of the software must also include this information. 
 * 
 * TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 * 
 * 
 * 1. Definitions.
 * 
 * "License" shall mean the terms and conditions for use, reproduction, and distribution as defined by Sections 1 through 9 of this document.
 * 
 * "Licensor" shall mean the copyright owner or entity authorized by the copyright owner that is granting the License.
 * 
 * "Legal Entity" shall mean the union of the acting entity and all other entities that control, are controlled by, or are under common control with that entity. For the purposes of this definition, "control" means (i) the power, direct or indirect, to cause the direction or management of such entity, whether by contract or otherwise, or (ii) ownership of fifty percent (50%) or more of the outstanding shares, or (iii) beneficial ownership of such entity.
 * 
 * "You" (or "Your") shall mean an individual or Legal Entity exercising permissions granted by this License.
 * 
 * "Source" form shall mean the preferred form for making modifications, including but not limited to software source code, documentation source, and configuration files.
 * 
 * "Object" form shall mean any form resulting from mechanical transformation or translation of a Source form, including but not limited to compiled object code, generated documentation, and conversions to other media types.
 * 
 * "Work" shall mean the work of authorship, whether in Source or Object form, made available under the License, as indicated by a copyright notice that is included in or attached to the work (an example is provided in the Appendix below).
 * 
 * "Derivative Works" shall mean any work, whether in Source or Object form, that is based on (or derived from) the Work and for which the editorial revisions, annotations, elaborations, or other modifications represent, as a whole, an original work of authorship. For the purposes of this License, Derivative Works shall not include works that remain separable from, or merely link (or bind by name) to the interfaces of, the Work and Derivative Works thereof.
 * 
 * "Contribution" shall mean any work of authorship, including the original version of the Work and any modifications or additions to that Work or Derivative Works thereof, that is intentionally submitted to Licensor for inclusion in the Work by the copyright owner or by an individual or Legal Entity authorized to submit on behalf of the copyright owner. For the purposes of this definition, "submitted" means any form of electronic, verbal, or written communication sent to the Licensor or its representatives, including but not limited to communication on electronic mailing lists, source code control systems, and issue tracking systems that are managed by, or on behalf of, the Licensor for the purpose of discussing and improving the Work, but excluding communication that is conspicuously marked or otherwise designated in writing by the copyright owner as "Not a Contribution."
 * 
 * "Contributor" shall mean Licensor and any individual or Legal Entity on behalf of whom a Contribution has been received by Licensor and subsequently incorporated within the Work.
 * 
 * 2. Grant of Copyright License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable copyright license to reproduce, prepare Derivative Works of, publicly display, publicly perform, sublicense, and distribute the Work and such Derivative Works in Source or Object form.
 * 
 * 3. Grant of Patent License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable (except as stated in this section) patent license to make, have made, use, offer to sell, sell, import, and otherwise transfer the Work, where such license applies only to those patent claims licensable by such Contributor that are necessarily infringed by their Contribution(s) alone or by combination of their Contribution(s) with the Work to which such Contribution(s) was submitted. If You institute patent litigation against any entity (including a cross-claim or counterclaim in a lawsuit) alleging that the Work or a Contribution incorporated within the Work constitutes direct or contributory patent infringement, then any patent licenses granted to You under this License for that Work shall terminate as of the date such litigation is filed.
 * 
 * 4. Redistribution. You may reproduce and distribute copies of the Work or Derivative Works thereof in any medium, with or without modifications, and in Source or Object form, provided that You meet the following conditions:
 * 
 * You must give any other recipients of the Work or Derivative Works a copy of this License; and
 * You must cause any modified files to carry prominent notices stating that You changed the files; and
 * You must retain, in the Source form of any Derivative Works that You distribute, all copyright, patent, trademark, and attribution notices from the Source form of the Work, excluding those notices that do not pertain to any part of the Derivative Works; and
 * If the Work includes a "NOTICE" text file as part of its distribution, then any Derivative Works that You distribute must include a readable copy of the attribution notices contained within such NOTICE file, excluding those notices that do not pertain to any part of the Derivative Works, in at least one of the following places: within a NOTICE text file distributed as part of the Derivative Works; within the Source form or documentation, if provided along with the Derivative Works; or, within a display generated by the Derivative Works, if and wherever such third-party notices normally appear. The contents of the NOTICE file are for informational purposes only and do not modify the License. You may add Your own attribution notices within Derivative Works that You distribute, alongside or as an addendum to the NOTICE text from the Work, provided that such additional attribution notices cannot be construed as modifying the License. 
 * 
 * You may add Your own copyright statement to Your modifications and may provide additional or different license terms and conditions for use, reproduction, or distribution of Your modifications, or for any such Derivative Works as a whole, provided Your use, reproduction, and distribution of the Work otherwise complies with the conditions stated in this License.
 * 5. Submission of Contributions. Unless You explicitly state otherwise, any Contribution intentionally submitted for inclusion in the Work by You to the Licensor shall be under the terms and conditions of this License, without any additional terms or conditions. Notwithstanding the above, nothing herein shall supersede or modify the terms of any separate license agreement you may have executed with Licensor regarding such Contributions.
 * 
 * 6. Trademarks. This License does not grant permission to use the trade names, trademarks, service marks, or product names of the Licensor, except as required for reasonable and customary use in describing the origin of the Work and reproducing the content of the NOTICE file.
 * 
 * 7. Disclaimer of Warranty. Unless required by applicable law or agreed to in writing, Licensor provides the Work (and each Contributor provides its Contributions) on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied, including, without limitation, any warranties or conditions of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A PARTICULAR PURPOSE. You are solely responsible for determining the appropriateness of using or redistributing the Work and assume any risks associated with Your exercise of permissions under this License.
 * 
 * 8. Limitation of Liability. In no event and under no legal theory, whether in tort (including negligence), contract, or otherwise, unless required by applicable law (such as deliberate and grossly negligent acts) or agreed to in writing, shall any Contributor be liable to You for damages, including any direct, indirect, special, incidental, or consequential damages of any character arising as a result of this License or out of the use or inability to use the Work (including but not limited to damages for loss of goodwill, work stoppage, computer failure or malfunction, or any and all other commercial damages or losses), even if such Contributor has been advised of the possibility of such damages.
 * 
 * 9. Accepting Warranty or Additional Liability. While redistributing the Work or Derivative Works thereof, You may choose to offer, and charge a fee for, acceptance of support, warranty, indemnity, or other liability obligations and/or rights consistent with this License. However, in accepting such obligations, You may act only on Your own behalf and on Your sole responsibility, not on behalf of any other Contributor, and only if You agree to indemnify, defend, and hold each Contributor harmless for any liability incurred by, or claims asserted against, such Contributor by reason of your accepting any such warranty or additional liability. 
 * 
 * END OF TERMS AND CONDITIONS
 ******************************************************************************/


package galileo.dht.hash;

import java.math.BigInteger;
import java.util.Set;
import java.util.TreeMap;

/**
 * Creates an evenly-spaced hash ring network topology.  Note that nodes in the
 * topology are spaced evenly apart, but of course this does not guarantee the
 * data will hash uniformly.
 *
 * @author malensek
 */
public class BalancedHashRing<T> implements HashRing<T> {

    /** If set to True, the first position in the hash ring is randomized. */
    protected boolean randomize = false;

    protected HashFunction<T> function;
    protected BigInteger maxHash;

    /** Maps hash ring positions to ring entries */
    protected TreeMap<BigInteger, HashRingEntry> entryMap = new TreeMap<>();

    /**
     * Creates a BalancedHashRing using the provided hash function.  The
     * function will determine the size of the hash space and where nodes will
     * be placed in the topology.
     *
     * @param function HashFunction that defines the hash space being used.
     */
    public BalancedHashRing(HashFunction<T> function) {
        this(function, false);
    }

    /**
     * Creates a BalancedHashRing using the provided hash function.  The
     * function will determine the size of the hash space and where nodes will
     * be placed in the topology.
     *
     * @param function HashFunction that defines the hash space being used.
     */
    public BalancedHashRing(HashFunction<T> function, boolean randomize) {
        this.function = function;
        this.randomize = randomize;
        maxHash = function.maxValue();
    }

    /**
     * Insert a node into the hash ring internal data structures.  Nodes are
     * relinked with the proper neighbors.
     *
     * @param position place in the hash space
     * @param predecessor the predecessor node in the hash space.
     */
    private void addRingEntry(BigInteger position, HashRingEntry predecessor)
    throws HashTopologyException {
        if (entryMap.get(position) != null) {
            /* Something is already here! */
            System.out.println(position);
            throw new HashTopologyException("Hash space exhausted!");
        }

        HashRingEntry newEntry
            = new HashRingEntry(position, predecessor.neighbor);
        predecessor.neighbor = newEntry;
        entryMap.put(position, newEntry);
    }

    /**
     * Add a node to the overlay network topology.
     *
     * @param data unused for this hash ring; nodes are placed evenly based on
     * current topology characteristics.  You may safely pass 'null' to this
     * method.
     *
     * @return the location of the new node in the hash space.
     */
    @Override
    public BigInteger addNode(T data)
    throws HashTopologyException, HashException {
        /* Edge case: when there are no entries in the hash ring yet. */
        if (entryMap.values().size() == 0) {
            BigInteger pos;

            if (randomize) {
                /* Find a random location to start with */
                pos = function.randomHash();
            } else {
                pos = BigInteger.ZERO;
            }

            HashRingEntry firstEntry = new HashRingEntry(pos);
            entryMap.put(pos, firstEntry);

            return pos;
        }

        /* Edge case: only one entry in the hash ring */
        if (entryMap.values().size() == 1) {
            HashRingEntry firstEntry = entryMap.values().iterator().next();
            BigInteger halfSize = maxHash.divide(BigInteger.valueOf(2));
            BigInteger secondPos = firstEntry.position.add(halfSize);

            if (secondPos.compareTo(maxHash) > 0) {
                secondPos = secondPos.subtract(maxHash);
            }

            HashRingEntry secondEntry
                = new HashRingEntry(secondPos, firstEntry);
            firstEntry.neighbor = secondEntry;
            entryMap.put(secondPos, secondEntry);

            return secondPos;
        }

        /* Find the largest empty span of hash space */
        BigInteger largestSpan = BigInteger.ZERO;
        HashRingEntry largestEntry = null;
        
        for (HashRingEntry entry : entryMap.values()) {
            BigInteger len = lengthBetween(entry, entry.neighbor);
            if (len.compareTo(largestSpan) > 0) {
                largestSpan = len;
                largestEntry = entry;
            }
        }

        if (largestEntry == null) {
            return BigInteger.ONE.negate();
        }

        /* Put the new node in the middle of the largest span */
        BigInteger half = half(largestEntry, largestEntry.neighbor);
        addRingEntry(half, largestEntry);
        return half;
    }

    /**
     * Find the hash location in the middle of a hash span.  When the hash space
     * is represented as a continuous circle, a span begins at an arbitrary
     * starting point and then proceeds clockwise until it reaches its endpoint.
     *
     * @param start beginning of the hash span
     * @param end end of the hash span
     *
     * @return hash position in the middle of the span.
     */
    private BigInteger half(HashRingEntry start, HashRingEntry end) {
        BigInteger length = lengthBetween(start, end);
        /* half = start + length / 2 */
        BigInteger half
            = start.position.add(length.divide(BigInteger.valueOf(2)));

        if (maxHash.compareTo(half) >= 0) {
            return half;
        } else {
            return half.subtract(maxHash);
        }
    }

    /**
     * Determines the number of hash values between two positions on a hash
     * ring.  The hash space is viewed as a ring, starting with 0 and proceeding
     * clockwise until it reaches its maximum value.  A hash span starts at an
     * arbitrary position, and then proceeds clockwise until it reaches its
     * ending position.  The number of values between the two points is the hash
     * span length.
     *
     * @param start the start of the hash span being measured
     * @param end the end of the hash span being measured
     *
     * @return the length of the hash span
     */
    private BigInteger lengthBetween(HashRingEntry start, HashRingEntry end) {
        BigInteger difference = end.position.subtract(start.position);

        if (difference.compareTo(BigInteger.ZERO) >= 0) {
            return difference;
        } else {
            /* Wraparound */
            /* wrapped = (MAX_HASH - start) + end */
            BigInteger wrapped
                = maxHash.subtract(start.position).add(end.position);
            return wrapped;
        }
    }

    @Override
    public BigInteger locate(T data) throws HashException {
        BigInteger hashLocation = function.hash(data);
        BigInteger node = entryMap.ceilingKey(hashLocation);

        /* Wraparound edge case */
        if (node == null) {
            node = entryMap.ceilingKey(BigInteger.ZERO);
        }

        return node;
    }
    
    @Override
    public Set<BigInteger> getPositions() {
    	return entryMap.keySet();
    }

    /**
     * Formats the hash ring as node-to-predecessor pair Strings.
     *
     * @return network links and nodes in String format.
     */
    @Override
    public String toString() {
        String str = "";
        HashRingEntry firstEntry = entryMap.values().iterator().next();
        HashRingEntry currentEntry = firstEntry;
        HashRingEntry nextEntry;

        do {
            nextEntry = currentEntry.neighbor;
            str += currentEntry.position + " -> " + nextEntry.position;
            str += System.lineSeparator();
            currentEntry = nextEntry;
        } while (currentEntry != firstEntry);

        return str;
    }
}
