/*******************************************************************************
 * Software in the Sustain Ecosystem are Released Under Terms of Apache Software License 
 * 
 * This research has been supported by funding from the US National Science Foundation's CSSI program through awards 1931363, 1931324, 1931335, and 1931283. The project is a joint effort involving Colorado State University, Arizona State University, the University of California-Irvine, and the University of Maryland - Baltimore County. All redistributions of the software must also include this information. 
 * 
 * TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 * 
 * 
 * 1. Definitions.
 * 
 * "License" shall mean the terms and conditions for use, reproduction, and distribution as defined by Sections 1 through 9 of this document.
 * 
 * "Licensor" shall mean the copyright owner or entity authorized by the copyright owner that is granting the License.
 * 
 * "Legal Entity" shall mean the union of the acting entity and all other entities that control, are controlled by, or are under common control with that entity. For the purposes of this definition, "control" means (i) the power, direct or indirect, to cause the direction or management of such entity, whether by contract or otherwise, or (ii) ownership of fifty percent (50%) or more of the outstanding shares, or (iii) beneficial ownership of such entity.
 * 
 * "You" (or "Your") shall mean an individual or Legal Entity exercising permissions granted by this License.
 * 
 * "Source" form shall mean the preferred form for making modifications, including but not limited to software source code, documentation source, and configuration files.
 * 
 * "Object" form shall mean any form resulting from mechanical transformation or translation of a Source form, including but not limited to compiled object code, generated documentation, and conversions to other media types.
 * 
 * "Work" shall mean the work of authorship, whether in Source or Object form, made available under the License, as indicated by a copyright notice that is included in or attached to the work (an example is provided in the Appendix below).
 * 
 * "Derivative Works" shall mean any work, whether in Source or Object form, that is based on (or derived from) the Work and for which the editorial revisions, annotations, elaborations, or other modifications represent, as a whole, an original work of authorship. For the purposes of this License, Derivative Works shall not include works that remain separable from, or merely link (or bind by name) to the interfaces of, the Work and Derivative Works thereof.
 * 
 * "Contribution" shall mean any work of authorship, including the original version of the Work and any modifications or additions to that Work or Derivative Works thereof, that is intentionally submitted to Licensor for inclusion in the Work by the copyright owner or by an individual or Legal Entity authorized to submit on behalf of the copyright owner. For the purposes of this definition, "submitted" means any form of electronic, verbal, or written communication sent to the Licensor or its representatives, including but not limited to communication on electronic mailing lists, source code control systems, and issue tracking systems that are managed by, or on behalf of, the Licensor for the purpose of discussing and improving the Work, but excluding communication that is conspicuously marked or otherwise designated in writing by the copyright owner as "Not a Contribution."
 * 
 * "Contributor" shall mean Licensor and any individual or Legal Entity on behalf of whom a Contribution has been received by Licensor and subsequently incorporated within the Work.
 * 
 * 2. Grant of Copyright License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable copyright license to reproduce, prepare Derivative Works of, publicly display, publicly perform, sublicense, and distribute the Work and such Derivative Works in Source or Object form.
 * 
 * 3. Grant of Patent License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable (except as stated in this section) patent license to make, have made, use, offer to sell, sell, import, and otherwise transfer the Work, where such license applies only to those patent claims licensable by such Contributor that are necessarily infringed by their Contribution(s) alone or by combination of their Contribution(s) with the Work to which such Contribution(s) was submitted. If You institute patent litigation against any entity (including a cross-claim or counterclaim in a lawsuit) alleging that the Work or a Contribution incorporated within the Work constitutes direct or contributory patent infringement, then any patent licenses granted to You under this License for that Work shall terminate as of the date such litigation is filed.
 * 
 * 4. Redistribution. You may reproduce and distribute copies of the Work or Derivative Works thereof in any medium, with or without modifications, and in Source or Object form, provided that You meet the following conditions:
 * 
 * You must give any other recipients of the Work or Derivative Works a copy of this License; and
 * You must cause any modified files to carry prominent notices stating that You changed the files; and
 * You must retain, in the Source form of any Derivative Works that You distribute, all copyright, patent, trademark, and attribution notices from the Source form of the Work, excluding those notices that do not pertain to any part of the Derivative Works; and
 * If the Work includes a "NOTICE" text file as part of its distribution, then any Derivative Works that You distribute must include a readable copy of the attribution notices contained within such NOTICE file, excluding those notices that do not pertain to any part of the Derivative Works, in at least one of the following places: within a NOTICE text file distributed as part of the Derivative Works; within the Source form or documentation, if provided along with the Derivative Works; or, within a display generated by the Derivative Works, if and wherever such third-party notices normally appear. The contents of the NOTICE file are for informational purposes only and do not modify the License. You may add Your own attribution notices within Derivative Works that You distribute, alongside or as an addendum to the NOTICE text from the Work, provided that such additional attribution notices cannot be construed as modifying the License. 
 * 
 * You may add Your own copyright statement to Your modifications and may provide additional or different license terms and conditions for use, reproduction, or distribution of Your modifications, or for any such Derivative Works as a whole, provided Your use, reproduction, and distribution of the Work otherwise complies with the conditions stated in this License.
 * 5. Submission of Contributions. Unless You explicitly state otherwise, any Contribution intentionally submitted for inclusion in the Work by You to the Licensor shall be under the terms and conditions of this License, without any additional terms or conditions. Notwithstanding the above, nothing herein shall supersede or modify the terms of any separate license agreement you may have executed with Licensor regarding such Contributions.
 * 
 * 6. Trademarks. This License does not grant permission to use the trade names, trademarks, service marks, or product names of the Licensor, except as required for reasonable and customary use in describing the origin of the Work and reproducing the content of the NOTICE file.
 * 
 * 7. Disclaimer of Warranty. Unless required by applicable law or agreed to in writing, Licensor provides the Work (and each Contributor provides its Contributions) on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied, including, without limitation, any warranties or conditions of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A PARTICULAR PURPOSE. You are solely responsible for determining the appropriateness of using or redistributing the Work and assume any risks associated with Your exercise of permissions under this License.
 * 
 * 8. Limitation of Liability. In no event and under no legal theory, whether in tort (including negligence), contract, or otherwise, unless required by applicable law (such as deliberate and grossly negligent acts) or agreed to in writing, shall any Contributor be liable to You for damages, including any direct, indirect, special, incidental, or consequential damages of any character arising as a result of this License or out of the use or inability to use the Work (including but not limited to damages for loss of goodwill, work stoppage, computer failure or malfunction, or any and all other commercial damages or losses), even if such Contributor has been advised of the possibility of such damages.
 * 
 * 9. Accepting Warranty or Additional Liability. While redistributing the Work or Derivative Works thereof, You may choose to offer, and charge a fee for, acceptance of support, warranty, indemnity, or other liability obligations and/or rights consistent with this License. However, in accepting such obligations, You may act only on Your own behalf and on Your sole responsibility, not on behalf of any other Contributor, and only if You agree to indemnify, defend, and hold each Contributor harmless for any liability incurred by, or claims asserted against, such Contributor by reason of your accepting any such warranty or additional liability. 
 * 
 * END OF TERMS AND CONDITIONS
 ******************************************************************************/
package galileo.graph;

import java.io.IOException;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.List;

import galileo.comm.TemporalType;
import galileo.dataset.Coordinates;
import galileo.fs.GeospatialFileSystem;
import galileo.serialization.ByteSerializable;
import galileo.serialization.SerializationException;
import galileo.serialization.SerializationInputStream;
import galileo.serialization.SerializationOutputStream;
import galileo.serialization.ByteSerializable.Deserialize;
import galileo.util.GeoHash;
import galileo.util.STRelatives;
import galileo.util.SpatialBorder;

public class CacheCell implements ByteSerializable{
	
	private SpatiotemporalHierarchicalCache cache;
	private String spatialInfo;
	private String temporalInfo;
	
	// MAINTAINED AS NON-ZERO NUMBERS
	private int spatialResolution;
	private int temporalResolution;
	
	private SummaryStatistics[] stats;
	
	private String spatialParent;
	private String temporalParent;
	
	private List<String> spatialChildren;
	private List<String> temporalChildren;
	
	private List<String> spatialNeighbors;
	private List<String> temporalNeighbors;

	private boolean hasParent;
	private boolean hasChildren;
	
	
	private float freshness;
	private long lastAccessed;
	private String lastEvent;
	
	private String cellKey;
	
	
	public String getCellKey() {
		return cellKey;
	}
	
	/**
	 * CREATING A NEW CACHE CELL
	 * UPDATE THE NEIGHBOR INFO AND OBJECT REFERENCES
	 * @param cache 
	 * @param stats
	 * @param numChildren
	 * @param numNeighbors
	 * @param numParents
	 * @param spatiotemporalInfo The spatial and temporal strings that define this cell block
	 * @param spatialResolution
	 * @param temporalResolution
	 * @param eventTime 
	 * @param eventId 
	 * @param eventId 
	 * @param freshnessMultiplier 
	 */
	
	public CacheCell(SpatiotemporalHierarchicalCache cache, SummaryStatistics[] stats, String spatiotemporalInfo,
			int spatialResolution, int temporalResolution, String eventId, long eventTime, int freshnessMultiplier) {
		
		this.cache = cache;
		this.stats = stats;
		this.spatialResolution = spatialResolution;
		this.temporalResolution = temporalResolution;
		
		this.cellKey = spatiotemporalInfo;
		String[] components = spatiotemporalInfo.split("\\$\\$");
		
		this.temporalInfo = components[0];
		this.spatialInfo = components[1];
		
		/* NEIGHBORS FOR THIS LEVEL */
		spatialNeighbors = GeoHash.getSpatialNeighboursSimplified(spatialInfo, GeospatialFileSystem.SPATIAL_SPREAD);
		temporalNeighbors = GeoHash.getTemporalNeighbors(temporalInfo, temporalResolution);
		
		/* PARENTS ONE LEVEL UP */
		if(spatialResolution > 1)
			spatialParent = GeoHash.getSpatialParent(spatialInfo, 1);
		else 
			spatialParent = null;
		
		if(temporalResolution > 1)
			temporalParent = GeoHash.getTemporalParent(temporalInfo, temporalResolution);
		else
			temporalParent = null;
		
		hasParent = true;
		
		if(spatialParent == null && temporalParent == null)
			hasParent = false;
		
		/* CHILDREN ONE LEVEL DOWN */
		if(spatialResolution < SpatiotemporalHierarchicalCache.totalSpatialLevels )
			spatialChildren = GeoHash.getSpatialChildren(spatialInfo);
		else 
			spatialChildren = null;
		
		if(temporalResolution < SpatiotemporalHierarchicalCache.totalTemporalLevels)
			temporalChildren = GeoHash.getTemporalChildren(temporalInfo, temporalResolution);
		else
			temporalChildren = null;
		
		hasChildren = true;
		
		if(spatialChildren == null && temporalChildren == null)
			hasChildren = false;
		
		this.freshness = 1*freshnessMultiplier;
		
		this.lastEvent = eventId;
		this.lastAccessed = eventTime;
		
	}
	
	
	public String toString() {
		
		return cellKey;
	}

	public String getSpatialInfo() {
		return spatialInfo;
	}


	public void setSpatialInfo(String spatialInfo) {
		this.spatialInfo = spatialInfo;
	}


	public String getTemporalInfo() {
		return temporalInfo;
	}

	public void setTemporalInfo(String temporalInfo) {
		this.temporalInfo = temporalInfo;
	}
	
	public int getSpatialResolution() {
		return spatialResolution;
	}

	public void setSpatialResolution(int spatialResolution) {
		this.spatialResolution = spatialResolution;
	}

	public int getTemporalResolution() {
		return temporalResolution;
	}

	public void setTemporalResolution(int temporalResolution) {
		this.temporalResolution = temporalResolution;
	}

	public SummaryStatistics[] getStats() {
		return stats;
	}

	public void setStats(SummaryStatistics[] stats) {
		this.stats = stats;
	}

	public float getFreshness() {
		return freshness;
	}
	
	/**
	 * SEE IF THIS IS STALE FOR THE GUEST CACHE
	 * @author sapmitra
	 * @param currentTime
	 * @param hELPER_TIME 
	 * @return
	 */
	public boolean checkForFreshness(long currentTime, long HELPER_TIME) {
		
		boolean toRemove = false;
		long timeInt = currentTime - lastAccessed;
		
		if (timeInt >= HELPER_TIME)
			toRemove = true;
		
		return toRemove;
	}
	
	/**
	 * ADDED TIME DECAY
	 * @author sapmitra
	 * @param currentTime 
	 * @return
	 */
	public float getCorrectedFreshness(long currentTime) {
		
		double timeInt = currentTime - lastAccessed;
		
		double multiplier = java.lang.Math.exp(-1d * timeInt/1000);
		
		float updatedFrashness = (float)(freshness*multiplier);
		
		return updatedFrashness;
	}

	public void setFreshness(float freshness) {
		this.freshness = freshness;
	}

	public String getSpatialParent() {
		return spatialParent;
	}

	
	public void setSpatialParent(String spatialParent) {
		this.spatialParent = spatialParent;
	}

	public String getTemporalParent() {
		return temporalParent;
	}

	public void setTemporalParent(String temporalParent) {
		this.temporalParent = temporalParent;
	}

	public boolean isHasParent() {
		return hasParent;
	}

	public void setHasParent(boolean hasParent) {
		this.hasParent = hasParent;
	}

	public boolean isHasChildren() {
		return hasChildren;
	}

	public void setHasChildren(boolean hasChildren) {
		this.hasChildren = hasChildren;
	}

	
	
	public void incrementFreshness(float val, String eventId, long eventTime) {
		this.freshness += val;
		this.lastAccessed = eventTime;
		this.lastEvent = eventId;
	}
	
	
	public static void main(String arg[]) {
		Calendar cal = Calendar.getInstance(); // creates calendar
	    cal.setTime(new Date()); // sets calendar time/date
	    cal.add(Calendar.HOUR_OF_DAY, 280); // adds one hour
	    System.out.println(cal.getTime()); 
		
	}

	public List<String> getSpatialNeighbors() {
		return spatialNeighbors;
	}

	public void setSpatialNeighbors(List<String> spatialNeighbors) {
		this.spatialNeighbors = spatialNeighbors;
	}

	/**
	 * ONLY THE NEIGHBORS THAT NEED TO BE DISPERSED FRESHNESS
	 * @author sapmitra
	 * @param polygon
	 * @return
	 */
	public List<String> getRefinedSpatialNeighbors(List<Coordinates> polygon) {
		
		List<String> refinedNeighbors = new ArrayList<String>();
		
		for(String gh: spatialNeighbors) {
			boolean hasIntersection = GeoHash.checkIntersection(polygon, gh);
			
			if(!hasIntersection) {
				refinedNeighbors.add(gh);
			}
		}
		
		return refinedNeighbors;
	}
	
	public List<String> getTemporalNeighbors() {
		return temporalNeighbors;
	}
	
	
	public List<String> getRefinedTemporalNeighbors(long qt1, long qt2) {
		
		List<String> refinedTemporalNeighbors = new ArrayList<String>();
		
		for(String ts : temporalNeighbors) {
			try {
				boolean temporalIntersection = GeoHash.checkTemporalIntersection(ts, qt1, qt2);
				if(!temporalIntersection)
					refinedTemporalNeighbors.add(ts);
				
			} catch (ParseException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			
			
		}
		return refinedTemporalNeighbors;
	}
	
	/**
	 * DISPERSING FRESHNESS VALUE TO A CELL'S RELATIVES
	 * 
	 * @author sapmitra
	 * @param polygon
	 * @param qt1
	 * @param qt2
	 * @param eventTime 
	 * @param eventId 
	 * @param freshnessMultiplier 
	 * @return
	 */
	public void freshenUpRelativesForCell(List<Coordinates> polygon, long qt1, long qt2, String eventId, long eventTime, int freshnessMultiplier) {
		
		freshenUpParents(eventId, eventTime, freshnessMultiplier);
		freshenUpChildren(eventId, eventTime, freshnessMultiplier);
		freshenUpNeighbors(polygon, qt1, qt2, eventId, eventTime, freshnessMultiplier);
		
		
	}
	

	public int getCellLevel() {
		
		return cache.getCacheLevel(spatialResolution, temporalResolution);
	}

	/**
	 * DISPERSE FRESHNESS AMONG NEIGHBORS
	 * @author sapmitra
	 * @param polygon
	 * @param qt1
	 * @param qt2
	 * @param eventTime 
	 * @param eventId 
	 */
	
	private void freshenUpNeighbors(List<Coordinates> polygon, long qt1, long qt2, String eventId, long eventTime, int freshnessMultiplier) {
		
		int cacheLevel = cache.getCacheLevel(spatialResolution, temporalResolution);
		// Only at the level of the current cell
		List<String> refinedSpatialNeighbors = getRefinedSpatialNeighbors(polygon);
		List<String> refinedTemporalNeighbors = getRefinedTemporalNeighbors(qt1, qt2);
		
		for(String s: refinedSpatialNeighbors) {
			for(String t: refinedTemporalNeighbors) {
				
				cache.disperseToCell(t+"$$"+s, cacheLevel, eventId, eventTime, freshnessMultiplier);
			}
		}
	}
	
	/**
	 * CHILDREN ARE CONSIDERED ONE SPATIAL/TEMPORAL LEVEL ABOVE, NOT BOTH
	 * DISPERSE FRESHNESS TO CHILDREN
	 * @author sapmitra
	 * @param eventId
	 * @param eventTime
	 * @param freshnessMultiplier 
	 */
	private void freshenUpChildren(String eventId, long eventTime, int freshnessMultiplier) {
		
		if(spatialChildren != null) {
			
			int cacheLevel = cache.getCacheLevel(spatialResolution+1, temporalResolution);
			
			for(String sc : spatialChildren) {
				
				cache.disperseToCell(temporalInfo+"$$"+sc, cacheLevel, eventId, eventTime, freshnessMultiplier);
				
			}
			
			
		} 
		
		if(temporalChildren != null) {
			
			int cacheLevel = cache.getCacheLevel(spatialResolution, temporalResolution+1);
			
			for(String tc: temporalChildren) {
				
				cache.disperseToCell(tc+"$$"+spatialInfo, cacheLevel, eventId, eventTime, freshnessMultiplier);
				
			}
			
		}
		
	}

	/**
	 * PARENTS ARE CONSIDERED ONE SPATIAL/TEMPORAL LEVEL ABOVE, NOT BOTH
	 * @author sapmitra
	 * @param eventTime 
	 * @param eventId 
	 * @param freshnessMultiplier 
	 * @return
	 */
	public void freshenUpParents(String eventId, long eventTime, int freshnessMultiplier) {
		
		String sp = getSpatialParent();
		String tp = getTemporalParent();
		
		if(sp != null) {
			
			// - 1 because we are loking for parents
			int cacheLevel = cache.getCacheLevel(spatialResolution-1, temporalResolution);
			
			cache.disperseToCell(temporalInfo+"$$"+sp, cacheLevel, eventId, eventTime, freshnessMultiplier);
			
		} 
		
		if(tp != null) {
			
			int cacheLevel = cache.getCacheLevel(spatialResolution, temporalResolution-1);
			cache.disperseToCell(tp+"$$"+spatialInfo, cacheLevel, eventId, eventTime, freshnessMultiplier);
			
		}
		
		/*if(sp != null && tp != null) {
			
			parentKeys.add(tp+"$$"+sp);
			
		}*/
		
	}
	
	/**
	 * Given a set of cache cells, returns the value, if any of an existing cell in the list
	 * matching a given key
	 * 
	 * @author sapmitra
	 * @param cells
	 * @param key
	 * @param currentTime 
	 * @return
	 */
	public static float getExistingCell(List<CacheCell> cells, String key, long currentTime) {
		
		for(CacheCell c : cells) {
			if(c.getCellKey().equals(key)) {
				return c.getCorrectedFreshness(currentTime);
			}
		}
		
		return 0;
		
	}
	
	
	

	public void setTemporalNeighbors(List<String> temporalNeighbors) {
		this.temporalNeighbors = temporalNeighbors;
	}

	public long getLastAccessed() {
		return lastAccessed;
	}

	public void setLastAccessed(long lastAccessed) {
		this.lastAccessed = lastAccessed;
	}

	public List<String> getSpatialChildren() {
		return spatialChildren;
	}

	public void setSpatialChildren(List<String> spatialChildren) {
		this.spatialChildren = spatialChildren;
	}

	public List<String> getTemporalChildren() {
		return temporalChildren;
	}

	public void setTemporalChildren(List<String> temporalChildren) {
		this.temporalChildren = temporalChildren;
	}

	public String getLastEvent() {
		return lastEvent;
	}

	public void setLastEvent(String lastEvent) {
		this.lastEvent = lastEvent;
	}

	@Override
	public void serialize(SerializationOutputStream out) throws IOException {
		
		out.writeString(cellKey);
		out.writeInt(spatialResolution);
		out.writeInt(temporalResolution);
		
		
		List<SummaryStatistics> summaries = new ArrayList<>();
		
		for(SummaryStatistics stat: stats) {
			summaries.add(stat);
		}
		out.writeSerializableCollection(summaries);
		
	}
	
	@Deserialize
	public CacheCell(SerializationInputStream in) throws IOException, SerializationException {
		
		this.cellKey = in.readString();
		this.spatialResolution = in.readInt();
		this.temporalResolution = in.readInt();
		
		List<SummaryStatistics> summaries = new ArrayList<>();
		in.readSerializableCollection(SummaryStatistics.class, summaries);
		
		stats = summaries.toArray(new SummaryStatistics[summaries.size()]);
		
		String[] components = cellKey.split("\\$\\$");
		
		this.temporalInfo = components[0];
		this.spatialInfo = components[1];
		
		/* NEIGHBORS FOR THIS LEVEL */
		spatialNeighbors = GeoHash.getSpatialNeighboursSimplified(spatialInfo, GeospatialFileSystem.SPATIAL_SPREAD);
		temporalNeighbors = GeoHash.getTemporalNeighbors(temporalInfo, temporalResolution);
		
		/* PARENTS ONE LEVEL UP */
		if(spatialResolution > 1)
			spatialParent = GeoHash.getSpatialParent(spatialInfo, 1);
		else 
			spatialParent = null;
		
		if(temporalResolution > 1)
			temporalParent = GeoHash.getTemporalParent(temporalInfo, temporalResolution);
		else
			temporalParent = null;
		
		hasParent = true;
		
		if(spatialParent == null && temporalParent == null)
			hasParent = false;
		
		/* CHILDREN ONE LEVEL DOWN */
		if(spatialResolution < SpatiotemporalHierarchicalCache.totalSpatialLevels )
			spatialChildren = GeoHash.getSpatialChildren(spatialInfo);
		else 
			spatialChildren = null;
		
		if(temporalResolution < SpatiotemporalHierarchicalCache.totalTemporalLevels)
			temporalChildren = GeoHash.getTemporalChildren(temporalInfo, temporalResolution);
		else
			temporalChildren = null;
		
		hasChildren = true;
		
		if(spatialChildren == null && temporalChildren == null)
			hasChildren = false;
		
		this.freshness = 0;
		
	}
	
	/**
	 * Checks if this cell is contained in the query area
	 * @author sapmitra
	 * @param polygon
	 * @param timeString
	 * @param tresolution
	 * @return
	 * @throws ParseException
	 */
	public boolean checkIntersection (List<Coordinates> polygon, String timeString) throws ParseException {
		
		boolean sInt = GeoHash.checkIntersection(polygon, spatialInfo);
		
		// START AND END TIME OF THE TILE
		long qt1 = GeoHash.getStartTimeStamp(temporalInfo, TemporalType.getTypeFromLevel(temporalResolution));
		long qt2 = GeoHash.getEndTimeStamp(temporalInfo, TemporalType.getTypeFromLevel(temporalResolution));
		
		boolean tInt = GeoHash.checkTemporalIntersection(timeString, qt1, qt2);
		
		if(sInt && tInt)
			return true;
		
		return false;
		
		
		
	}


}
