/*******************************************************************************
 * Software in the Sustain Ecosystem are Released Under Terms of Apache Software License 
 * 
 * This research has been supported by funding from the US National Science Foundation's CSSI program through awards 1931363, 1931324, 1931335, and 1931283. The project is a joint effort involving Colorado State University, Arizona State University, the University of California-Irvine, and the University of Maryland - Baltimore County. All redistributions of the software must also include this information. 
 * 
 * TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 * 
 * 
 * 1. Definitions.
 * 
 * "License" shall mean the terms and conditions for use, reproduction, and distribution as defined by Sections 1 through 9 of this document.
 * 
 * "Licensor" shall mean the copyright owner or entity authorized by the copyright owner that is granting the License.
 * 
 * "Legal Entity" shall mean the union of the acting entity and all other entities that control, are controlled by, or are under common control with that entity. For the purposes of this definition, "control" means (i) the power, direct or indirect, to cause the direction or management of such entity, whether by contract or otherwise, or (ii) ownership of fifty percent (50%) or more of the outstanding shares, or (iii) beneficial ownership of such entity.
 * 
 * "You" (or "Your") shall mean an individual or Legal Entity exercising permissions granted by this License.
 * 
 * "Source" form shall mean the preferred form for making modifications, including but not limited to software source code, documentation source, and configuration files.
 * 
 * "Object" form shall mean any form resulting from mechanical transformation or translation of a Source form, including but not limited to compiled object code, generated documentation, and conversions to other media types.
 * 
 * "Work" shall mean the work of authorship, whether in Source or Object form, made available under the License, as indicated by a copyright notice that is included in or attached to the work (an example is provided in the Appendix below).
 * 
 * "Derivative Works" shall mean any work, whether in Source or Object form, that is based on (or derived from) the Work and for which the editorial revisions, annotations, elaborations, or other modifications represent, as a whole, an original work of authorship. For the purposes of this License, Derivative Works shall not include works that remain separable from, or merely link (or bind by name) to the interfaces of, the Work and Derivative Works thereof.
 * 
 * "Contribution" shall mean any work of authorship, including the original version of the Work and any modifications or additions to that Work or Derivative Works thereof, that is intentionally submitted to Licensor for inclusion in the Work by the copyright owner or by an individual or Legal Entity authorized to submit on behalf of the copyright owner. For the purposes of this definition, "submitted" means any form of electronic, verbal, or written communication sent to the Licensor or its representatives, including but not limited to communication on electronic mailing lists, source code control systems, and issue tracking systems that are managed by, or on behalf of, the Licensor for the purpose of discussing and improving the Work, but excluding communication that is conspicuously marked or otherwise designated in writing by the copyright owner as "Not a Contribution."
 * 
 * "Contributor" shall mean Licensor and any individual or Legal Entity on behalf of whom a Contribution has been received by Licensor and subsequently incorporated within the Work.
 * 
 * 2. Grant of Copyright License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable copyright license to reproduce, prepare Derivative Works of, publicly display, publicly perform, sublicense, and distribute the Work and such Derivative Works in Source or Object form.
 * 
 * 3. Grant of Patent License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable (except as stated in this section) patent license to make, have made, use, offer to sell, sell, import, and otherwise transfer the Work, where such license applies only to those patent claims licensable by such Contributor that are necessarily infringed by their Contribution(s) alone or by combination of their Contribution(s) with the Work to which such Contribution(s) was submitted. If You institute patent litigation against any entity (including a cross-claim or counterclaim in a lawsuit) alleging that the Work or a Contribution incorporated within the Work constitutes direct or contributory patent infringement, then any patent licenses granted to You under this License for that Work shall terminate as of the date such litigation is filed.
 * 
 * 4. Redistribution. You may reproduce and distribute copies of the Work or Derivative Works thereof in any medium, with or without modifications, and in Source or Object form, provided that You meet the following conditions:
 * 
 * You must give any other recipients of the Work or Derivative Works a copy of this License; and
 * You must cause any modified files to carry prominent notices stating that You changed the files; and
 * You must retain, in the Source form of any Derivative Works that You distribute, all copyright, patent, trademark, and attribution notices from the Source form of the Work, excluding those notices that do not pertain to any part of the Derivative Works; and
 * If the Work includes a "NOTICE" text file as part of its distribution, then any Derivative Works that You distribute must include a readable copy of the attribution notices contained within such NOTICE file, excluding those notices that do not pertain to any part of the Derivative Works, in at least one of the following places: within a NOTICE text file distributed as part of the Derivative Works; within the Source form or documentation, if provided along with the Derivative Works; or, within a display generated by the Derivative Works, if and wherever such third-party notices normally appear. The contents of the NOTICE file are for informational purposes only and do not modify the License. You may add Your own attribution notices within Derivative Works that You distribute, alongside or as an addendum to the NOTICE text from the Work, provided that such additional attribution notices cannot be construed as modifying the License. 
 * 
 * You may add Your own copyright statement to Your modifications and may provide additional or different license terms and conditions for use, reproduction, or distribution of Your modifications, or for any such Derivative Works as a whole, provided Your use, reproduction, and distribution of the Work otherwise complies with the conditions stated in this License.
 * 5. Submission of Contributions. Unless You explicitly state otherwise, any Contribution intentionally submitted for inclusion in the Work by You to the Licensor shall be under the terms and conditions of this License, without any additional terms or conditions. Notwithstanding the above, nothing herein shall supersede or modify the terms of any separate license agreement you may have executed with Licensor regarding such Contributions.
 * 
 * 6. Trademarks. This License does not grant permission to use the trade names, trademarks, service marks, or product names of the Licensor, except as required for reasonable and customary use in describing the origin of the Work and reproducing the content of the NOTICE file.
 * 
 * 7. Disclaimer of Warranty. Unless required by applicable law or agreed to in writing, Licensor provides the Work (and each Contributor provides its Contributions) on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied, including, without limitation, any warranties or conditions of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A PARTICULAR PURPOSE. You are solely responsible for determining the appropriateness of using or redistributing the Work and assume any risks associated with Your exercise of permissions under this License.
 * 
 * 8. Limitation of Liability. In no event and under no legal theory, whether in tort (including negligence), contract, or otherwise, unless required by applicable law (such as deliberate and grossly negligent acts) or agreed to in writing, shall any Contributor be liable to You for damages, including any direct, indirect, special, incidental, or consequential damages of any character arising as a result of this License or out of the use or inability to use the Work (including but not limited to damages for loss of goodwill, work stoppage, computer failure or malfunction, or any and all other commercial damages or losses), even if such Contributor has been advised of the possibility of such damages.
 * 
 * 9. Accepting Warranty or Additional Liability. While redistributing the Work or Derivative Works thereof, You may choose to offer, and charge a fee for, acceptance of support, warranty, indemnity, or other liability obligations and/or rights consistent with this License. However, in accepting such obligations, You may act only on Your own behalf and on Your sole responsibility, not on behalf of any other Contributor, and only if You agree to indemnify, defend, and hold each Contributor harmless for any liability incurred by, or claims asserted against, such Contributor by reason of your accepting any such warranty or additional liability. 
 * 
 * END OF TERMS AND CONDITIONS
 ******************************************************************************/
package galileo.graph;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Logger;

import org.joda.time.DateTime;
import org.joda.time.DateTimeZone;

import galileo.bmp.CorrectedBitmap;
import galileo.comm.TemporalType;
import galileo.fs.GeospatialFileSystem;
import galileo.net.NetworkDestination;
import galileo.serialization.ByteSerializable;
import galileo.serialization.SerializationException;
import galileo.serialization.SerializationInputStream;
import galileo.serialization.SerializationOutputStream;
import galileo.util.GeoHash;

public class CliqueContainer implements ByteSerializable{
	
	private String geohashKey;
	private List<Integer> levels = null;
	private List<List<CacheCell>> cells = null;
	
	private List<CorrectedBitmap> bitmaps;
	
	private String geohashAntipode;
	private int direction;
	
	private int totalCliqueSize = 0;
	
	private NetworkDestination replicatedNode = null;
	
	private static final Logger logger = Logger.getLogger("galileo");
	
	public CliqueContainer(String key) {
		
		this.geohashKey = key;
		levels = new ArrayList<Integer>();
		cells = new ArrayList<List<CacheCell>>();
		
	}
	
	public String toString() {
		String ret = geohashKey;
		ret += "****"+levels +"****";
		ret += cells;
		
		return ret;
	}
	
	/**
	 * FOR A SPATIOTEMPORAL LEVEL, RETURN THE BITMAP AT THAT LEVEL
	 * @author sapmitra
	 * @return
	 */
	public CorrectedBitmap getBitmapAtLevel(int level) {
		
		CorrectedBitmap ret = null;
		
		int index = levels.indexOf(level);
		
		if(index >= 0) {
			
			return bitmaps.get(index);
			
		}
		
		return ret;
		
		
	}
	
	
	public List<CacheCell> getCacheCellsAtLevel(int level) {
		
		List<CacheCell> retCells = null;
		
		int index = levels.indexOf(level);
		
		if(index >= 0) {
			
			return cells.get(index);
			
		}
		
		return retCells;
	}
	
	public void addCells(int level, List<CacheCell> cells) {

		this.levels.add(level);
		
		this.cells.add(cells);
		
		totalCliqueSize += cells.size();
		
		
	}
	
	/**
	 * CREATES A LEVEL-WISE BITMAP OF ALL THE CELLS CONTAINED IN A CLIQUE 
	 * @author sapmitra
	 * @param fs
	 * @return
	 */
	
	
	public void calculateBitmap(GeospatialFileSystem fs) {
		
		List<CorrectedBitmap> bitmaps = new ArrayList<CorrectedBitmap>();
		
		int count = 0;
		
		for(List<CacheCell> cellRow: cells) {
			
			int currentLevel = levels.get(count);
			
			CorrectedBitmap bm = null;
			
			int[] levels = fs.getStCache().getSpatioTemporalResolutionsFromLevel(currentLevel);
			
			if(count == 0) {
				logger.info("RIKI: TESTING: "+currentLevel + " : BROKEN: "+levels[0]+" "+levels[1]+" :KEY:"+cellRow.get(0).getCellKey());
			}
			
			if(fs.isSubBlockLevel(levels[0], levels[1])) {
				
				// CREATE A SET OF TEMPORARY BITMAPS
				bm = createBitmapForClique(fs, levels[0], levels[1], cellRow);
				
				bitmaps.add(bm);
				
			} else {
				bitmaps.add(bm);
			}
			
			
			
			count++;
			
		}
		
		this.bitmaps = bitmaps;
		
		//logger.info("RIKI: BITMAP CALCULATED: "+bitmaps);
	}
	
	
	/**
	 * POPULATES A BITMAP FOR A SINGLE LEVEL OF CACHE CELLS IN A CLIQUE.
	 * THIS IS TO BE MAINTAINED ON THE DISTRESSED NODE SIDE
	 * 
	 * @author sapmitra
	 * @param temporaryBitmap
	 * @param currentLevel
	 * @param spatialLevel
	 * @param temporalLevel
	 * @param cellsInThisLevel
	 */
	public CorrectedBitmap createBitmapForClique(GeospatialFileSystem fs, int spatialLevel, int temporalLevel, List<CacheCell> cellsInThisLevel) {
		
		CorrectedBitmap temporaryBitmap = new CorrectedBitmap();
		
		int i = 0;
		
		for(CacheCell cell : cellsInThisLevel) {
			
			String tokens[] = cell.getCellKey().split("\\$\\$");
			
			//logger.info("RIKI: CACHE CELL KEY: "+cell.getCellKey());
			
			int fsSpatialResolution = fs.getGeohashPrecision();
			//int fsTemporalResolution = fs.getTemporalType().getType();
			
			String choppedGeohash = tokens[1].substring(fsSpatialResolution);
			
			int spatialSize = (int)java.lang.Math.pow(32, choppedGeohash.length());
			
			String dateString = tokens[0];
			
			String timeTokens[] = dateString.split("-");
			
			// THE START TIME FOR THE BLOCK ASSOSSIATED WITH THIS CLIQUE
			long cliqueStartTimeStamp = GeoHash.getStartTimeStamp(timeTokens[0], timeTokens[1], timeTokens[2], timeTokens[3], fs.getTemporalType());
			
			long cellTimestamp = GeoHash.getStartTimeStamp(dateString, TemporalType.getTypeFromLevel(temporalLevel));
			
			// returns a number between 0 and 31 for single character
			
			long spatialIndex = 0;
			
			if(choppedGeohash.length() > 0)
				 spatialIndex = GeoHash.hashToLong(choppedGeohash);
			
			DateTime startDate = new DateTime(cliqueStartTimeStamp, DateTimeZone.UTC);
			
			long temporalIndex = TemporalType.getTemporalIndex(startDate, cellTimestamp, temporalLevel);
			
			int bitIndex = (int)(temporalIndex*spatialSize + spatialIndex);
			
			temporaryBitmap.set(bitIndex);
			
			if(i == 0) {
				logger.info("RIKI: CHECKING: LEVELS: "+spatialLevel+" "+temporalLevel+" :::INFO:"+cell.getCellKey()+" "+spatialIndex+" "+spatialSize + " "+ temporalIndex);
			}
			
			i++;
				
				
			
		}
		temporaryBitmap.applyUpdates();
		
		return temporaryBitmap;
				
		
	}
	
	
	
	
	
	public List<Integer> getLevels() {
		return levels;
	}
	public void setLevels(List<Integer> levels) {
		this.levels = levels;
	}
	public List<List<CacheCell>> getCells() {
		return cells;
	}
	public void setCells(List<List<CacheCell>> cells) {
		this.cells = cells;
	}

	public String getGeohashKey() {
		return geohashKey;
	}

	public void setGeohashKey(String geohashKey) {
		this.geohashKey = geohashKey;
	}

	public int getTotalCliqueSize() {
		return totalCliqueSize;
	}

	public void setTotalCliqueSize(int totalCliqueSize) {
		this.totalCliqueSize = totalCliqueSize;
	}

	public String getGeohashAntipode() {
		return geohashAntipode;
	}

	public void setGeohashAntipode(String geohashAntipode) {
		this.geohashAntipode = geohashAntipode;
	}

	public int getDirection() {
		return direction;
	}

	public void setDirection(int direction) {
		this.direction = direction;
	}

	@Override
	public void serialize(SerializationOutputStream out) throws IOException {
		
		out.writeString(geohashKey);
		out.writeString(geohashAntipode);
		out.writeInt(direction);
		out.writeInt(totalCliqueSize);
		
		out.writeIntegerCollection(levels);
		
		//out.writeInt(levels.size());
		
		for(List<CacheCell> cellRow: cells)
			out.writeSerializableCollection(cellRow);
		
	}
	
	@Deserialize
	public CliqueContainer(SerializationInputStream in) throws IOException, SerializationException  {
		
		this.geohashKey = in.readString();
		this.geohashAntipode = in.readString();
		this.direction = in.readInt();
		this.totalCliqueSize = in.readInt();
		
		levels = new ArrayList<Integer>();
		
		in.readIntegerCollection(levels);
		
		cells = new ArrayList<List<CacheCell>>();
		
		for(int i=0; i< levels.size(); i++) {
			
			List<CacheCell> lCells = new ArrayList<>();
			
			in.readSerializableCollection(CacheCell.class, lCells);
			
			cells.add(lCells);
			
		}
		
		
	}

	public List<CorrectedBitmap> getBitmaps() {
		return bitmaps;
	}

	public void setBitmaps(List<CorrectedBitmap> bitmaps) {
		this.bitmaps = bitmaps;
	}

	public NetworkDestination getReplicatedNode() {
		return replicatedNode;
	}

	public void setReplicatedNode(NetworkDestination replicatedNode) {
		this.replicatedNode = replicatedNode;
	}

	
	

}
