/*******************************************************************************
 * Software in the Sustain Ecosystem are Released Under Terms of Apache Software License 
 * 
 * This research has been supported by funding from the US National Science Foundation's CSSI program through awards 1931363, 1931324, 1931335, and 1931283. The project is a joint effort involving Colorado State University, Arizona State University, the University of California-Irvine, and the University of Maryland - Baltimore County. All redistributions of the software must also include this information. 
 * 
 * TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 * 
 * 
 * 1. Definitions.
 * 
 * "License" shall mean the terms and conditions for use, reproduction, and distribution as defined by Sections 1 through 9 of this document.
 * 
 * "Licensor" shall mean the copyright owner or entity authorized by the copyright owner that is granting the License.
 * 
 * "Legal Entity" shall mean the union of the acting entity and all other entities that control, are controlled by, or are under common control with that entity. For the purposes of this definition, "control" means (i) the power, direct or indirect, to cause the direction or management of such entity, whether by contract or otherwise, or (ii) ownership of fifty percent (50%) or more of the outstanding shares, or (iii) beneficial ownership of such entity.
 * 
 * "You" (or "Your") shall mean an individual or Legal Entity exercising permissions granted by this License.
 * 
 * "Source" form shall mean the preferred form for making modifications, including but not limited to software source code, documentation source, and configuration files.
 * 
 * "Object" form shall mean any form resulting from mechanical transformation or translation of a Source form, including but not limited to compiled object code, generated documentation, and conversions to other media types.
 * 
 * "Work" shall mean the work of authorship, whether in Source or Object form, made available under the License, as indicated by a copyright notice that is included in or attached to the work (an example is provided in the Appendix below).
 * 
 * "Derivative Works" shall mean any work, whether in Source or Object form, that is based on (or derived from) the Work and for which the editorial revisions, annotations, elaborations, or other modifications represent, as a whole, an original work of authorship. For the purposes of this License, Derivative Works shall not include works that remain separable from, or merely link (or bind by name) to the interfaces of, the Work and Derivative Works thereof.
 * 
 * "Contribution" shall mean any work of authorship, including the original version of the Work and any modifications or additions to that Work or Derivative Works thereof, that is intentionally submitted to Licensor for inclusion in the Work by the copyright owner or by an individual or Legal Entity authorized to submit on behalf of the copyright owner. For the purposes of this definition, "submitted" means any form of electronic, verbal, or written communication sent to the Licensor or its representatives, including but not limited to communication on electronic mailing lists, source code control systems, and issue tracking systems that are managed by, or on behalf of, the Licensor for the purpose of discussing and improving the Work, but excluding communication that is conspicuously marked or otherwise designated in writing by the copyright owner as "Not a Contribution."
 * 
 * "Contributor" shall mean Licensor and any individual or Legal Entity on behalf of whom a Contribution has been received by Licensor and subsequently incorporated within the Work.
 * 
 * 2. Grant of Copyright License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable copyright license to reproduce, prepare Derivative Works of, publicly display, publicly perform, sublicense, and distribute the Work and such Derivative Works in Source or Object form.
 * 
 * 3. Grant of Patent License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable (except as stated in this section) patent license to make, have made, use, offer to sell, sell, import, and otherwise transfer the Work, where such license applies only to those patent claims licensable by such Contributor that are necessarily infringed by their Contribution(s) alone or by combination of their Contribution(s) with the Work to which such Contribution(s) was submitted. If You institute patent litigation against any entity (including a cross-claim or counterclaim in a lawsuit) alleging that the Work or a Contribution incorporated within the Work constitutes direct or contributory patent infringement, then any patent licenses granted to You under this License for that Work shall terminate as of the date such litigation is filed.
 * 
 * 4. Redistribution. You may reproduce and distribute copies of the Work or Derivative Works thereof in any medium, with or without modifications, and in Source or Object form, provided that You meet the following conditions:
 * 
 * You must give any other recipients of the Work or Derivative Works a copy of this License; and
 * You must cause any modified files to carry prominent notices stating that You changed the files; and
 * You must retain, in the Source form of any Derivative Works that You distribute, all copyright, patent, trademark, and attribution notices from the Source form of the Work, excluding those notices that do not pertain to any part of the Derivative Works; and
 * If the Work includes a "NOTICE" text file as part of its distribution, then any Derivative Works that You distribute must include a readable copy of the attribution notices contained within such NOTICE file, excluding those notices that do not pertain to any part of the Derivative Works, in at least one of the following places: within a NOTICE text file distributed as part of the Derivative Works; within the Source form or documentation, if provided along with the Derivative Works; or, within a display generated by the Derivative Works, if and wherever such third-party notices normally appear. The contents of the NOTICE file are for informational purposes only and do not modify the License. You may add Your own attribution notices within Derivative Works that You distribute, alongside or as an addendum to the NOTICE text from the Work, provided that such additional attribution notices cannot be construed as modifying the License. 
 * 
 * You may add Your own copyright statement to Your modifications and may provide additional or different license terms and conditions for use, reproduction, or distribution of Your modifications, or for any such Derivative Works as a whole, provided Your use, reproduction, and distribution of the Work otherwise complies with the conditions stated in this License.
 * 5. Submission of Contributions. Unless You explicitly state otherwise, any Contribution intentionally submitted for inclusion in the Work by You to the Licensor shall be under the terms and conditions of this License, without any additional terms or conditions. Notwithstanding the above, nothing herein shall supersede or modify the terms of any separate license agreement you may have executed with Licensor regarding such Contributions.
 * 
 * 6. Trademarks. This License does not grant permission to use the trade names, trademarks, service marks, or product names of the Licensor, except as required for reasonable and customary use in describing the origin of the Work and reproducing the content of the NOTICE file.
 * 
 * 7. Disclaimer of Warranty. Unless required by applicable law or agreed to in writing, Licensor provides the Work (and each Contributor provides its Contributions) on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied, including, without limitation, any warranties or conditions of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A PARTICULAR PURPOSE. You are solely responsible for determining the appropriateness of using or redistributing the Work and assume any risks associated with Your exercise of permissions under this License.
 * 
 * 8. Limitation of Liability. In no event and under no legal theory, whether in tort (including negligence), contract, or otherwise, unless required by applicable law (such as deliberate and grossly negligent acts) or agreed to in writing, shall any Contributor be liable to You for damages, including any direct, indirect, special, incidental, or consequential damages of any character arising as a result of this License or out of the use or inability to use the Work (including but not limited to damages for loss of goodwill, work stoppage, computer failure or malfunction, or any and all other commercial damages or losses), even if such Contributor has been advised of the possibility of such damages.
 * 
 * 9. Accepting Warranty or Additional Liability. While redistributing the Work or Derivative Works thereof, You may choose to offer, and charge a fee for, acceptance of support, warranty, indemnity, or other liability obligations and/or rights consistent with this License. However, in accepting such obligations, You may act only on Your own behalf and on Your sole responsibility, not on behalf of any other Contributor, and only if You agree to indemnify, defend, and hold each Contributor harmless for any liability incurred by, or claims asserted against, such Contributor by reason of your accepting any such warranty or additional liability. 
 * 
 * END OF TERMS AND CONDITIONS
 ******************************************************************************/
package galileo.graph;

import java.util.Arrays;
import java.util.Calendar;
import java.util.logging.Logger;

import org.joda.time.DateTime;
import org.json.JSONArray;
import org.json.JSONObject;

import com.googlecode.javaewah.EWAHCompressedBitmap;

import galileo.bmp.Bitmap;
import galileo.bmp.CorrectedBitmap;
import galileo.comm.TemporalType;
import galileo.dataset.TemporalProperties;
import galileo.dht.hash.TemporalHash;
import galileo.fs.GeospatialFileSystem;
import galileo.util.GeoHash;

public class SubBlockLevelBitmaps {
	private static final Logger logger = Logger.getLogger("galileo");
	
	// FS block level 
	private int spatialFSLevel;
	private int temporalFSLevel;
	
	// Number of sub block spatial levels
	private int spatialSubLevels;
	// Number of sub block level temporalLevels
	private int temporalSubLevels;
	private CorrectedBitmap[] spatioTemporalBitmaps;
	
	public SubBlockLevelBitmaps(int spatialSubLevels, int temporalSubLevels, int geohashPrecision, int temporalFSLevel) {
		
		// +1 for when either spatial or temporal level is same as FSLEVEL
		this.spatialSubLevels = spatialSubLevels + 1;
		this.temporalSubLevels = temporalSubLevels + 1;
		this.spatialFSLevel = geohashPrecision;
		this.temporalFSLevel = temporalFSLevel;
		this.spatioTemporalBitmaps = new CorrectedBitmap[(this.spatialSubLevels)*(this.spatialSubLevels)];
		
	}
	
	public SubBlockLevelBitmaps() {}
	
	/**
	 * CREATING JSON REPRESENTATION OF A SUBBLOCK BITMAP
	 * @author sapmitra
	 * @param path
	 * @return
	 */
	public JSONObject createJsonObject(String path) {
    	
    	JSONObject blockBitmap = new JSONObject();
    	
    	blockBitmap.put("path", path);
    	blockBitmap.put("spatialFSLevel", spatialFSLevel);
    	blockBitmap.put("temporalFSLevel", temporalFSLevel);
    	blockBitmap.put("spatialSubLevels", spatialSubLevels);
    	blockBitmap.put("temporalSubLevels", temporalSubLevels);
    	
    	
    	
		JSONArray spbMaps = new JSONArray();
		for(CorrectedBitmap bm : spatioTemporalBitmaps) {
			if(bm == null) {
				spbMaps.put(JSONObject.NULL);
			} else {
				
				JSONObject sJSON = bm.createJsonObject();
				spbMaps.put(sJSON);
			}
		}
		
	
    	blockBitmap.put("allbitmaps", spbMaps);
    	
    	return blockBitmap;
    }
    
	/**
	 * REGENERATING SUBBLOCK BITMAP FROM JSON OBJECT
	 * @author sapmitra
	 * @param jsonObj
	 */
    public void populateFromJson(JSONObject jsonObj) {
    	
    	this.spatialFSLevel = jsonObj.getInt("spatialFSLevel");
    	this.temporalFSLevel = jsonObj.getInt("temporalFSLevel");
    	this.spatialSubLevels = jsonObj.getInt("spatialSubLevels");
    	this.temporalSubLevels = jsonObj.getInt("temporalSubLevels");
    	
    	this.spatioTemporalBitmaps = new CorrectedBitmap[(this.spatialSubLevels)*(this.spatialSubLevels)];
    	
    	
    	JSONArray bmaps = jsonObj.getJSONArray("allbitmaps");
    	
		for (int i = 0; i < bmaps.length(); i++) {
			
			if(bmaps.get(i) != JSONObject.NULL) {
				JSONObject object = bmaps.getJSONObject(i);
				CorrectedBitmap b = new CorrectedBitmap();
				b.populateFromJson(object);
				spatioTemporalBitmaps[i] = b;
			} else {
				spatioTemporalBitmaps[i] = null;
			}
		}
    	
    }
	
	/**
	 * RETURNS THE BITMAP FOR A REQUESTED SPATIOTEMPORAL LEVEL
	 * @author sapmitra
	 * @param spatialLevel - actual sp level, not relative
	 * @param temporalLevel = actual tmp level, not relative
	 * @return
	 */
	public CorrectedBitmap getBitMapForParticularLevel(int spatialLevel, int temporalLevel) {
		int indx = getMapIndex(spatialLevel, temporalLevel);
		
		return spatioTemporalBitmaps[indx];
		
	}
	
	/**
	 * RETURNS THE BITMAP FOR A REQUESTED SPATIOTEMPORAL LEVEL
	 * @author sapmitra
	 * @param spatialLevel
	 * @param temporalLevel
	 * @return
	 */
	public CorrectedBitmap getBitMapForParticularLevel(int spatiotemporalLevel) {
		
		return spatioTemporalBitmaps[spatiotemporalLevel];
		
	}
	
	/**
	 * Returns the index for which particular bitmap to be used
	 * @author sapmitra
	 * @param spatialLevel - actual spatial level, not relative
	 * @param temporalLevel - actual temporal level, not relative
	 * @return
	 */
	
	public int getMapIndex(int spatialLevel, int temporalLevel) {
		// The following are relative level numbers
		int spLvl = spatialLevel - spatialFSLevel;
		int tLvl = temporalLevel - temporalFSLevel;
		
		int indx = (tLvl)*spatialSubLevels + spLvl;
		return indx;
	}
	
	/**
	 * 
	 * @author sapmitra
	 * @param records
	 * @param temporalPosn 
	 * @param spatialPosn2 
	 * @param spatialPosn1 
	 * @param startDate 
	 * @param removeLength 
	 */
	public void populateTemporaryBitmapUsingRecords (String[] records, int spatialPosn1, int spatialPosn2, int temporalPosn, int removeLength, DateTime startDate) {
		
		CorrectedBitmap[] temporaryBitmaps  = new CorrectedBitmap[(spatialSubLevels)*(temporalSubLevels)];
		
		for(String record: records) {
			
			String[] fields = record.split(",");
			
			// Highest resolution geohash allowed by visualization application
			String geoHash = GeoHash.encode(GeospatialFileSystem.parseFloat(fields[spatialPosn1]),
					GeospatialFileSystem.parseFloat(fields[spatialPosn2]), spatialFSLevel+spatialSubLevels-1);
			
			long timestamp = GeospatialFileSystem.reformatDatetime(fields[temporalPosn]);
			
			String choppedGeohash = geoHash.substring(removeLength);
			
			// RIKI-REMOVE
			//logger.info("RIKI: GEOHASH: "+geoHash+" CHOPPED GEOHASH: "+choppedGeohash);
			
			// CREATE A SET OF TEMPORARY BITMAPS
			populateBitmapUsingRecord(startDate, timestamp, choppedGeohash, temporaryBitmaps);
			
		}
		
		synchronized(spatioTemporalBitmaps) {
			// POPULATE THE ACTUAL BITMAPS USING THE TEMPORARY BITMAPS EXTRACTED FROM THE BLOCK
			for(int i = 0; i < temporaryBitmaps.length; i++) {
				if(temporaryBitmaps[i] != null) {
					
					temporaryBitmaps[i].applyUpdates();
					//RIKI-REMOVE
					//logger.info("BITMAP FOR LEVEL "+i+" IS "+temporaryBitmaps[i].toString());
					
					if(spatioTemporalBitmaps[i] == null) {
						spatioTemporalBitmaps[i] = temporaryBitmaps[i];
					} else {
						spatioTemporalBitmaps[i].applyUpdates(temporaryBitmaps[i].bmp);
					}
					
					//RIKI-REMOVE
					//logger.info("BITMAP FOR LEVEL "+i+" IS POPULATED");
				} else {
					//RIKI-REMOVE
					//logger.info("BITMAP FOR LEVEL "+i+" IS NOT AVAILABLE");
				}
		}
		}
	}
	
	/**
	 * 
	 * @author sapmitra
	 * @param startTime
	 * @param recordTimestamp
	 * @param choppedGeohash The remainder portion of the geohash below block level
	 */
	public void populateBitmapUsingRecord(DateTime startTime, long recordTimestamp, String choppedGeohash, CorrectedBitmap[] temporaryBitmaps) {
		
		//logger.info("===========CHOPPER GEOHASH:"+choppedGeohash+"===================");
		// Update each bitmap corresponding to a record
		for(int i = 0; i < spatialSubLevels; i++) {
			
			for(int j = 0; j < temporalSubLevels; j++) {
				
				if(i == 0 && j == 0) {
					continue;
				}
				
				int currentSpatialLevel = spatialFSLevel+i;
				int currentTemporalLevel = temporalFSLevel+j;
				
				int bitmapArrayIndx = getMapIndex(currentSpatialLevel, currentTemporalLevel);
				
				// THIS SPATIAL INDEX IS THE INDEX WITHIN AN INDIVIDUAL BITMAP 
				// AND SHOULD NOT BE CONFUSED WITH BITMAP ARRAY INDEX
				long spatialBitIndex = 0;
				
				// If we do not go down the spatial resolution but only in temporal resolution
				if(i > 0) {
					
					String partialGeohash = choppedGeohash.substring(0, i);
					
					// returns a number between 0 and 31 for single character
					spatialBitIndex = GeoHash.hashToLong(partialGeohash);
					
					// RIKI-REMOVE
					//logger.info("RIKI: PARTIAL GEOHASH " + partialGeohash + " SPATIAL INDEX: "+spatialBitIndex);
				}
				
				long temporalBitIndex = 0;
				
				if(j > 0) {
					
					temporalBitIndex = TemporalType.getTemporalIndex(startTime, recordTimestamp, currentTemporalLevel);
					
				}
				
				// RIKI-REMOVE
				//logger.info("RIKI: HERE");
				CorrectedBitmap bm = temporaryBitmaps[bitmapArrayIndx];
				
				if(bm == null) {
					bm = new CorrectedBitmap();
					temporaryBitmaps[bitmapArrayIndx] = bm;
				}
				
				int spatialSize = (int)java.lang.Math.pow(32, i);
				
				// RIKI-REMOVE
				//logger.info("RIKI: HERE1");
				
				int bitIndex = (int)(temporalBitIndex*spatialSize + spatialBitIndex);
				
				bm.set(bitIndex);
				
				//logger.info("RIKI: POPULATING BITMAP ARRAY LEVEL: "+bitmapArrayIndx+" SP "+spatialBitIndex+" TEMP: "+temporalBitIndex);
				
			}
		}
		
	}
	
	/**
	 * Given a bitmap bit position and resolution, this returns the key for a cell that that bit represents
	 * Cell key is in a time$$space format
	 * @author sapmitra
	 * @param bitmapIndex
	 * @param sresolution - actual, not relative
	 * @param tresolution - actual, not relative. This is the temporal resolution of the cells being fetched, not the block.
	 * @param sFSLevel - spatial resolution of a block
	 * @param tFSLevels
	 * @return
	 */
	public static String getKeyFromBitmapIndex(int bitmapIndex, String blockGeohash, int sresolution, int tresolution, int sFSLevel, long startTimestamp) {
		
		int spatialSize = (int)java.lang.Math.pow(32, sresolution - sFSLevel);
		
		int spatialIndex = bitmapIndex%spatialSize;
		
		int temporalIndex = bitmapIndex/spatialSize;
		
		//GeoHash.hashToLong(hash);
		
		String geohashString = blockGeohash;
		if(sresolution > sFSLevel)
			geohashString += GeoHash.fromLongToString(spatialIndex, sresolution-sFSLevel);
		
		long newTimestamp = TemporalType.addTime(startTimestamp, temporalIndex, tresolution);
		String temporalString = getTemporalString(newTimestamp, tresolution);
		
		return temporalString+"$$"+geohashString;
	}
	
	
	/**
	 * Create a temporal string with xx for missing fields
	 */
	private static String getTemporalString(long timestamp, int temporalLevel) {
		
		Calendar c = Calendar.getInstance();
		c.setTimeZone(TemporalHash.TIMEZONE);
		c.setTimeInMillis(timestamp);
		int hour = c.get(Calendar.HOUR_OF_DAY);
		int day = c.get(Calendar.DAY_OF_MONTH);
		int month = c.get(Calendar.MONTH) + 1;
		int year = c.get(Calendar.YEAR);
		
		switch (temporalLevel) {
		case 4:
			return String.format("%d-%d-%d-%d", year, month, day, hour);
		case 3:
			return String.format("%d-%d-%d-xx", year, month, day);
		case 2:
			return String.format("%d-%d-xx-xx", year, month);
		case 1:
			return String.format("%d-xx-xx-xx", year);
		}
		return String.format("%d-%d-%d-xx", year, month, day);
	}
	
	public static void main(String arg[]) {
		
		//long hashToLong = GeoHash.hashToLong("0");
		//System.out.println("0 "+hashToLong+GeoHash.fromLongToString(hashToLong));
		
		long hashToLong = GeoHash.hashToLong("cx0");
		System.out.println("cx "+hashToLong+" "+GeoHash.fromLongToString(hashToLong,3));
		
		
		//hashToLong = GeoHash.hashToLong("f");
		//System.out.println("f "+hashToLong+GeoHash.fromLongToString(hashToLong));
		//hashToLong = GeoHash.hashToLong("g");
		//System.out.println("g "+hashToLong+GeoHash.fromLongToString(hashToLong));
		//hashToLong = GeoHash.hashToLong("u");
		//System.out.println("u "+hashToLong+GeoHash.fromLongToString(hashToLong));
		//hashToLong = GeoHash.hashToLong("z");
		//System.out.println("z "+hashToLong+GeoHash.fromLongToString(hashToLong));
		
	}

	public int getSpatialFSLevel() {
		return spatialFSLevel;
	}

	public void setSpatialFSLevel(int spatialFSLevel) {
		this.spatialFSLevel = spatialFSLevel;
	}

	public int getTemporalFSLevel() {
		return temporalFSLevel;
	}

	public void setTemporalFSLevel(int temporalFSLevel) {
		this.temporalFSLevel = temporalFSLevel;
	}

	public int getSpatialLevels() {
		return spatialSubLevels;
	}

	public void setSpatialLevels(int spatialLevels) {
		this.spatialSubLevels = spatialLevels;
	}

	public int getTemporalLevels() {
		return temporalSubLevels;
	}

	public void setTemporalLevels(int temporalLevels) {
		this.temporalSubLevels = temporalLevels;
	}

	public CorrectedBitmap[] getSpatioTemporalBitmaps() {
		return spatioTemporalBitmaps;
	}

	public void setSpatioTemporalBitmaps(CorrectedBitmap[] spatioTemporalBitmaps) {
		this.spatioTemporalBitmaps = spatioTemporalBitmaps;
	}

}
