/*******************************************************************************
 * Software in the Sustain Ecosystem are Released Under Terms of Apache Software License 
 * 
 * This research has been supported by funding from the US National Science Foundation's CSSI program through awards 1931363, 1931324, 1931335, and 1931283. The project is a joint effort involving Colorado State University, Arizona State University, the University of California-Irvine, and the University of Maryland - Baltimore County. All redistributions of the software must also include this information. 
 * 
 * TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 * 
 * 
 * 1. Definitions.
 * 
 * "License" shall mean the terms and conditions for use, reproduction, and distribution as defined by Sections 1 through 9 of this document.
 * 
 * "Licensor" shall mean the copyright owner or entity authorized by the copyright owner that is granting the License.
 * 
 * "Legal Entity" shall mean the union of the acting entity and all other entities that control, are controlled by, or are under common control with that entity. For the purposes of this definition, "control" means (i) the power, direct or indirect, to cause the direction or management of such entity, whether by contract or otherwise, or (ii) ownership of fifty percent (50%) or more of the outstanding shares, or (iii) beneficial ownership of such entity.
 * 
 * "You" (or "Your") shall mean an individual or Legal Entity exercising permissions granted by this License.
 * 
 * "Source" form shall mean the preferred form for making modifications, including but not limited to software source code, documentation source, and configuration files.
 * 
 * "Object" form shall mean any form resulting from mechanical transformation or translation of a Source form, including but not limited to compiled object code, generated documentation, and conversions to other media types.
 * 
 * "Work" shall mean the work of authorship, whether in Source or Object form, made available under the License, as indicated by a copyright notice that is included in or attached to the work (an example is provided in the Appendix below).
 * 
 * "Derivative Works" shall mean any work, whether in Source or Object form, that is based on (or derived from) the Work and for which the editorial revisions, annotations, elaborations, or other modifications represent, as a whole, an original work of authorship. For the purposes of this License, Derivative Works shall not include works that remain separable from, or merely link (or bind by name) to the interfaces of, the Work and Derivative Works thereof.
 * 
 * "Contribution" shall mean any work of authorship, including the original version of the Work and any modifications or additions to that Work or Derivative Works thereof, that is intentionally submitted to Licensor for inclusion in the Work by the copyright owner or by an individual or Legal Entity authorized to submit on behalf of the copyright owner. For the purposes of this definition, "submitted" means any form of electronic, verbal, or written communication sent to the Licensor or its representatives, including but not limited to communication on electronic mailing lists, source code control systems, and issue tracking systems that are managed by, or on behalf of, the Licensor for the purpose of discussing and improving the Work, but excluding communication that is conspicuously marked or otherwise designated in writing by the copyright owner as "Not a Contribution."
 * 
 * "Contributor" shall mean Licensor and any individual or Legal Entity on behalf of whom a Contribution has been received by Licensor and subsequently incorporated within the Work.
 * 
 * 2. Grant of Copyright License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable copyright license to reproduce, prepare Derivative Works of, publicly display, publicly perform, sublicense, and distribute the Work and such Derivative Works in Source or Object form.
 * 
 * 3. Grant of Patent License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable (except as stated in this section) patent license to make, have made, use, offer to sell, sell, import, and otherwise transfer the Work, where such license applies only to those patent claims licensable by such Contributor that are necessarily infringed by their Contribution(s) alone or by combination of their Contribution(s) with the Work to which such Contribution(s) was submitted. If You institute patent litigation against any entity (including a cross-claim or counterclaim in a lawsuit) alleging that the Work or a Contribution incorporated within the Work constitutes direct or contributory patent infringement, then any patent licenses granted to You under this License for that Work shall terminate as of the date such litigation is filed.
 * 
 * 4. Redistribution. You may reproduce and distribute copies of the Work or Derivative Works thereof in any medium, with or without modifications, and in Source or Object form, provided that You meet the following conditions:
 * 
 * You must give any other recipients of the Work or Derivative Works a copy of this License; and
 * You must cause any modified files to carry prominent notices stating that You changed the files; and
 * You must retain, in the Source form of any Derivative Works that You distribute, all copyright, patent, trademark, and attribution notices from the Source form of the Work, excluding those notices that do not pertain to any part of the Derivative Works; and
 * If the Work includes a "NOTICE" text file as part of its distribution, then any Derivative Works that You distribute must include a readable copy of the attribution notices contained within such NOTICE file, excluding those notices that do not pertain to any part of the Derivative Works, in at least one of the following places: within a NOTICE text file distributed as part of the Derivative Works; within the Source form or documentation, if provided along with the Derivative Works; or, within a display generated by the Derivative Works, if and wherever such third-party notices normally appear. The contents of the NOTICE file are for informational purposes only and do not modify the License. You may add Your own attribution notices within Derivative Works that You distribute, alongside or as an addendum to the NOTICE text from the Work, provided that such additional attribution notices cannot be construed as modifying the License. 
 * 
 * You may add Your own copyright statement to Your modifications and may provide additional or different license terms and conditions for use, reproduction, or distribution of Your modifications, or for any such Derivative Works as a whole, provided Your use, reproduction, and distribution of the Work otherwise complies with the conditions stated in this License.
 * 5. Submission of Contributions. Unless You explicitly state otherwise, any Contribution intentionally submitted for inclusion in the Work by You to the Licensor shall be under the terms and conditions of this License, without any additional terms or conditions. Notwithstanding the above, nothing herein shall supersede or modify the terms of any separate license agreement you may have executed with Licensor regarding such Contributions.
 * 
 * 6. Trademarks. This License does not grant permission to use the trade names, trademarks, service marks, or product names of the Licensor, except as required for reasonable and customary use in describing the origin of the Work and reproducing the content of the NOTICE file.
 * 
 * 7. Disclaimer of Warranty. Unless required by applicable law or agreed to in writing, Licensor provides the Work (and each Contributor provides its Contributions) on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied, including, without limitation, any warranties or conditions of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A PARTICULAR PURPOSE. You are solely responsible for determining the appropriateness of using or redistributing the Work and assume any risks associated with Your exercise of permissions under this License.
 * 
 * 8. Limitation of Liability. In no event and under no legal theory, whether in tort (including negligence), contract, or otherwise, unless required by applicable law (such as deliberate and grossly negligent acts) or agreed to in writing, shall any Contributor be liable to You for damages, including any direct, indirect, special, incidental, or consequential damages of any character arising as a result of this License or out of the use or inability to use the Work (including but not limited to damages for loss of goodwill, work stoppage, computer failure or malfunction, or any and all other commercial damages or losses), even if such Contributor has been advised of the possibility of such damages.
 * 
 * 9. Accepting Warranty or Additional Liability. While redistributing the Work or Derivative Works thereof, You may choose to offer, and charge a fee for, acceptance of support, warranty, indemnity, or other liability obligations and/or rights consistent with this License. However, in accepting such obligations, You may act only on Your own behalf and on Your sole responsibility, not on behalf of any other Contributor, and only if You agree to indemnify, defend, and hold each Contributor harmless for any liability incurred by, or claims asserted against, such Contributor by reason of your accepting any such warranty or additional liability. 
 * 
 * END OF TERMS AND CONDITIONS
 ******************************************************************************/


package galileo.net;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.ClosedSelectorException;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.SocketChannel;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Provides client-side message routing capabilities. This includes connecting
 * to a remote server, and transmitting messages in non-blocking mode.
 *
 * @author malensek
 */
public class ClientMessageRouter extends MessageRouter {

    protected static final Logger logger = Logger.getLogger("galileo");

    protected Map<NetworkDestination, SocketChannel> destinationToSocket
        = new HashMap<>();
    protected Map<SocketChannel, NetworkDestination> socketToDestination
        = new HashMap<>();
    protected Map<SocketChannel, TransmissionTracker> socketToTracker
        = new HashMap<>();

    protected Queue<SocketChannel> pendingRegistrations
        = new ConcurrentLinkedQueue<>();
    
    private boolean messageFailure;

    public ClientMessageRouter()
    throws IOException {
        super();
        initializeSelector();
    }
    
    public ClientMessageRouter(boolean msgFail)
    	    throws IOException {
        this();
        this.messageFailure = msgFail;
    }

    public ClientMessageRouter(int readBufferSize, int maxWriteQueueSize)
    throws IOException {
        super(readBufferSize, maxWriteQueueSize);
        initializeSelector();
    }

    private void initializeSelector()
    throws IOException {
        this.selector = Selector.open();
        this.online = true;
        Thread selectorThread = new Thread(this);
        selectorThread.start();
    }

    @Override
    public void run() {
        while (online) {
            try {
                processPendingRegistrations();
                updateInterestOps();
                processSelectionKeys();
            } catch(ClosedSelectorException sce){
            	//ignore
            } catch (IOException e) {
                e.printStackTrace();
                if(this.messageFailure)
                	dispatchMessage(null);
            }
        }
    }

    /**
     * Handles pending registration operations on the Selector thread.
     */
    private void processPendingRegistrations()
    throws ClosedChannelException {
        Iterator<SocketChannel> it = pendingRegistrations.iterator();
        while (it.hasNext() == true) {
            SocketChannel channel = it.next();
            it.remove();

            TransmissionTracker tracker = socketToTracker.get(channel);
            channel.register(selector, SelectionKey.OP_CONNECT, tracker);
        }
    }

    /**
     * Ensures that a particular {@link NetworkDestination} has been connected
     * to, and retrieves its relevant {@link TransmissionTracker} instance.
     *
     * @param destination The NetworkDestination to ensure this MessageRouter is
     * connected to
     *
     * @return The TransmissionTracker associated with the NetworkDestination.
     */
    private TransmissionTracker ensureConnected(NetworkDestination destination)
    throws IOException {
        SocketChannel channel = destinationToSocket.get(destination);
        if (channel != null) {
            return socketToTracker.get(channel);
        }

        channel = SocketChannel.open();
        channel.configureBlocking(false);
        InetSocketAddress address = new InetSocketAddress(
                destination.getHostname(), destination.getPort());
        channel.connect(address);

        /* Update data structures for mapping between sockets/keys/trackers */
        destinationToSocket.put(destination, channel);
        socketToDestination.put(channel, destination);
        TransmissionTracker tracker = new TransmissionTracker(writeQueueSize);
        socketToTracker.put(channel, tracker);

        /* Finally, put this registration in the pending queue */
        pendingRegistrations.add(channel);

        return tracker;
    }

    /**
     * Sends a message to multiple network destinations.
     */
    public List<Transmission> broadcastMessage(
            Iterable<NetworkDestination> destinations, GalileoMessage message)
    throws IOException {
        List<Transmission> transmissions = new ArrayList<>();
        for (NetworkDestination destination : destinations) {
            Transmission trans = sendMessage(destination, message);
            transmissions.add(trans);
        }
        return transmissions;
    }

    /**
     * Sends a message to the specified network destination.  Connections are
     * completed during the first send operation.
     */
    public Transmission sendMessage(NetworkDestination destination,
            GalileoMessage message)
    throws IOException {

        /* Make sure this destination has been connected.  If not, this kicks
         * off the connection process. */
        TransmissionTracker tracker = ensureConnected(destination);

        /* Queue the data to be written */
        Transmission trans = null;
        ByteBuffer payload = wrapWithPrefix(message);
        try {
            trans = tracker.queueOutgoingData(payload);
        } catch (InterruptedException e) {
            e.printStackTrace();
        }

        /* Request for interestOps change. */
        SocketChannel channel = destinationToSocket.get(destination);
        if (channel.isRegistered() && channel.isConnected()) {
            changeInterest.put(channel.keyFor(this.selector),
                    SelectionKey.OP_READ | SelectionKey.OP_WRITE);
        }

        selector.wakeup();
        return trans;
    }

    @Override
    protected void disconnect(SelectionKey key) {
        /* Update our ClientMessageRouter-specific data structures when
         * disconnected. */
        SocketChannel channel = (SocketChannel) key.channel();
        NetworkDestination destination = socketToDestination.get(channel);
        destinationToSocket.remove(destination);
        socketToDestination.remove(channel);
        socketToTracker.remove(channel);
        super.disconnect(key);
    }

    /**
     * Forcibly shuts down the message processor and disconnects from any
     * connected server(s).  If pending writes have been queued, they will be
     * discarded.
     */
    public void forceShutdown() {
        shutdown(true);
    }

    /**
     * Shuts down the message processor and disconnects from the server(s).  If
     * pending writes have been queued, this method will block until the queue
     * is empty.
     */
    public void shutdown() {
        shutdown(false);
    }

    /**
     * Shuts down the message processor and disconnects from the server(s).
     *
     * @param forcible Whether or not to forcibly shut down (discard queued
     * messages).
     */
    private void shutdown(boolean forcible) {
        //TODO we need to start refusing send operations here (before doing a
        //final flush of the outgoing queues)
        Iterator<SelectionKey> keys = this.selector.keys().iterator();
        while(keys.hasNext()){
        	SelectionKey key = keys.next();
        	if (forcible == false) {
                safeShutdown(key);
            }
        	disconnect(key);
        }
        this.online = false;
        this.selector.wakeup();
        try {
			this.selector.close();
		} catch (IOException e) {
			logger.log(Level.SEVERE, "Failed to close the selector in ClientMessageRouter", e);
		}
    }

    /**
     * This method checks a given SelectionKey's write queue for pending
     * writes, and then does a series of sleep-checks until the queue is empty.
     *
     * @param key SelectionKey to monitor for pending writes.
     */
    private void safeShutdown(SelectionKey key) {
        TransmissionTracker tracker = TransmissionTracker.fromKey(key);
        Iterator<Transmission> it = tracker.pendingTransmissionIterator();
        while (it.hasNext()) {
            Transmission t = it.next();
            try {
                t.finish();
            } catch (InterruptedException e) {
                logger.warning("Interrupted during safe shutdown");
            }
        }
    }
}
