/*******************************************************************************
 * Software in the Sustain Ecosystem are Released Under Terms of Apache Software License 
 * 
 * This research has been supported by funding from the US National Science Foundation's CSSI program through awards 1931363, 1931324, 1931335, and 1931283. The project is a joint effort involving Colorado State University, Arizona State University, the University of California-Irvine, and the University of Maryland - Baltimore County. All redistributions of the software must also include this information. 
 * 
 * TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 * 
 * 
 * 1. Definitions.
 * 
 * "License" shall mean the terms and conditions for use, reproduction, and distribution as defined by Sections 1 through 9 of this document.
 * 
 * "Licensor" shall mean the copyright owner or entity authorized by the copyright owner that is granting the License.
 * 
 * "Legal Entity" shall mean the union of the acting entity and all other entities that control, are controlled by, or are under common control with that entity. For the purposes of this definition, "control" means (i) the power, direct or indirect, to cause the direction or management of such entity, whether by contract or otherwise, or (ii) ownership of fifty percent (50%) or more of the outstanding shares, or (iii) beneficial ownership of such entity.
 * 
 * "You" (or "Your") shall mean an individual or Legal Entity exercising permissions granted by this License.
 * 
 * "Source" form shall mean the preferred form for making modifications, including but not limited to software source code, documentation source, and configuration files.
 * 
 * "Object" form shall mean any form resulting from mechanical transformation or translation of a Source form, including but not limited to compiled object code, generated documentation, and conversions to other media types.
 * 
 * "Work" shall mean the work of authorship, whether in Source or Object form, made available under the License, as indicated by a copyright notice that is included in or attached to the work (an example is provided in the Appendix below).
 * 
 * "Derivative Works" shall mean any work, whether in Source or Object form, that is based on (or derived from) the Work and for which the editorial revisions, annotations, elaborations, or other modifications represent, as a whole, an original work of authorship. For the purposes of this License, Derivative Works shall not include works that remain separable from, or merely link (or bind by name) to the interfaces of, the Work and Derivative Works thereof.
 * 
 * "Contribution" shall mean any work of authorship, including the original version of the Work and any modifications or additions to that Work or Derivative Works thereof, that is intentionally submitted to Licensor for inclusion in the Work by the copyright owner or by an individual or Legal Entity authorized to submit on behalf of the copyright owner. For the purposes of this definition, "submitted" means any form of electronic, verbal, or written communication sent to the Licensor or its representatives, including but not limited to communication on electronic mailing lists, source code control systems, and issue tracking systems that are managed by, or on behalf of, the Licensor for the purpose of discussing and improving the Work, but excluding communication that is conspicuously marked or otherwise designated in writing by the copyright owner as "Not a Contribution."
 * 
 * "Contributor" shall mean Licensor and any individual or Legal Entity on behalf of whom a Contribution has been received by Licensor and subsequently incorporated within the Work.
 * 
 * 2. Grant of Copyright License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable copyright license to reproduce, prepare Derivative Works of, publicly display, publicly perform, sublicense, and distribute the Work and such Derivative Works in Source or Object form.
 * 
 * 3. Grant of Patent License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable (except as stated in this section) patent license to make, have made, use, offer to sell, sell, import, and otherwise transfer the Work, where such license applies only to those patent claims licensable by such Contributor that are necessarily infringed by their Contribution(s) alone or by combination of their Contribution(s) with the Work to which such Contribution(s) was submitted. If You institute patent litigation against any entity (including a cross-claim or counterclaim in a lawsuit) alleging that the Work or a Contribution incorporated within the Work constitutes direct or contributory patent infringement, then any patent licenses granted to You under this License for that Work shall terminate as of the date such litigation is filed.
 * 
 * 4. Redistribution. You may reproduce and distribute copies of the Work or Derivative Works thereof in any medium, with or without modifications, and in Source or Object form, provided that You meet the following conditions:
 * 
 * You must give any other recipients of the Work or Derivative Works a copy of this License; and
 * You must cause any modified files to carry prominent notices stating that You changed the files; and
 * You must retain, in the Source form of any Derivative Works that You distribute, all copyright, patent, trademark, and attribution notices from the Source form of the Work, excluding those notices that do not pertain to any part of the Derivative Works; and
 * If the Work includes a "NOTICE" text file as part of its distribution, then any Derivative Works that You distribute must include a readable copy of the attribution notices contained within such NOTICE file, excluding those notices that do not pertain to any part of the Derivative Works, in at least one of the following places: within a NOTICE text file distributed as part of the Derivative Works; within the Source form or documentation, if provided along with the Derivative Works; or, within a display generated by the Derivative Works, if and wherever such third-party notices normally appear. The contents of the NOTICE file are for informational purposes only and do not modify the License. You may add Your own attribution notices within Derivative Works that You distribute, alongside or as an addendum to the NOTICE text from the Work, provided that such additional attribution notices cannot be construed as modifying the License. 
 * 
 * You may add Your own copyright statement to Your modifications and may provide additional or different license terms and conditions for use, reproduction, or distribution of Your modifications, or for any such Derivative Works as a whole, provided Your use, reproduction, and distribution of the Work otherwise complies with the conditions stated in this License.
 * 5. Submission of Contributions. Unless You explicitly state otherwise, any Contribution intentionally submitted for inclusion in the Work by You to the Licensor shall be under the terms and conditions of this License, without any additional terms or conditions. Notwithstanding the above, nothing herein shall supersede or modify the terms of any separate license agreement you may have executed with Licensor regarding such Contributions.
 * 
 * 6. Trademarks. This License does not grant permission to use the trade names, trademarks, service marks, or product names of the Licensor, except as required for reasonable and customary use in describing the origin of the Work and reproducing the content of the NOTICE file.
 * 
 * 7. Disclaimer of Warranty. Unless required by applicable law or agreed to in writing, Licensor provides the Work (and each Contributor provides its Contributions) on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied, including, without limitation, any warranties or conditions of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A PARTICULAR PURPOSE. You are solely responsible for determining the appropriateness of using or redistributing the Work and assume any risks associated with Your exercise of permissions under this License.
 * 
 * 8. Limitation of Liability. In no event and under no legal theory, whether in tort (including negligence), contract, or otherwise, unless required by applicable law (such as deliberate and grossly negligent acts) or agreed to in writing, shall any Contributor be liable to You for damages, including any direct, indirect, special, incidental, or consequential damages of any character arising as a result of this License or out of the use or inability to use the Work (including but not limited to damages for loss of goodwill, work stoppage, computer failure or malfunction, or any and all other commercial damages or losses), even if such Contributor has been advised of the possibility of such damages.
 * 
 * 9. Accepting Warranty or Additional Liability. While redistributing the Work or Derivative Works thereof, You may choose to offer, and charge a fee for, acceptance of support, warranty, indemnity, or other liability obligations and/or rights consistent with this License. However, in accepting such obligations, You may act only on Your own behalf and on Your sole responsibility, not on behalf of any other Contributor, and only if You agree to indemnify, defend, and hold each Contributor harmless for any liability incurred by, or claims asserted against, such Contributor by reason of your accepting any such warranty or additional liability. 
 * 
 * END OF TERMS AND CONDITIONS
 ******************************************************************************/


package galileo.net;

import java.io.IOException;
import java.net.Socket;
import java.nio.BufferUnderflowException;
import java.nio.ByteBuffer;
import java.nio.channels.CancelledKeyException;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.ServerSocketChannel;
import java.nio.channels.SocketChannel;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.ConcurrentHashMap;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Provides an abstract implementation for consuming and publishing messages on
 * both the server and client side.
 *
 * @author malensek
 */
public abstract class MessageRouter implements Runnable {

    protected static final Logger logger = Logger.getLogger("galileo");

    /** The size (in bytes) of the message prefix used in the system. */
    public static final int PREFIX_SZ = Integer.SIZE / Byte.SIZE;

    /** The default read buffer size is 8 MB. */
    public static final int DEFAULT_READ_BUFFER_SIZE = 8388608;

    /** The default write queue allows 100 items to be inserted before it
     * starts blocking.  This prevents situations where the MessageRouter is
     * overwhelmed by an extreme number of write requests, exhausting available
     * resources. */
    public static final int DEFAULT_WRITE_QUEUE_SIZE = 100;

    /** System property that overrides the read buffer size. */
    public static final String READ_BUFFER_PROPERTY
        = "galileo.net.MessageRouter.readBufferSize";

    /** System property that overrides the write queue maximum size. */
    public static final String WRITE_QUEUE_PROPERTY
        = "galileo.net.MessageRouter.writeQueueSize";

    /** Flag used to determine whether the Selector thread should run */
    protected boolean online;

    private List<MessageListener> listeners = new ArrayList<>();

    protected Selector selector;

    protected int readBufferSize;
    protected int writeQueueSize;
    private ByteBuffer readBuffer;

    protected ConcurrentHashMap<SelectionKey, Integer> changeInterest
        = new ConcurrentHashMap<>();

    public MessageRouter() {
        this(DEFAULT_READ_BUFFER_SIZE, DEFAULT_WRITE_QUEUE_SIZE);
    }

    public MessageRouter(int readBufferSize, int maxWriteQueueSize) {
        String readSz = System.getProperty(READ_BUFFER_PROPERTY);
        if (readSz == null) {
            this.readBufferSize = readBufferSize;
        } else {
            this.readBufferSize = Integer.parseInt(readSz);
        }

        String queueSz = System.getProperty(WRITE_QUEUE_PROPERTY);
        if (queueSz == null) {
            this.writeQueueSize = maxWriteQueueSize;
        } else {
            this.writeQueueSize = Integer.parseInt(queueSz);
        }

        readBuffer = ByteBuffer.allocateDirect(this.readBufferSize);
    }

    /**
     * As long as the MessageRouter is online, monitor connection operations
     * through the Selector instance.
     */
    @Override
    public void run() {
        while (online) {
            try {
                updateInterestOps();
                processSelectionKeys();
            } catch (IOException e) {
                logger.log(Level.WARNING, "Error in selector thread", e);
            }
        }
    }

    /**
     * Updates interest sets for any SelectionKey instances that require
     * changes.  This allows external threads to queue up changes to the
     * interest sets that will be fulfilled by the selector thread.
     */
    protected void updateInterestOps() {
        Iterator<SelectionKey> it = changeInterest.keySet().iterator();
        while (it.hasNext()) {
            SelectionKey key = it.next();
            if (key.isValid()) {
                SocketChannel channel = (SocketChannel) key.channel();
                if (channel.isConnected() == false
                        || channel.isRegistered() == false) {
                    continue;
                }
                key.interestOps(changeInterest.get(key));
            }
            changeInterest.remove(key);
        }
    }

    /**
     * Performs a select operation, and then processes the resulting
     * SelectionKey set based on interest ops.
     */
    protected void processSelectionKeys()
    throws IOException {

        selector.select();

        Iterator<SelectionKey> keys = selector.selectedKeys().iterator();
        while (keys.hasNext()) {
            SelectionKey key = keys.next();
            keys.remove();

            if (key.isValid() == false) {
                continue;
            }

            try {

                if (key.isAcceptable()) {
                    accept(key);
                    continue;
                }

                if (key.isConnectable()) {
                    connect(key);
                    continue;
                }

                if (key.isWritable()) {
                    write(key);
                }

                if (key.isReadable()) {
                    read(key);
                }

            } catch (CancelledKeyException e) {
                /* SelectionKey was cancelled by another thread. */
                continue;
            }
        }
    }

    /**
     * Accepts new connections.
     *
     * @param key The SelectionKey for the connecting client.
     */
    protected void accept(SelectionKey key)
    throws IOException {
        ServerSocketChannel servSocket = (ServerSocketChannel) key.channel();
        SocketChannel channel = servSocket.accept();
        logger.info("Accepted connection: " + getClientString(channel));

        TransmissionTracker tracker = new TransmissionTracker(writeQueueSize);
        channel.configureBlocking(false);
        channel.register(selector, SelectionKey.OP_READ, tracker);

        dispatchConnect(getDestination(channel));
    }

    /**
     * Finishes setting up a connection on a SocketChannel.
     *
     * @param key SelectionKey for the SocketChannel.
     * @throws IOException 
     */
    protected void connect(SelectionKey key) throws IOException {
        try {
            SocketChannel channel = (SocketChannel) key.channel();

            if (channel.finishConnect()) {
                TransmissionTracker tracker = TransmissionTracker.fromKey(key);
                if (tracker.hasPendingData() == false) {
                    changeInterest.put(key, SelectionKey.OP_READ);
                } else {
                    /* Data has already been queued up; start writing */
                    changeInterest.put(key,
                            SelectionKey.OP_READ | SelectionKey.OP_WRITE);
                }
            }

            dispatchConnect(getDestination(channel));
        } catch (IOException e) {
            logger.log(Level.INFO, "Connection finalization failed", e);
            disconnect(key);
            throw e;
        }
    }

    /**
     * Read data from a SocketChannel.
     *
     * @param key SelectionKey for the SocketChannel.
     */
    protected void read(SelectionKey key) {
        SocketChannel channel = (SocketChannel) key.channel();
        readBuffer.clear();

        int bytesRead = 0;

        try {
            /* Read data from the channel */
            while ((bytesRead = channel.read(readBuffer)) > 0) {
                readBuffer.flip();
                processIncomingMessage(key);
            }
        } catch (IOException e) {
            logger.log(Level.FINE, "Abnormal remote termination", e);
            disconnect(key);
            return;
        } catch (BufferUnderflowException e) {
            /* Incoming packets lied about their size! */
            logger.log(Level.WARNING, "Incoming packet size mismatch", e);
        }

        if (bytesRead == -1) {
            /* Connection was terminated by the client. */
            logger.fine("Reached EOF in channel input stream");
            disconnect(key);
            return;
        }
    }

    /**
     * Process data received from a client SocketChannel.  This method is
     * chiefly concerned with processing incoming data streams into
     * GalileoMessage packets to be consumed by the system.
     *
     * @param key SelectionKey for the client.
     */
    protected void processIncomingMessage(SelectionKey key) {
        TransmissionTracker transmission = TransmissionTracker.fromKey(key);
        if (transmission.expectedBytes == 0) {
            /* We don't know how much data the client is sending yet.
             * Read the message prefix to determine the payload size. */
            boolean ready = readPrefix(readBuffer, transmission);

            /* Check if we have read the payload size prefix yet.  If
             * not, then we're done for now. */
            if (ready == false) {
                return;
            }
        }

        int readSize = transmission.expectedBytes - transmission.readPointer;
        if (readSize > readBuffer.remaining()) {
            readSize = readBuffer.remaining();
        }

        readBuffer.get(transmission.payload,
                transmission.readPointer, readSize);
        transmission.readPointer += readSize;

        if (transmission.readPointer == transmission.expectedBytes) {
            /* The payload has been read */
            GalileoMessage msg = new GalileoMessage(
                    transmission.payload,
                    new MessageContext(this, key));
            dispatchMessage(msg);
            transmission.resetCounters();

            if (readBuffer.hasRemaining()) {
                /* There is another payload to read */
                processIncomingMessage(key);
                /* Note: this process continues until we reach the end of the
                 * buffer.  Not doing so would cause us to lose data. */
            }
        }
    }

    /**
     * Read the payload size prefix from a channel.
     * Each message in Galileo is prefixed with a payload size field; this is
     * read to allocate buffers for the incoming message.
     *
     * @return true if the payload size has been determined; false otherwise.
     */
    protected static boolean readPrefix(ByteBuffer buffer,
            TransmissionTracker transmission) {
        /* Make sure the prefix hasn't already been read. */
        if (transmission.expectedBytes != 0) {
            return true;
        }

        /* Can we determine the payload size in one shot? (we must read at least
         * PREFIX_SZ bytes) */
        if (transmission.prefixPointer == 0
                && buffer.remaining() >= PREFIX_SZ) {
            transmission.expectedBytes = buffer.getInt();
            transmission.allocatePayload();
            return true;
        } else {
            /* Keep reading until we have at least PREFIX_SZ bytes to determine
             * the payload size.  */

            int prefixLeft = PREFIX_SZ - transmission.prefixPointer;
            if (buffer.remaining() < prefixLeft) {
                prefixLeft = buffer.remaining();
            }

            buffer.get(transmission.prefix,
                    transmission.prefixPointer, prefixLeft);
            transmission.prefixPointer += prefixLeft;

            if (transmission.prefixPointer >= PREFIX_SZ) {
                ByteBuffer buf = ByteBuffer.wrap(transmission.prefix);
                transmission.expectedBytes = buf.getInt();
                transmission.allocatePayload();
                return true;
            }
        }

        return false;
    }

    /**
     * Wraps a given message in a {@link ByteBuffer}, including the payload size
     * prefix.  Data produced by this method will be subsequently read by the
     * readPrefix() method.
     */
    protected static ByteBuffer wrapWithPrefix(GalileoMessage message) {
        int messageSize = message.getPayload().length;
        ByteBuffer buffer = ByteBuffer.allocate(messageSize + 4);
        buffer.putInt(messageSize);
        buffer.put(message.getPayload());
        buffer.flip();
        return buffer;
    }

    /**
     * Adds a message to the pending write queue for a particular SelectionKey
     * and submits a change request for its interest set. Pending data is placed
     * in a blocking queue, so this function may block to prevent queueing an
     * excessive amount of data.
     * <p>
     * The system property <em>galileo.net.MessageRouter.writeQueueSize</em>
     * tunes the maximum amount of data that can be queued.
     *
     * @param key SelectionKey for the channel.
     * @param message GalileoMessage to publish on the channel.
     *
     * @return {@link Transmission} instance representing the send operation.
     */
    public Transmission sendMessage(SelectionKey key, GalileoMessage message)
    throws IOException {
        //TODO reduce the visibility of this method to protected
        if (this.isOnline() == false) {
            throw new IOException("MessageRouter is not online.");
        }

        TransmissionTracker tracker = TransmissionTracker.fromKey(key);
        ByteBuffer payload = wrapWithPrefix(message);

        Transmission trans = null;
        try {
            tracker.queueOutgoingData(payload);
        } catch (InterruptedException e) {
            Thread.currentThread().interrupt();
            throw new IOException("Interrupted while waiting to queue data");
        }

        changeInterest.put(key, SelectionKey.OP_READ | SelectionKey.OP_WRITE);
        selector.wakeup();
        return trans;
    }

    /**
     * When a {@link SelectionKey} is writable, push as much pending data
     * out on the channel as possible.
     *
     * @param key {@link SelectionKey} of the channel to write to.
     */
    private void write(SelectionKey key) {
        TransmissionTracker tracker = TransmissionTracker.fromKey(key);
        SocketChannel channel = (SocketChannel) key.channel();

        while (tracker.hasPendingData() == true) {
            Transmission trans = tracker.getNextTransmission();
            ByteBuffer buffer = trans.getPayload();
            if (buffer == null) {
                break;
            }

            int written = 0;
            while (buffer.hasRemaining()) {
                try {
                    written = channel.write(buffer);
                } catch (IOException e) {
                    /* Broken pipe */
                    disconnect(key);
                    return;
                }

                if (buffer.hasRemaining() == false) {
                    /* Done writing */
                    tracker.transmissionFinished();
                }

                if (written == 0) {
                    /* Return now, to keep our OP_WRITE interest op set. */
                    return;
                }
            }
        }

        /* At this point, the queue is empty. */
        key.interestOps(SelectionKey.OP_READ);
        return;
    }

    /**
     * Handle termination of connections.
     *
     * @param key The SelectionKey of the SocketChannel that has disconnected.
     */
    protected void disconnect(SelectionKey key) {
        if (key.isValid() == false) {
            return;
        }

        SocketChannel channel = (SocketChannel) key.channel();
        NetworkDestination destination = getDestination(channel);
        logger.info("Terminating connection: " + destination.toString());

        try {
            key.cancel();
            key.channel().close();
        } catch (IOException e) {
            logger.log(Level.WARNING, "Failed to disconnect channel", e);
        }

        dispatchDisconnect(destination);
    }

    /**
     * Adds a message listener (consumer) to this MessageRouter.  Listeners
     * receive messages that are published by this MessageRouter.
     *
     * @param listener {@link MessageListener} that will consume messages
     * published by this MessageRouter.
     */
    public void addListener(MessageListener listener) {
        listeners.add(listener);
    }

    /**
     * Dispatches a message to all listening consumers.
     *
     * @param message {@link GalileoMessage} to dispatch.
     */
    protected void dispatchMessage(GalileoMessage message) {
        for (MessageListener listener : listeners) {
            listener.onMessage(message);
        }
    }

    /**
     * Informs all listening consumers that a connection to a remote endpoint
     * has been made.
     */
    protected void dispatchConnect(NetworkDestination endpoint) {
        for (MessageListener listener : listeners) {
            listener.onConnect(endpoint);
        }
    }

    /**
     * Informs all listening consumers that a connection to a remote endpoint
     * has been terminated.
     */
    protected void dispatchDisconnect(NetworkDestination endpoint) {
        for (MessageListener listener : listeners) {
            listener.onDisconnect(endpoint);
        }
    }

    /**
     * Determines whether or not this MessageRouter is online.  As long as the
     * router is online, the selector thread will continue to run.
     *
     * @return true if the MessageRouter instance is online and running.
     */
    public boolean isOnline() {
        return this.online;
    }

    /**
     * Determines a connection's hostname and port, then concatenates the two
     * values, separated by a colon (:).
     *
     * @param channel Channel to get client information about.
     */
    protected static String getClientString(SocketChannel channel) {
        Socket socket = channel.socket();
        return socket.getInetAddress().getHostName() + ":" + socket.getPort();
    }

    /**
     * Determines a connection's endpoint information (hostname and port) and
     * encapsulates them in a {@link NetworkDestination}.
     *
     * @param channel The SocketChannel of the network endpoint.
     *
     * @return NetworkDestination representation of the endpoint.
     */
    protected static NetworkDestination getDestination(SocketChannel channel) {
        Socket socket = channel.socket();
        return new NetworkDestination(
                socket.getInetAddress().getHostName(),
                socket.getPort());
    }
}
