/*******************************************************************************
 * Software in the Sustain Ecosystem are Released Under Terms of Apache Software License 
 * 
 * This research has been supported by funding from the US National Science Foundation's CSSI program through awards 1931363, 1931324, 1931335, and 1931283. The project is a joint effort involving Colorado State University, Arizona State University, the University of California-Irvine, and the University of Maryland - Baltimore County. All redistributions of the software must also include this information. 
 * 
 * TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 * 
 * 
 * 1. Definitions.
 * 
 * "License" shall mean the terms and conditions for use, reproduction, and distribution as defined by Sections 1 through 9 of this document.
 * 
 * "Licensor" shall mean the copyright owner or entity authorized by the copyright owner that is granting the License.
 * 
 * "Legal Entity" shall mean the union of the acting entity and all other entities that control, are controlled by, or are under common control with that entity. For the purposes of this definition, "control" means (i) the power, direct or indirect, to cause the direction or management of such entity, whether by contract or otherwise, or (ii) ownership of fifty percent (50%) or more of the outstanding shares, or (iii) beneficial ownership of such entity.
 * 
 * "You" (or "Your") shall mean an individual or Legal Entity exercising permissions granted by this License.
 * 
 * "Source" form shall mean the preferred form for making modifications, including but not limited to software source code, documentation source, and configuration files.
 * 
 * "Object" form shall mean any form resulting from mechanical transformation or translation of a Source form, including but not limited to compiled object code, generated documentation, and conversions to other media types.
 * 
 * "Work" shall mean the work of authorship, whether in Source or Object form, made available under the License, as indicated by a copyright notice that is included in or attached to the work (an example is provided in the Appendix below).
 * 
 * "Derivative Works" shall mean any work, whether in Source or Object form, that is based on (or derived from) the Work and for which the editorial revisions, annotations, elaborations, or other modifications represent, as a whole, an original work of authorship. For the purposes of this License, Derivative Works shall not include works that remain separable from, or merely link (or bind by name) to the interfaces of, the Work and Derivative Works thereof.
 * 
 * "Contribution" shall mean any work of authorship, including the original version of the Work and any modifications or additions to that Work or Derivative Works thereof, that is intentionally submitted to Licensor for inclusion in the Work by the copyright owner or by an individual or Legal Entity authorized to submit on behalf of the copyright owner. For the purposes of this definition, "submitted" means any form of electronic, verbal, or written communication sent to the Licensor or its representatives, including but not limited to communication on electronic mailing lists, source code control systems, and issue tracking systems that are managed by, or on behalf of, the Licensor for the purpose of discussing and improving the Work, but excluding communication that is conspicuously marked or otherwise designated in writing by the copyright owner as "Not a Contribution."
 * 
 * "Contributor" shall mean Licensor and any individual or Legal Entity on behalf of whom a Contribution has been received by Licensor and subsequently incorporated within the Work.
 * 
 * 2. Grant of Copyright License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable copyright license to reproduce, prepare Derivative Works of, publicly display, publicly perform, sublicense, and distribute the Work and such Derivative Works in Source or Object form.
 * 
 * 3. Grant of Patent License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable (except as stated in this section) patent license to make, have made, use, offer to sell, sell, import, and otherwise transfer the Work, where such license applies only to those patent claims licensable by such Contributor that are necessarily infringed by their Contribution(s) alone or by combination of their Contribution(s) with the Work to which such Contribution(s) was submitted. If You institute patent litigation against any entity (including a cross-claim or counterclaim in a lawsuit) alleging that the Work or a Contribution incorporated within the Work constitutes direct or contributory patent infringement, then any patent licenses granted to You under this License for that Work shall terminate as of the date such litigation is filed.
 * 
 * 4. Redistribution. You may reproduce and distribute copies of the Work or Derivative Works thereof in any medium, with or without modifications, and in Source or Object form, provided that You meet the following conditions:
 * 
 * You must give any other recipients of the Work or Derivative Works a copy of this License; and
 * You must cause any modified files to carry prominent notices stating that You changed the files; and
 * You must retain, in the Source form of any Derivative Works that You distribute, all copyright, patent, trademark, and attribution notices from the Source form of the Work, excluding those notices that do not pertain to any part of the Derivative Works; and
 * If the Work includes a "NOTICE" text file as part of its distribution, then any Derivative Works that You distribute must include a readable copy of the attribution notices contained within such NOTICE file, excluding those notices that do not pertain to any part of the Derivative Works, in at least one of the following places: within a NOTICE text file distributed as part of the Derivative Works; within the Source form or documentation, if provided along with the Derivative Works; or, within a display generated by the Derivative Works, if and wherever such third-party notices normally appear. The contents of the NOTICE file are for informational purposes only and do not modify the License. You may add Your own attribution notices within Derivative Works that You distribute, alongside or as an addendum to the NOTICE text from the Work, provided that such additional attribution notices cannot be construed as modifying the License. 
 * 
 * You may add Your own copyright statement to Your modifications and may provide additional or different license terms and conditions for use, reproduction, or distribution of Your modifications, or for any such Derivative Works as a whole, provided Your use, reproduction, and distribution of the Work otherwise complies with the conditions stated in this License.
 * 5. Submission of Contributions. Unless You explicitly state otherwise, any Contribution intentionally submitted for inclusion in the Work by You to the Licensor shall be under the terms and conditions of this License, without any additional terms or conditions. Notwithstanding the above, nothing herein shall supersede or modify the terms of any separate license agreement you may have executed with Licensor regarding such Contributions.
 * 
 * 6. Trademarks. This License does not grant permission to use the trade names, trademarks, service marks, or product names of the Licensor, except as required for reasonable and customary use in describing the origin of the Work and reproducing the content of the NOTICE file.
 * 
 * 7. Disclaimer of Warranty. Unless required by applicable law or agreed to in writing, Licensor provides the Work (and each Contributor provides its Contributions) on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied, including, without limitation, any warranties or conditions of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A PARTICULAR PURPOSE. You are solely responsible for determining the appropriateness of using or redistributing the Work and assume any risks associated with Your exercise of permissions under this License.
 * 
 * 8. Limitation of Liability. In no event and under no legal theory, whether in tort (including negligence), contract, or otherwise, unless required by applicable law (such as deliberate and grossly negligent acts) or agreed to in writing, shall any Contributor be liable to You for damages, including any direct, indirect, special, incidental, or consequential damages of any character arising as a result of this License or out of the use or inability to use the Work (including but not limited to damages for loss of goodwill, work stoppage, computer failure or malfunction, or any and all other commercial damages or losses), even if such Contributor has been advised of the possibility of such damages.
 * 
 * 9. Accepting Warranty or Additional Liability. While redistributing the Work or Derivative Works thereof, You may choose to offer, and charge a fee for, acceptance of support, warranty, indemnity, or other liability obligations and/or rights consistent with this License. However, in accepting such obligations, You may act only on Your own behalf and on Your sole responsibility, not on behalf of any other Contributor, and only if You agree to indemnify, defend, and hold each Contributor harmless for any liability incurred by, or claims asserted against, such Contributor by reason of your accepting any such warranty or additional liability. 
 * 
 * END OF TERMS AND CONDITIONS
 ******************************************************************************/
package galileo.util;

import java.util.ArrayList;
import java.util.List;

import org.json.JSONArray;
import org.json.JSONObject;

public class BorderingProperties {
	
	private List<String> n;
	private List<String> s;
	private List<String> e;
	private List<String> w;
	private String ne;
	private String se;
	private String nw;
	private String sw;
	
	/* Remember Up refers to upper time, so essentially the end part of this time span */
	private long up1;
	private long up2;
	private long down1;
	private long down2;
	
	private long totalRecords;
	
	private List<Long> upTimeEntries;
	private List<Long> downTimeEntries;
	//private List<Integer> fringeEntries;
	
	private List<Long> northEntries;
	private List<Long> southEntries;
	private List<Long> eastEntries;
	private List<Long> westEntries;
	private List<Long> neEntries;
	private List<Long> nwEntries;
	private List<Long> seEntries;
	private List<Long> swEntries;
	
	public String toString() {
		String ret = "";
		ret+=n+" "+s+" "+e+" "+w+" "+down1+" "+down2+" "+up1+" "+up2+" "+northEntries+" "
				+southEntries+" "+eastEntries+" "+westEntries+" "+neEntries+" "+nwEntries+" "+seEntries+" "+swEntries+" ";
		return ret;
	}
	
	public JSONObject getJsonStringRepresentation(String path) {
		JSONObject bmap = new JSONObject();
		
		bmap.put("blockName", path);
		bmap.put("n",n);
		bmap.put("s",s);
		bmap.put("e",e);
		bmap.put("w",w);
		bmap.put("ne",ne);
		bmap.put("nw",nw);
		bmap.put("se",se);
		bmap.put("sw",sw);
		
		bmap.put("northEntries",northEntries);
		bmap.put("southEntries",southEntries);
		bmap.put("eastEntries",eastEntries);
		bmap.put("westEntries",westEntries);
		bmap.put("neEntries",neEntries);
		bmap.put("nwEntries",nwEntries);
		bmap.put("seEntries",seEntries);
		bmap.put("swEntries",swEntries);
		
		bmap.put("up1", up1);
		bmap.put("up2", up2);
		bmap.put("down1", down1);
		bmap.put("down2", down2);
		
		bmap.put("totalRecords", totalRecords);
		
		bmap.put("upTimeEntries",upTimeEntries);
		bmap.put("downTimeEntries",downTimeEntries);
		//bmap.put("fringeEntries",fringeEntries);
		
		
		return bmap;
	}
	
	public void populateObject(JSONObject jsonObj) {
		this.up1 = jsonObj.getLong("up1");
		this.up2 = jsonObj.getLong("up2");
		this.down1 = jsonObj.getLong("down1");
		this.down2 = jsonObj.getLong("down2");
		this.totalRecords = jsonObj.getLong("totalRecords");
		
		JSONArray ns = jsonObj.getJSONArray("n");
		for (int i = 0; i < ns.length(); i++)
			n.add(ns.getString(i));
		
		JSONArray ss = jsonObj.getJSONArray("s");
		for (int i = 0; i < ss.length(); i++)
			s.add(ss.getString(i));
		
		JSONArray es = jsonObj.getJSONArray("e");
		for (int i = 0; i < es.length(); i++)
			e.add(es.getString(i));
		
		JSONArray ws = jsonObj.getJSONArray("w");
		for (int i = 0; i < ws.length(); i++)
			w.add(ws.getString(i));
		
		
		this.ne = jsonObj.getString("ne");
		this.nw = jsonObj.getString("nw");
		this.se = jsonObj.getString("se");
		this.sw = jsonObj.getString("sw");
		
		
		JSONArray nEntries = jsonObj.getJSONArray("northEntries");
		for (int i = 0; i < nEntries.length(); i++)
			northEntries.add(nEntries.getLong(i));
		
		JSONArray sEntries = jsonObj.getJSONArray("southEntries");
		for (int i = 0; i < sEntries.length(); i++)
			southEntries.add(sEntries.getLong(i));
		
		JSONArray eEntries = jsonObj.getJSONArray("eastEntries");
		for (int i = 0; i < eEntries.length(); i++)
			eastEntries.add(eEntries.getLong(i));
		
		JSONArray wEntries = jsonObj.getJSONArray("westEntries");
		for (int i = 0; i < wEntries.length(); i++)
			westEntries.add(wEntries.getLong(i));
		
		JSONArray neEntriess = jsonObj.getJSONArray("neEntries");
		for (int i = 0; i < neEntriess.length(); i++)
			neEntries.add(neEntriess.getLong(i));
		
		JSONArray nwEntriess = jsonObj.getJSONArray("nwEntries");
		for (int i = 0; i < nwEntriess.length(); i++)
			nwEntries.add(nwEntriess.getLong(i));
		
		JSONArray seEntriess = jsonObj.getJSONArray("seEntries");
		for (int i = 0; i < seEntriess.length(); i++)
			seEntries.add(seEntriess.getLong(i));
		
		JSONArray swEntriess = jsonObj.getJSONArray("swEntries");
		for (int i = 0; i < swEntriess.length(); i++)
			swEntries.add(swEntriess.getLong(i));
		
		JSONArray upTimeEntriess = jsonObj.getJSONArray("upTimeEntries");
		for (int i = 0; i < upTimeEntriess.length(); i++)
			upTimeEntries.add(upTimeEntriess.getLong(i));
		
		JSONArray downTimeEntriess = jsonObj.getJSONArray("downTimeEntries");
		for (int i = 0; i < downTimeEntriess.length(); i++)
			downTimeEntries.add(downTimeEntriess.getLong(i));
		/*
		
		JSONArray fringeEntriess = jsonObj.getJSONArray("fringeEntries");
		for (int i = 0; i < fringeEntriess.length(); i++)
			fringeEntries.add(fringeEntriess.getInt(i));
		*/
	}
	
	public BorderingProperties() {
		
		n = new ArrayList<String>();
		s = new ArrayList<String>();
		e = new ArrayList<String>();
		w = new ArrayList<String>();
		
		totalRecords = 0;
		
		northEntries = new ArrayList<Long>();
		southEntries = new ArrayList<Long>();
		eastEntries = new ArrayList<Long>();
		westEntries = new ArrayList<Long>();
		neEntries = new ArrayList<Long>();
		nwEntries = new ArrayList<Long>();
		seEntries = new ArrayList<Long>();
		swEntries = new ArrayList<Long>();
		
		upTimeEntries = new ArrayList<Long>();
		downTimeEntries = new ArrayList<Long>();
		//fringeEntries = new ArrayList<Integer>();
	}
	
	public synchronized void updateRecordCount(long n) {
		totalRecords+=n;
	}

	public List<String> getN() {
		return n;
	}

	public void setN(List<String> n) {
		this.n = n;
	}

	public List<String> getS() {
		return s;
	}

	public void setS(List<String> s) {
		this.s = s;
	}

	public List<String> getE() {
		return e;
	}

	public void setE(List<String> e) {
		this.e = e;
	}

	public List<String> getW() {
		return w;
	}

	public void setW(List<String> w) {
		this.w = w;
	}

	public String getNe() {
		return ne;
	}

	public void setNe(String ne) {
		this.ne = ne;
	}

	public String getSe() {
		return se;
	}

	public void setSe(String se) {
		this.se = se;
	}

	public String getNw() {
		return nw;
	}

	public void setNw(String nw) {
		this.nw = nw;
	}

	public String getSw() {
		return sw;
	}

	public void setSw(String sw) {
		this.sw = sw;
	}
	
	public void addE(String str) {
		e.add(str);
	}
	
	public void addW(String str) {
		w.add(str);
	}
	
	public void addN(String str) {
		n.add(str);
	}
	
	public void addS(String str) {
		s.add(str);
	}

	public List<Long> getNorthEntries() {
		return northEntries;
	}

	public void setNorthEntries(List<Long> northEntries) {
		this.northEntries = northEntries;
	}

	public List<Long> getSouthEntries() {
		return southEntries;
	}

	public void setSouthEntries(List<Long> southEntries) {
		this.southEntries = southEntries;
	}

	public List<Long> getEastEntries() {
		return eastEntries;
	}

	public void setEastEntries(List<Long> eastEntries) {
		this.eastEntries = eastEntries;
	}

	public List<Long> getWestEntries() {
		return westEntries;
	}

	public void setWestEntries(List<Long> westEntries) {
		this.westEntries = westEntries;
	}

	public List<Long> getNeEntries() {
		return neEntries;
	}

	public void setNeEntries(List<Long> neEntries) {
		this.neEntries = neEntries;
	}

	public List<Long> getNwEntries() {
		return nwEntries;
	}

	public void setNwEntries(List<Long> nwEntries) {
		this.nwEntries = nwEntries;
	}

	public List<Long> getSeEntries() {
		return seEntries;
	}

	public void setSeEntries(List<Long> seEntries) {
		this.seEntries = seEntries;
	}

	public List<Long> getSwEntries() {
		return swEntries;
	}

	public void setSwEntries(List<Long> swEntries) {
		this.swEntries = swEntries;
	}
	
	public void addNorthEntries(long n) {
		northEntries.add(n);
	}
	
	public void addSouthEntries(long n) {
		southEntries.add(n);
	}
	
	public void addEastEntries(long n) {
		eastEntries.add(n);
	}
	
	public void addWestEntries(long n) {
		westEntries.add(n);
	}
	
	public void addNEEntries(long n) {
		neEntries.add(n);
	}
	
	public void addNWEntries(long n) {
		nwEntries.add(n);
	}
	
	public void addSEEntries(long n) {
		seEntries.add(n);
	}
	
	public void addSWEntries(long n) {
		swEntries.add(n);
	}

	public long getUp1() {
		return up1;
	}

	public void setUp1(long up1) {
		this.up1 = up1;
	}

	public long getUp2() {
		return up2;
	}

	public void setUp2(long up2) {
		this.up2 = up2;
	}

	public long getDown1() {
		return down1;
	}

	public void setDown1(long down1) {
		this.down1 = down1;
	}

	public long getDown2() {
		return down2;
	}

	public void setDown2(long down2) {
		this.down2 = down2;
	}

	public List<Long> getUpTimeEntries() {
		return upTimeEntries;
	}

	public void setUpTimeEntries(List<Long> upTimeEntries) {
		this.upTimeEntries = upTimeEntries;
	}
	

	public void addUpTimeEntries(long n) {
		upTimeEntries.add(n);
	}

	public List<Long> getDownTimeEntries() {
		return downTimeEntries;
	}

	public void setDownTimeEntries(List<Long> downTimeEntries) {
		this.downTimeEntries = downTimeEntries;
	}
	
	public void addDownTimeEntries(long n) {
		downTimeEntries.add(n);
	}

	public long getTotalRecords() {
		return totalRecords;
	}

	public void setTotalRecords(long totalRecords) {
		this.totalRecords = totalRecords;
	}

	/*public List<Integer> getFringeEntries() {
		return fringeEntries;
	}

	public void setFringeEntries(List<Integer> fringeEntries) {
		this.fringeEntries = fringeEntries;
	}
	
	public void addFringeEntries(int n) {
		this.fringeEntries.add(n);
	}*/
 
}
