/*******************************************************************************
 * Software in the Sustain Ecosystem are Released Under Terms of Apache Software License 
 * 
 * This research has been supported by funding from the US National Science Foundation's CSSI program through awards 1931363, 1931324, 1931335, and 1931283. The project is a joint effort involving Colorado State University, Arizona State University, the University of California-Irvine, and the University of Maryland - Baltimore County. All redistributions of the software must also include this information. 
 * 
 * TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 * 
 * 
 * 1. Definitions.
 * 
 * "License" shall mean the terms and conditions for use, reproduction, and distribution as defined by Sections 1 through 9 of this document.
 * 
 * "Licensor" shall mean the copyright owner or entity authorized by the copyright owner that is granting the License.
 * 
 * "Legal Entity" shall mean the union of the acting entity and all other entities that control, are controlled by, or are under common control with that entity. For the purposes of this definition, "control" means (i) the power, direct or indirect, to cause the direction or management of such entity, whether by contract or otherwise, or (ii) ownership of fifty percent (50%) or more of the outstanding shares, or (iii) beneficial ownership of such entity.
 * 
 * "You" (or "Your") shall mean an individual or Legal Entity exercising permissions granted by this License.
 * 
 * "Source" form shall mean the preferred form for making modifications, including but not limited to software source code, documentation source, and configuration files.
 * 
 * "Object" form shall mean any form resulting from mechanical transformation or translation of a Source form, including but not limited to compiled object code, generated documentation, and conversions to other media types.
 * 
 * "Work" shall mean the work of authorship, whether in Source or Object form, made available under the License, as indicated by a copyright notice that is included in or attached to the work (an example is provided in the Appendix below).
 * 
 * "Derivative Works" shall mean any work, whether in Source or Object form, that is based on (or derived from) the Work and for which the editorial revisions, annotations, elaborations, or other modifications represent, as a whole, an original work of authorship. For the purposes of this License, Derivative Works shall not include works that remain separable from, or merely link (or bind by name) to the interfaces of, the Work and Derivative Works thereof.
 * 
 * "Contribution" shall mean any work of authorship, including the original version of the Work and any modifications or additions to that Work or Derivative Works thereof, that is intentionally submitted to Licensor for inclusion in the Work by the copyright owner or by an individual or Legal Entity authorized to submit on behalf of the copyright owner. For the purposes of this definition, "submitted" means any form of electronic, verbal, or written communication sent to the Licensor or its representatives, including but not limited to communication on electronic mailing lists, source code control systems, and issue tracking systems that are managed by, or on behalf of, the Licensor for the purpose of discussing and improving the Work, but excluding communication that is conspicuously marked or otherwise designated in writing by the copyright owner as "Not a Contribution."
 * 
 * "Contributor" shall mean Licensor and any individual or Legal Entity on behalf of whom a Contribution has been received by Licensor and subsequently incorporated within the Work.
 * 
 * 2. Grant of Copyright License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable copyright license to reproduce, prepare Derivative Works of, publicly display, publicly perform, sublicense, and distribute the Work and such Derivative Works in Source or Object form.
 * 
 * 3. Grant of Patent License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable (except as stated in this section) patent license to make, have made, use, offer to sell, sell, import, and otherwise transfer the Work, where such license applies only to those patent claims licensable by such Contributor that are necessarily infringed by their Contribution(s) alone or by combination of their Contribution(s) with the Work to which such Contribution(s) was submitted. If You institute patent litigation against any entity (including a cross-claim or counterclaim in a lawsuit) alleging that the Work or a Contribution incorporated within the Work constitutes direct or contributory patent infringement, then any patent licenses granted to You under this License for that Work shall terminate as of the date such litigation is filed.
 * 
 * 4. Redistribution. You may reproduce and distribute copies of the Work or Derivative Works thereof in any medium, with or without modifications, and in Source or Object form, provided that You meet the following conditions:
 * 
 * You must give any other recipients of the Work or Derivative Works a copy of this License; and
 * You must cause any modified files to carry prominent notices stating that You changed the files; and
 * You must retain, in the Source form of any Derivative Works that You distribute, all copyright, patent, trademark, and attribution notices from the Source form of the Work, excluding those notices that do not pertain to any part of the Derivative Works; and
 * If the Work includes a "NOTICE" text file as part of its distribution, then any Derivative Works that You distribute must include a readable copy of the attribution notices contained within such NOTICE file, excluding those notices that do not pertain to any part of the Derivative Works, in at least one of the following places: within a NOTICE text file distributed as part of the Derivative Works; within the Source form or documentation, if provided along with the Derivative Works; or, within a display generated by the Derivative Works, if and wherever such third-party notices normally appear. The contents of the NOTICE file are for informational purposes only and do not modify the License. You may add Your own attribution notices within Derivative Works that You distribute, alongside or as an addendum to the NOTICE text from the Work, provided that such additional attribution notices cannot be construed as modifying the License. 
 * 
 * You may add Your own copyright statement to Your modifications and may provide additional or different license terms and conditions for use, reproduction, or distribution of Your modifications, or for any such Derivative Works as a whole, provided Your use, reproduction, and distribution of the Work otherwise complies with the conditions stated in this License.
 * 5. Submission of Contributions. Unless You explicitly state otherwise, any Contribution intentionally submitted for inclusion in the Work by You to the Licensor shall be under the terms and conditions of this License, without any additional terms or conditions. Notwithstanding the above, nothing herein shall supersede or modify the terms of any separate license agreement you may have executed with Licensor regarding such Contributions.
 * 
 * 6. Trademarks. This License does not grant permission to use the trade names, trademarks, service marks, or product names of the Licensor, except as required for reasonable and customary use in describing the origin of the Work and reproducing the content of the NOTICE file.
 * 
 * 7. Disclaimer of Warranty. Unless required by applicable law or agreed to in writing, Licensor provides the Work (and each Contributor provides its Contributions) on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied, including, without limitation, any warranties or conditions of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A PARTICULAR PURPOSE. You are solely responsible for determining the appropriateness of using or redistributing the Work and assume any risks associated with Your exercise of permissions under this License.
 * 
 * 8. Limitation of Liability. In no event and under no legal theory, whether in tort (including negligence), contract, or otherwise, unless required by applicable law (such as deliberate and grossly negligent acts) or agreed to in writing, shall any Contributor be liable to You for damages, including any direct, indirect, special, incidental, or consequential damages of any character arising as a result of this License or out of the use or inability to use the Work (including but not limited to damages for loss of goodwill, work stoppage, computer failure or malfunction, or any and all other commercial damages or losses), even if such Contributor has been advised of the possibility of such damages.
 * 
 * 9. Accepting Warranty or Additional Liability. While redistributing the Work or Derivative Works thereof, You may choose to offer, and charge a fee for, acceptance of support, warranty, indemnity, or other liability obligations and/or rights consistent with this License. However, in accepting such obligations, You may act only on Your own behalf and on Your sole responsibility, not on behalf of any other Contributor, and only if You agree to indemnify, defend, and hold each Contributor harmless for any liability incurred by, or claims asserted against, such Contributor by reason of your accepting any such warranty or additional liability. 
 * 
 * END OF TERMS AND CONDITIONS
 ******************************************************************************/


package galileo.util;

import java.awt.Graphics2D;
import java.awt.Polygon;
import java.awt.Rectangle;
import java.awt.geom.Area;
import java.awt.geom.Rectangle2D;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Set;
import java.util.TimeZone;
import java.util.concurrent.ThreadLocalRandom;
import java.util.logging.Logger;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import galileo.bmp.Bitmap;
import galileo.comm.TemporalType;
import galileo.dataset.Coordinates;
import galileo.dataset.Point;
import galileo.dataset.SpatialRange;
import galileo.dataset.TemporalProperties;
import galileo.dht.Partitioner;
import galileo.dht.StandardDHTPartitioner;
import galileo.dht.hash.TemporalHash;
import galileo.fs.GeospatialFileSystem;
//import math.geom2d.Point2D;
//import math.geom2d.polygon.SimplePolygon2D;

/**
 * This class provides an implementation of the GeoHash (http://www.geohash.org)
 * algorithm.
 *
 * See http://en.wikipedia.org/wiki/Geohash for implementation details.
 */
public class GeoHash {

	private static final Logger logger = Logger.getLogger("galileo");
	public final static byte BITS_PER_CHAR = 5;
	public final static int LATITUDE_RANGE = 90;
	public final static int LONGITUDE_RANGE = 180;
	public final static int MAX_PRECISION = 30; // 6 character precision = 30 (~
												// 1.2km x 0.61km)

	public static List<Character> evenChars = Arrays.asList('b', 'c', 'f', 'g', 'u', 'v', 'y', 'z', '8', '9', 'd', 'e',
			's', 't', 'w', 'x', '2', '3', '6', '7', 'k', 'm', 'q', 'r', '0', '1', '4', '5', 'h', 'j', 'n', 'p');
	public static List<Character> oddChars = Arrays.asList('p', 'r', 'x', 'z', 'n', 'q', 'w', 'y', 'j', 'm', 't', 'v',
			'h', 'k', 's', 'u', '5', '7', 'e', 'g', '4', '6', 'd', 'f', '1', '3', '9', 'c', '0', '2', '8', 'b');

	private static final String BASE32 = "0123456789bcdefghjkmnpqrstuvwxyz";
	/**
	 * This character array maps integer values (array indices) to their GeoHash
	 * base32 alphabet equivalents.
	 */
	public final static char[] charMap = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'b', 'c', 'd', 'e', 'f',
			'g', 'h', 'j', 'k', 'm', 'n', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z' };

	/**
	 * Allows lookups from a GeoHash character to its integer index value.
	 */
	public final static HashMap<Character, Integer> charLookupTable = new HashMap<Character, Integer>();

	/**
	 * Initialize HashMap for character to integer lookups.
	 */
	static {
		for (int i = 0; i < charMap.length; ++i) {
			charLookupTable.put(charMap[i], i);
		}
	}

	private String binaryHash;
	private Rectangle2D bounds;

	public GeoHash() {
		this("");
	}

	public GeoHash(String binaryString) {
		this.binaryHash = binaryString;
		ArrayList<Boolean> bits = new ArrayList<>();
		for (char bit : this.binaryHash.toCharArray())
			bits.add(bit == '0' ? false : true);
		float[] longitude = decodeBits(bits, false);
		float[] latitude = decodeBits(bits, true);
		SpatialRange range = new SpatialRange(latitude[0], latitude[1], longitude[0], longitude[1]);
		Pair<Coordinates, Coordinates> coordsPair = range.get2DCoordinates();
		Point<Integer> upLeft = coordinatesToXY(coordsPair.a);
		Point<Integer> lowRight = coordinatesToXY(coordsPair.b);
		this.bounds = new Rectangle(upLeft.X(), upLeft.Y(), lowRight.X() - upLeft.X(), lowRight.Y() - upLeft.Y());
	}

	public int getPrecision() {
		return this.binaryHash.length();
	}

	public String getBinaryHash() {
		return this.binaryHash;
	}

	public String[] getValues(int precision) {
		String[] values = null;
		String hash = "";
		for (int i = 0; i < this.binaryHash.length(); i += 5) {
			String hashChar = this.binaryHash.substring(i, java.lang.Math.min(i + 5, this.binaryHash.length()));
			if (hashChar.length() == 5)
				hash += charMap[Integer.parseInt(hashChar, 2)];
			else {
				String beginHash = hashChar;
				String endHash = hashChar;
				while (beginHash.length() < BITS_PER_CHAR) {
					beginHash += "0";
					endHash += "1";
				}
				values = new String[2];
				values[0] = hash + charMap[Integer.parseInt(beginHash, 2)];
				values[1] = hash + charMap[Integer.parseInt(endHash, 2)];
				while (values[0].length() < precision) {
					values[0] += "0";
					values[1] += "z";
				}
			}
		}
		if (values == null) {
			if (hash.length() < precision) {
				String beginHash = hash;
				String endHash = hash;
				while (beginHash.length() < precision) {
					beginHash += "0";
					endHash += "z";
				}
				values = new String[] { beginHash, endHash };
			} else {
				values = new String[] { hash };
			}
		}
		return values;
	}

	public Rectangle2D getRectangle() {
		return this.bounds;
	}

	@Override
	public boolean equals(Object obj) {
		if (obj instanceof GeoHash) {
			GeoHash other = (GeoHash) obj;
			return this.binaryHash.equals(other.binaryHash);
		}
		return false;
	}

	@Override
	public int hashCode() {
		return this.binaryHash.hashCode();
	}

	/**
	 * GET STARTIMESTAMP FROM A TEMPORAL STRING
	 * 
	 * @author sapmitra
	 * @param temporalString
	 * @param temporalType
	 * @return
	 * @throws ParseException
	 */
	public static long getStartTimeStamp(String temporalString, TemporalType temporalType) {

		String timeTokens[] = temporalString.split("-");
		return getStartTimeStamp(timeTokens[0], timeTokens[1], timeTokens[2], timeTokens[3], temporalType);

	}

	/**
	 * GET STARTIMESTAMP FROM A TEMPORAL STRING
	 * 
	 * @author sapmitra
	 * @param temporalString
	 * @param temporalType
	 * @return
	 * @throws ParseException
	 */
	public static long getEndTimeStamp(String temporalString, TemporalType temporalType) {

		String timeTokens[] = temporalString.split("-");
		return getEndTimeStamp(timeTokens[0], timeTokens[1], timeTokens[2], timeTokens[3], temporalType);

	}

	/**
	 * 
	 * @author sapmitra
	 * @param year
	 * @param month
	 * @param day
	 * @param hour
	 * @param temporalType
	 * @return
	 * @throws ParseException
	 */
	public static long getStartTimeStamp(String year, String month, String day, String hour,
			TemporalType temporalType) {

		Calendar calendar = Calendar.getInstance(TemporalHash.TIMEZONE);
		calendar.setTimeZone(TemporalHash.TIMEZONE);
		// calendar.setLenient(false);

		if (month.contains("x"))
			month = "00";
		if (day.contains("x"))
			day = "01";
		if (hour.contains("x"))
			hour = "00";

		switch (temporalType) {
		case HOUR_OF_DAY:
			calendar.set(Calendar.YEAR, Integer.parseInt(year));
			calendar.set(Calendar.MONTH, Integer.parseInt(month) - 1);
			calendar.set(Calendar.DAY_OF_MONTH, Integer.parseInt(day));
			calendar.set(Calendar.HOUR_OF_DAY, Integer.parseInt(hour));
			calendar.set(Calendar.MINUTE, 0);
			calendar.set(Calendar.SECOND, 0);
			calendar.set(Calendar.MILLISECOND, 0);
			break;
		case DAY_OF_MONTH:
			calendar.set(Calendar.YEAR, Integer.parseInt(year));
			calendar.set(Calendar.MONTH, Integer.parseInt(month) - 1);
			calendar.set(Calendar.DAY_OF_MONTH, Integer.parseInt(day));
			calendar.set(Calendar.HOUR_OF_DAY, 0);
			calendar.set(Calendar.MINUTE, 0);
			calendar.set(Calendar.SECOND, 0);
			calendar.set(Calendar.MILLISECOND, 0);
			break;
		case MONTH:
			calendar.set(Calendar.YEAR, Integer.parseInt(year));
			calendar.set(Calendar.MONTH, Integer.parseInt(month) - 1);
			calendar.set(Calendar.DAY_OF_MONTH, 1);
			calendar.set(Calendar.HOUR_OF_DAY, 0);
			calendar.set(Calendar.MINUTE, 0);
			calendar.set(Calendar.SECOND, 0);
			calendar.set(Calendar.MILLISECOND, 0);
			break;
		case YEAR:
			calendar.set(Calendar.YEAR, Integer.parseInt(year));
			calendar.set(Calendar.MONTH, 0);
			calendar.set(Calendar.DAY_OF_MONTH, 1);
			calendar.set(Calendar.HOUR_OF_DAY, 0);
			calendar.set(Calendar.MINUTE, 0);
			calendar.set(Calendar.SECOND, 0);
			calendar.set(Calendar.MILLISECOND, 0);
			break;

		}

		// long offset = TemporalHash.TIMEZONE.getRawOffset() -
		// TimeZone.getDefault().getRawOffset();

		return calendar.getTimeInMillis();

	}

	public static long getStartTimeStamp_BackupWithOffset(String year, String month, String day, String hour,
			TemporalType temporalType) throws ParseException {

		Calendar calendar = Calendar.getInstance(TemporalHash.TIMEZONE);
		calendar.setTimeZone(TemporalHash.TIMEZONE);
		// calendar.setLenient(false);

		if (month.contains("x"))
			month = "00";
		if (day.contains("x"))
			day = "01";
		if (hour.contains("x"))
			hour = "00";

		switch (temporalType) {
		case HOUR_OF_DAY:
			calendar.set(Calendar.YEAR, Integer.parseInt(year));
			calendar.set(Calendar.MONTH, Integer.parseInt(month) - 1);
			calendar.set(Calendar.DAY_OF_MONTH, Integer.parseInt(day));
			calendar.set(Calendar.HOUR_OF_DAY, Integer.parseInt(hour));
			calendar.set(Calendar.MINUTE, 0);
			calendar.set(Calendar.SECOND, 0);
			calendar.set(Calendar.MILLISECOND, 0);
			break;
		case DAY_OF_MONTH:
			calendar.set(Calendar.YEAR, Integer.parseInt(year));
			calendar.set(Calendar.MONTH, Integer.parseInt(month) - 1);
			calendar.set(Calendar.DAY_OF_MONTH, Integer.parseInt(day));
			calendar.set(Calendar.HOUR_OF_DAY, 0);
			calendar.set(Calendar.MINUTE, 0);
			calendar.set(Calendar.SECOND, 0);
			calendar.set(Calendar.MILLISECOND, 0);
			break;
		case MONTH:
			calendar.set(Calendar.YEAR, Integer.parseInt(year));
			calendar.set(Calendar.MONTH, Integer.parseInt(month) - 1);
			calendar.set(Calendar.DAY_OF_MONTH, 1);
			calendar.set(Calendar.HOUR_OF_DAY, 0);
			calendar.set(Calendar.MINUTE, 0);
			calendar.set(Calendar.SECOND, 0);
			calendar.set(Calendar.MILLISECOND, 0);
			break;
		case YEAR:
			calendar.set(Calendar.YEAR, Integer.parseInt(year));
			calendar.set(Calendar.MONTH, 0);
			calendar.set(Calendar.DAY_OF_MONTH, 1);
			calendar.set(Calendar.HOUR_OF_DAY, 0);
			calendar.set(Calendar.MINUTE, 0);
			calendar.set(Calendar.SECOND, 0);
			calendar.set(Calendar.MILLISECOND, 0);
			break;

		}

		long offset = TemporalHash.TIMEZONE.getRawOffset() - TimeZone.getDefault().getRawOffset();

		return calendar.getTimeInMillis() + offset;

	}

	/**
	 * 
	 * @author sapmitra
	 * @param year
	 * @param month
	 * @param day
	 * @param hour
	 * @param temporalType
	 * @return
	 */
	public static long getEndTimeStamp(String year, String month, String day, String hour, TemporalType temporalType) {

		Calendar calendar = Calendar.getInstance(TemporalHash.TIMEZONE);
		calendar.setTimeZone(TemporalHash.TIMEZONE);
		// calendar.setLenient(false);

		if (month.contains("x"))
			month = "11";

		int mn = Integer.parseInt(month);

		if (day.contains("x")) {
			if (mn == 1 || mn == 3 || mn == 5 || mn == 7 || mn == 8 || mn == 10 || mn == 12)
				day = "31";
			else if (mn == 2) {
				int yr = Integer.parseInt(year);
				if (yr % 4 == 0)
					day = "29";
				else
					day = "28";
			}
		}

		if (hour.contains("x"))
			hour = "23";

		switch (temporalType) {
		case HOUR_OF_DAY:
			calendar.set(Calendar.YEAR, Integer.parseInt(year));
			calendar.set(Calendar.MONTH, Integer.parseInt(month) - 1);
			calendar.set(Calendar.DAY_OF_MONTH, Integer.parseInt(day));
			calendar.set(Calendar.HOUR_OF_DAY, Integer.parseInt(hour));
			calendar.set(Calendar.MINUTE, 59);
			calendar.set(Calendar.SECOND, 59);
			calendar.set(Calendar.MILLISECOND, 999);
			break;
		case DAY_OF_MONTH:
			calendar.set(Calendar.YEAR, Integer.parseInt(year));
			calendar.set(Calendar.MONTH, Integer.parseInt(month) - 1);
			calendar.set(Calendar.DAY_OF_MONTH, Integer.parseInt(day));
			calendar.set(Calendar.HOUR_OF_DAY, 23);
			calendar.set(Calendar.MINUTE, 59);
			calendar.set(Calendar.SECOND, 59);
			calendar.set(Calendar.MILLISECOND, 999);
			break;
		case MONTH:
			calendar.set(Calendar.YEAR, Integer.parseInt(year));
			calendar.set(Calendar.MONTH, Integer.parseInt(month) - 1);

			int m = Integer.parseInt(month);
			if (m == 1 || m == 3 || m == 5 || m == 7 || m == 8 || m == 10 || m == 12)
				calendar.set(Calendar.DAY_OF_MONTH, 31);
			else if (m == 2) {
				int yr = calendar.get(Calendar.YEAR);
				if (yr % 4 == 0)
					calendar.set(Calendar.DAY_OF_MONTH, 29);
				else
					calendar.set(Calendar.DAY_OF_MONTH, 28);
			} else
				calendar.set(Calendar.DAY_OF_MONTH, 30);

			calendar.set(Calendar.HOUR_OF_DAY, 23);
			calendar.set(Calendar.MINUTE, 59);
			calendar.set(Calendar.SECOND, 59);
			calendar.set(Calendar.MILLISECOND, 999);
			break;
		case YEAR:
			calendar.set(Calendar.YEAR, Integer.parseInt(year));
			calendar.set(Calendar.MONTH, 11);
			calendar.set(Calendar.DAY_OF_MONTH, 31);
			calendar.set(Calendar.HOUR_OF_DAY, 23);
			calendar.set(Calendar.MINUTE, 59);
			calendar.set(Calendar.SECOND, 59);
			calendar.set(Calendar.MILLISECOND, 999);
			break;

		}

		// System.out.println(calendar.getTime());

		// System.out.println("CONVERTED:" + d1);

		// System.out.println(calendar.getTimeInMillis() + offset);
		return calendar.getTimeInMillis();

	}

	/**
	 * 
	 * @author sapmitra
	 * @param year
	 * @param month
	 * @param day
	 * @param hour
	 * @param temporalType
	 * @return
	 */
	public static long getEndTimeStamp_BackupWithOffset(String year, String month, String day, String hour,
			TemporalType temporalType) {

		Calendar calendar = Calendar.getInstance(TemporalHash.TIMEZONE);
		calendar.setTimeZone(TemporalHash.TIMEZONE);
		// calendar.setLenient(false);

		if (month.contains("x"))
			month = "11";

		int mn = Integer.parseInt(month);

		if (day.contains("x")) {
			if (mn == 1 || mn == 3 || mn == 5 || mn == 7 || mn == 8 || mn == 10 || mn == 12)
				day = "31";
			else if (mn == 2) {
				int yr = Integer.parseInt(year);
				if (yr % 4 == 0)
					day = "29";
				else
					day = "28";
			}
		}

		if (hour.contains("x"))
			hour = "23";

		switch (temporalType) {
		case HOUR_OF_DAY:
			calendar.set(Calendar.YEAR, Integer.parseInt(year));
			calendar.set(Calendar.MONTH, Integer.parseInt(month) - 1);
			calendar.set(Calendar.DAY_OF_MONTH, Integer.parseInt(day));
			calendar.set(Calendar.HOUR_OF_DAY, Integer.parseInt(hour));
			calendar.set(Calendar.MINUTE, 59);
			calendar.set(Calendar.SECOND, 59);
			calendar.set(Calendar.MILLISECOND, 999);
			break;
		case DAY_OF_MONTH:
			calendar.set(Calendar.YEAR, Integer.parseInt(year));
			calendar.set(Calendar.MONTH, Integer.parseInt(month) - 1);
			calendar.set(Calendar.DAY_OF_MONTH, Integer.parseInt(day));
			calendar.set(Calendar.HOUR_OF_DAY, 23);
			calendar.set(Calendar.MINUTE, 59);
			calendar.set(Calendar.SECOND, 59);
			calendar.set(Calendar.MILLISECOND, 999);
			break;
		case MONTH:
			calendar.set(Calendar.YEAR, Integer.parseInt(year));
			calendar.set(Calendar.MONTH, Integer.parseInt(month) - 1);

			int m = Integer.parseInt(month);
			if (m == 1 || m == 3 || m == 5 || m == 7 || m == 8 || m == 10 || m == 12)
				calendar.set(Calendar.DAY_OF_MONTH, 31);
			else if (m == 2) {
				int yr = calendar.get(Calendar.YEAR);
				if (yr % 4 == 0)
					calendar.set(Calendar.DAY_OF_MONTH, 29);
				else
					calendar.set(Calendar.DAY_OF_MONTH, 28);
			} else
				calendar.set(Calendar.DAY_OF_MONTH, 30);

			calendar.set(Calendar.HOUR_OF_DAY, 23);
			calendar.set(Calendar.MINUTE, 59);
			calendar.set(Calendar.SECOND, 59);
			calendar.set(Calendar.MILLISECOND, 999);
			break;
		case YEAR:
			calendar.set(Calendar.YEAR, Integer.parseInt(year));
			calendar.set(Calendar.MONTH, 11);
			calendar.set(Calendar.DAY_OF_MONTH, 31);
			calendar.set(Calendar.HOUR_OF_DAY, 23);
			calendar.set(Calendar.MINUTE, 59);
			calendar.set(Calendar.SECOND, 59);
			calendar.set(Calendar.MILLISECOND, 999);
			break;

		}

		// System.out.println(calendar.getTime());

		long offset = TemporalHash.TIMEZONE.getRawOffset() - TimeZone.getDefault().getRawOffset();

		Date d1 = new Date(calendar.getTimeInMillis() + offset);
		// System.out.println("CONVERTED:" + d1);

		// System.out.println(calendar.getTimeInMillis() + offset);
		return calendar.getTimeInMillis() + offset;

	}

	/**
	 * Encode a set of {@link Coordinates} into a GeoHash string.
	 *
	 * @param coords    Coordinates to get GeoHash for.
	 *
	 * @param precision Desired number of characters in the returned GeoHash String.
	 *                  More characters means more precision.
	 *
	 * @return GeoHash string.
	 */
	public static String encode(Coordinates coords, int precision) {
		return encode(coords.getLatitude(), coords.getLongitude(), precision);
	}

	/**
	 * Encode {@link SpatialRange} into a GeoHash string.
	 *
	 * @param range     SpatialRange to get GeoHash for.
	 *
	 * @param precision Number of characters in the returned GeoHash String. More
	 *                  characters is more precise.
	 *
	 * @return GeoHash string.
	 */
	public static String encode(SpatialRange range, int precision) {
		Coordinates rangeCoords = range.getCenterPoint();
		return encode(rangeCoords.getLatitude(), rangeCoords.getLongitude(), precision);
	}

	/**
	 * Encode latitude and longitude into a GeoHash string.
	 *
	 * @param latitude  Latitude coordinate, in degrees.
	 *
	 * @param longitude Longitude coordinate, in degrees.
	 *
	 * @param precision Number of characters in the returned GeoHash String. More
	 *                  characters is more precise.
	 *
	 * @return resulting GeoHash String.
	 */
	public static String encode(float latitude, float longitude, int precision) {
		while (latitude < -90f || latitude > 90f)
			latitude = latitude < -90f ? 180.0f + latitude : latitude > 90f ? -180f + latitude : latitude;
		while (longitude < -180f || longitude > 180f)
			longitude = longitude < -180f ? 360.0f + longitude : longitude > 180f ? -360f + longitude : longitude;
		/*
		 * Set up 2-element arrays for longitude and latitude that we can flip between
		 * while encoding
		 */
		float[] high = new float[2];
		float[] low = new float[2];
		float[] value = new float[2];

		high[0] = LONGITUDE_RANGE;
		high[1] = LATITUDE_RANGE;
		low[0] = -LONGITUDE_RANGE;
		low[1] = -LATITUDE_RANGE;
		value[0] = longitude;
		value[1] = latitude;

		String hash = "";

		for (int p = 0; p < precision; ++p) {

			float middle = 0.0f;
			int charBits = 0;
			for (int b = 0; b < BITS_PER_CHAR; ++b) {
				int bit = (p * BITS_PER_CHAR) + b;

				charBits <<= 1;

				middle = (high[bit % 2] + low[bit % 2]) / 2;
				if (value[bit % 2] > middle) {
					charBits |= 1;
					low[bit % 2] = middle;
				} else {
					high[bit % 2] = middle;
				}
			}

			hash += charMap[charBits];
		}

		return hash;
	}

	/**
	 * Convert a GeoHash String to a long integer.
	 *
	 * @param hash GeoHash String to convert.
	 *
	 * @return The GeoHash as a long integer.
	 */
	public static long hashToLong(String hash) {
		long longForm = 0;

		/* Long can fit 12 GeoHash characters worth of precision. */
		if (hash.length() > 12) {
			hash = hash.substring(0, 12);
		}

		for (char c : hash.toCharArray()) {
			longForm <<= BITS_PER_CHAR;
			longForm |= charLookupTable.get(c);
		}

		return longForm;
	}

	public static String fromLongToString(long hash, int precision) {
		String retStr = "";
		String binaryString = Long.toBinaryString(hash);

		if (binaryString.length() < precision * BITS_PER_CHAR) {
			int ln = precision * BITS_PER_CHAR - binaryString.length();

			for (int i = 0; i < ln; i++) {
				binaryString = "0" + binaryString;
			}
		}

		for (int i = 0; i < binaryString.length(); i = i + 5) {
			String sb = binaryString.substring(i, i + 5);
			int foo = Integer.parseInt(sb, 2);

			retStr += BASE32.charAt(foo);
			// System.out.println(BASE32.charAt(foo));
		}

		return retStr;
	}

	/**
	 * Decode a GeoHash to an approximate bounding box that contains the original
	 * GeoHashed point.
	 *
	 * @param geoHash GeoHash string
	 *
	 * @return Spatial Range (bounding box) of the GeoHash.
	 */
	public static SpatialRange decodeHash(String geoHash) {
		ArrayList<Boolean> bits = getBits(geoHash);

		float[] longitude = decodeBits(bits, false);
		float[] latitude = decodeBits(bits, true);

		return new SpatialRange(latitude[0], latitude[1], longitude[0], longitude[1]);
	}

	/**
	 * @param geohash   - geohash of the region for which the neighbors are needed
	 * @param direction - one of nw, n, ne, w, e, sw, s, se
	 * @return
	 */
	public static String getNeighbour(String geohash, String direction) {
		if (geohash == null || geohash.trim().length() == 0)
			throw new IllegalArgumentException("Invalid Geohash");
		geohash = geohash.trim();
		int precision = geohash.length();
		SpatialRange boundingBox = decodeHash(geohash);
		Coordinates centroid = boundingBox.getCenterPoint();
		float widthDiff = boundingBox.getUpperBoundForLongitude() - centroid.getLongitude();
		float heightDiff = boundingBox.getUpperBoundForLatitude() - centroid.getLatitude();
		switch (direction) {
		case "nw":
			return encode(boundingBox.getUpperBoundForLatitude() + heightDiff,
					boundingBox.getLowerBoundForLongitude() - widthDiff, precision);
		case "n":
			return encode(boundingBox.getUpperBoundForLatitude() + heightDiff, centroid.getLongitude(), precision);
		case "ne":
			return encode(boundingBox.getUpperBoundForLatitude() + heightDiff,
					boundingBox.getUpperBoundForLongitude() + widthDiff, precision);
		case "w":
			return encode(centroid.getLatitude(), boundingBox.getLowerBoundForLongitude() - widthDiff, precision);
		case "e":
			return encode(centroid.getLatitude(), boundingBox.getUpperBoundForLongitude() + widthDiff, precision);
		case "sw":
			return encode(boundingBox.getLowerBoundForLatitude() - heightDiff,
					boundingBox.getLowerBoundForLongitude() - widthDiff, precision);
		case "s":
			return encode(boundingBox.getLowerBoundForLatitude() - heightDiff, centroid.getLongitude(), precision);
		case "se":
			return encode(boundingBox.getLowerBoundForLatitude() - heightDiff,
					boundingBox.getUpperBoundForLongitude() + widthDiff, precision);
		default:
			return "";
		}
	}

	/**
	 * 
	 * Neighbors are returned in the order nw,n,ne,w,e,sw,s,se
	 * 
	 * @param geoHash
	 * @return
	 */
	public static String[] getNeighbours(String geoHash) {
		String[] neighbors = new String[8];
		if (geoHash == null || geoHash.trim().length() == 0)
			throw new IllegalArgumentException("Invalid Geohash");
		geoHash = geoHash.trim();
		int precision = geoHash.length();
		SpatialRange boundingBox = decodeHash(geoHash);
		Coordinates centroid = boundingBox.getCenterPoint();
		float widthDiff = boundingBox.getUpperBoundForLongitude() - centroid.getLongitude();
		float heightDiff = boundingBox.getUpperBoundForLatitude() - centroid.getLatitude();
		neighbors[0] = encode(boundingBox.getUpperBoundForLatitude() + heightDiff,
				boundingBox.getLowerBoundForLongitude() - widthDiff, precision);
		neighbors[1] = encode(boundingBox.getUpperBoundForLatitude() + heightDiff, centroid.getLongitude(), precision);
		neighbors[2] = encode(boundingBox.getUpperBoundForLatitude() + heightDiff,
				boundingBox.getUpperBoundForLongitude() + widthDiff, precision);
		neighbors[3] = encode(centroid.getLatitude(), boundingBox.getLowerBoundForLongitude() - widthDiff, precision);
		neighbors[4] = encode(centroid.getLatitude(), boundingBox.getUpperBoundForLongitude() + widthDiff, precision);
		neighbors[5] = encode(boundingBox.getLowerBoundForLatitude() - heightDiff,
				boundingBox.getLowerBoundForLongitude() - widthDiff, precision);
		neighbors[6] = encode(boundingBox.getLowerBoundForLatitude() - heightDiff, centroid.getLongitude(), precision);
		neighbors[7] = encode(boundingBox.getLowerBoundForLatitude() - heightDiff,
				boundingBox.getUpperBoundForLongitude() + widthDiff, precision);
		return neighbors;
	}

	/**
	 * @param coordinates - latitude and longitude values
	 * @return Point - x, y pair obtained from a geohash precision of 12. x,y values
	 *         range from [0, 4096)
	 */
	public static Point<Integer> coordinatesToXY(Coordinates coords) {
		int width = 1 << MAX_PRECISION;
		float xDiff = coords.getLongitude() + 180;
		float yDiff = 90 - coords.getLatitude();
		int x = (int) (xDiff * width / 360);
		int y = (int) (yDiff * width / 180);
		return new Point<>(x, y);
	}

	public static Coordinates xyToCoordinates(int x, int y) {
		int width = 1 << MAX_PRECISION;
		return new Coordinates(90 - y * 180f / width, x * 360f / width - 180f);
	}

	/**
	 * 
	 * @param polygon bounding polygon
	 * @return
	 */
	public static String[] getIntersectingGeohashes(List<Coordinates> polygon) {
		Set<String> hashes = new HashSet<String>();
		Polygon geometry = new Polygon();

		// Getting MBR out of the POLYGON in geometry object
		for (Coordinates coords : polygon) {
			Point<Integer> point = coordinatesToXY(coords);
			geometry.addPoint(point.X(), point.Y());
		}
		// center may not lie inside polygon so start with any vertex of the
		// polygon
		Coordinates spatialCenter = polygon.get(0);

		// GETS MBR WITH XY COORDINATES FOR THE FULL POLYGON
		Rectangle2D box = geometry.getBounds2D();

		// Take any point(0) from the polygon and convert it to a 2 character geohash

		String geohash = encode(spatialCenter, Partitioner.SPATIAL_PRECISION);
		Queue<String> hashQue = new LinkedList<String>();
		Set<String> computedHashes = new HashSet<String>();

		/* Start by inserting this initial geohash for point0 into the queue */
		hashQue.offer(geohash);

		while (!hashQue.isEmpty()) {
			String hash = hashQue.poll();
			computedHashes.add(hash);
			/* Returns the bounds for this geohash in terms of latlong */
			SpatialRange hashRange = decodeHash(hash);
			/* Get the top left and bottom right points as coordinates */
			Pair<Coordinates, Coordinates> coordsPair = hashRange.get2DCoordinates();
			Point<Integer> upLeft = coordinatesToXY(coordsPair.a);
			Point<Integer> lowRight = coordinatesToXY(coordsPair.b);
			Rectangle2D hashRect = new Rectangle(upLeft.X(), upLeft.Y(), lowRight.X() - upLeft.X(),
					lowRight.Y() - upLeft.Y());
			/* If the geohash fully encloses the MBR */
			/* box corresponds to the entire polygon */
			if (hash.equals(geohash) && hashRect.contains(box)) {
				hashes.add(hash);
				break;
			}
			/* If it does not fully enclose */
			if (geometry.intersects(hashRect)) {
				hashes.add(hash);
				String[] neighbors = getNeighbours(hash);
				for (String neighbour : neighbors)
					if (!computedHashes.contains(neighbour) && !hashQue.contains(neighbour))
						hashQue.offer(neighbour);
			}
		}
		return hashes.size() > 0 ? hashes.toArray(new String[hashes.size()]) : new String[] {};
	}

	/**
	 * Gives the list of geohashes of a required precision that intersect with a
	 * given polygon Use this if you are sure your polygon is a convex polygon
	 * 
	 * @author sapmitra
	 * @param polygon
	 * @param precision
	 * @return
	 */
	public static String[] getIntersectingGeohashesForConvexBoundingPolygon(List<Coordinates> polygon, int precision) {
		Set<String> hashes = new HashSet<String>();
		Polygon geometry = new Polygon();
		for (Coordinates coords : polygon) {
			Point<Integer> point = coordinatesToXY(coords);
			geometry.addPoint(point.X(), point.Y());
		}
		Coordinates spatialCenter = new SpatialRange(polygon).getCenterPoint();
		Rectangle2D box = geometry.getBounds2D();
		String geohash = encode(spatialCenter, precision);
		Queue<String> hashQue = new LinkedList<String>();
		Set<String> computedHashes = new HashSet<String>();
		hashQue.offer(geohash);
		while (!hashQue.isEmpty()) {
			String hash = hashQue.poll();
			computedHashes.add(hash);
			SpatialRange hashRange = decodeHash(hash);
			Pair<Coordinates, Coordinates> coordsPair = hashRange.get2DCoordinates();
			Point<Integer> upLeft = coordinatesToXY(coordsPair.a);
			Point<Integer> lowRight = coordinatesToXY(coordsPair.b);
			Rectangle2D hashRect = new Rectangle(upLeft.X(), upLeft.Y(), lowRight.X() - upLeft.X(),
					lowRight.Y() - upLeft.Y());
			if (hash.equals(geohash) && hashRect.contains(box)) {
				hashes.add(hash);
				break;
			}
			if (geometry.intersects(hashRect)) {
				hashes.add(hash);
				String[] neighbors = getNeighbours(hash);
				for (String neighbour : neighbors)
					if (!computedHashes.contains(neighbour) && !hashQue.contains(neighbour))
						hashQue.offer(neighbour);
			}
		}
		return hashes.size() > 0 ? hashes.toArray(new String[hashes.size()]) : new String[] {};
	}

	/* Same as the previous one. This only lets you add a precision */
	public static String[] getIntersectingGeohashes(List<Coordinates> polygon, int userPrecision) {
		Set<String> hashes = new HashSet<String>();
		Polygon geometry = new Polygon();
		for (Coordinates coords : polygon) {
			Point<Integer> point = coordinatesToXY(coords);
			geometry.addPoint(point.X(), point.Y());
		}
		// center may not lie inside polygon so start with any vertex of the
		// polygon
		Coordinates spatialCenter = polygon.get(0);

		// GETS MBR WITH XY COORDINATES
		Rectangle2D box = geometry.getBounds2D();

		// Take any point(0) from the polygon and convert it to a 2 character geohash

		String geohash = encode(spatialCenter, userPrecision);
		Queue<String> hashQue = new LinkedList<String>();
		Set<String> computedHashes = new HashSet<String>();
		hashQue.offer(geohash);
		while (!hashQue.isEmpty()) {
			String hash = hashQue.poll();
			computedHashes.add(hash);
			/* Returns the bounds for this geohash in terms of latlong */
			SpatialRange hashRange = decodeHash(hash);
			/* Get the top left and bottom right points as coordinates */
			Pair<Coordinates, Coordinates> coordsPair = hashRange.get2DCoordinates();
			Point<Integer> upLeft = coordinatesToXY(coordsPair.a);
			Point<Integer> lowRight = coordinatesToXY(coordsPair.b);
			Rectangle2D hashRect = new Rectangle(upLeft.X(), upLeft.Y(), lowRight.X() - upLeft.X(),
					lowRight.Y() - upLeft.Y());
			/* If the geohash fully encloses the MBR */
			if (hash.equals(geohash) && hashRect.contains(box)) {
				hashes.add(hash);
				break;
			}
			/* If it does not fully enclose */
			if (geometry.intersects(hashRect)) {
				hashes.add(hash);
				String[] neighbors = getNeighbours(hash);
				for (String neighbour : neighbors)
					if (!computedHashes.contains(neighbour) && !hashQue.contains(neighbour))
						hashQue.offer(neighbour);
			}
		}
		return hashes.size() > 0 ? hashes.toArray(new String[hashes.size()]) : new String[] {};
	}

	/**
	 * 
	 * @author sapmitra
	 * @param polygon
	 * @param flank
	 * @return
	 */
	public static boolean checkIntersection(List<Coordinates> polygon, List<Coordinates> flank) {
		Polygon geometry = new Polygon();
		for (Coordinates coords : polygon) {
			Point<Integer> point = coordinatesToXY(coords);
			geometry.addPoint(point.X(), point.Y());
		}

		Polygon geometryFlank = new Polygon();
		for (Coordinates coords : flank) {
			Point<Integer> point = coordinatesToXY(coords);
			geometryFlank.addPoint(point.X(), point.Y());
		}

		Area a1 = new Area(geometry);
		Area a2 = new Area(geometryFlank);

		a1.intersect(a2);

		if (a1.isEmpty()) {
			return false;
		} else {
			return true;
		}

	}

	/**
	 * returns true if the flank lies completely inside polygon
	 * 
	 * @author sapmitra
	 * @param polygon
	 * @param flank
	 * @return
	 */
	public static boolean checkEnclosure(List<Coordinates> polygon, List<Coordinates> flank) {

		Polygon geometry = new Polygon();
		for (Coordinates coords : polygon) {
			Point<Integer> point = coordinatesToXY(coords);
			geometry.addPoint(point.X(), point.Y());
		}

		Polygon geometrySmaller = new Polygon();
		for (Coordinates coords : flank) {
			Point<Integer> point = coordinatesToXY(coords);
			geometrySmaller.addPoint(point.X(), point.Y());
		}
		Area aPrev = new Area(geometrySmaller);
		Area a1 = new Area(geometry);
		Area a2 = new Area(geometrySmaller);

		a1.intersect(a2);

		if (a1.equals(aPrev)) {

			return true;
		} else {
			return false;
		}

	}

	/**
	 * returns true if the geohash2 lies completely inside geohash1
	 * 
	 * @author sapmitra
	 * @param geohash1 bigger geohash
	 * @param geohash2 smaller geohash
	 * @return
	 */
	public static boolean checkEnclosure(String geohash1, String geohash2) {

		if (geohash1.equals(geohash2))
			return true;

		if (geohash2.length() < geohash1.length()) {
			return false;
		}

		String g2_dash = geohash2.substring(0, geohash1.length());

		if (geohash1.equals(g2_dash)) {
			return true;
		} else {
			return false;
		}

	}

	/**
	 * check if two geohashes overlap
	 * 
	 * @author sapmitra
	 * @param geohash1
	 * @param geohash2
	 * @return
	 */
	public static boolean checkIntersection(String geohash1, String geohash2) {

		SpatialRange range1 = decodeHash(geohash1);

		Coordinates c1 = new Coordinates(range1.getLowerBoundForLatitude(), range1.getLowerBoundForLongitude());
		Coordinates c2 = new Coordinates(range1.getUpperBoundForLatitude(), range1.getLowerBoundForLongitude());
		Coordinates c3 = new Coordinates(range1.getUpperBoundForLatitude(), range1.getUpperBoundForLongitude());
		Coordinates c4 = new Coordinates(range1.getLowerBoundForLatitude(), range1.getUpperBoundForLongitude());

		ArrayList<Coordinates> cs1 = new ArrayList<Coordinates>();
		cs1.add(c1);
		cs1.add(c2);
		cs1.add(c3);
		cs1.add(c4);

		SpatialRange range2 = decodeHash(geohash2);

		Coordinates c12 = new Coordinates(range2.getLowerBoundForLatitude(), range2.getLowerBoundForLongitude());
		Coordinates c22 = new Coordinates(range2.getUpperBoundForLatitude(), range2.getLowerBoundForLongitude());
		Coordinates c32 = new Coordinates(range2.getUpperBoundForLatitude(), range2.getUpperBoundForLongitude());
		Coordinates c42 = new Coordinates(range2.getLowerBoundForLatitude(), range2.getUpperBoundForLongitude());

		ArrayList<Coordinates> cs2 = new ArrayList<Coordinates>();
		cs2.add(c12);
		cs2.add(c22);
		cs2.add(c32);
		cs2.add(c42);

		return checkIntersection(cs1, cs2);
	}

	/**
	 * CHECK IF A GEOHASH IS ENCLOSED IN A POLYGON
	 * 
	 * @author sapmitra
	 * @param geohash
	 * @param polygon
	 * @return
	 */

	public static boolean checkEnclosure(List<Coordinates> polygon, String geohash) {

		SpatialRange range = decodeHash(geohash);

		Coordinates c1 = new Coordinates(range.getLowerBoundForLatitude(), range.getLowerBoundForLongitude());
		Coordinates c2 = new Coordinates(range.getUpperBoundForLatitude(), range.getLowerBoundForLongitude());
		Coordinates c3 = new Coordinates(range.getUpperBoundForLatitude(), range.getUpperBoundForLongitude());
		Coordinates c4 = new Coordinates(range.getLowerBoundForLatitude(), range.getUpperBoundForLongitude());

		ArrayList<Coordinates> cs1 = new ArrayList<Coordinates>();
		cs1.add(c1);
		cs1.add(c2);
		cs1.add(c3);
		cs1.add(c4);

		return checkEnclosure(polygon, cs1);
	}

	/**
	 * CHECK IF A GEOHASH INTERSECTS A POLYGON
	 * 
	 * @author sapmitra
	 * @param geohash
	 * @param polygon
	 * @return
	 */

	public static boolean checkIntersection(List<Coordinates> polygon, String geohash) {

		SpatialRange range = decodeHash(geohash);

		Coordinates c1 = new Coordinates(range.getLowerBoundForLatitude(), range.getLowerBoundForLongitude());
		Coordinates c2 = new Coordinates(range.getUpperBoundForLatitude(), range.getLowerBoundForLongitude());
		Coordinates c3 = new Coordinates(range.getUpperBoundForLatitude(), range.getUpperBoundForLongitude());
		Coordinates c4 = new Coordinates(range.getLowerBoundForLatitude(), range.getUpperBoundForLongitude());

		ArrayList<Coordinates> cs1 = new ArrayList<Coordinates>();
		cs1.add(c1);
		cs1.add(c2);
		cs1.add(c3);
		cs1.add(c4);

		return checkIntersection(polygon, cs1);
	}

	public static void main2(String arg[]) {

		Coordinates c11 = new Coordinates(50.18f, -103.29f);
		Coordinates c21 = new Coordinates(36.78f, -117.88f);
		Coordinates c31 = new Coordinates(36.85f, -97.21f);
		List<Coordinates> cl1 = new ArrayList<Coordinates>();
		cl1.add(c11);
		cl1.add(c21);
		cl1.add(c31);

		SpatialRange range = decodeHash("9xn");

		Coordinates c1 = new Coordinates(range.getLowerBoundForLatitude(), range.getLowerBoundForLongitude());
		Coordinates c2 = new Coordinates(range.getUpperBoundForLatitude(), range.getLowerBoundForLongitude());
		Coordinates c3 = new Coordinates(range.getUpperBoundForLatitude(), range.getUpperBoundForLongitude());
		Coordinates c4 = new Coordinates(range.getLowerBoundForLatitude(), range.getUpperBoundForLongitude());

		List<Coordinates> cl = new ArrayList<Coordinates>();
		cl.add(c1);
		cl.add(c2);
		cl.add(c3);
		cl.add(c4);

		// String[] intersectingGeohashes = getIntersectingGeohashes(cl, 6);
		System.out.println(checkIntersection(cl1, cl));
		/*
		 * for(String s: intersectingGeohashes) { if(s.startsWith("9x"))
		 * System.out.print(s+","); }
		 */

		// getSuperCubeGeohashBounds("sp", null, null, null, null, 3);

		// GeoHash.getBorderingGeoHashes("9r", 3, 1);

	}

	/**
	 * Decode GeoHash bits from a binary GeoHash.
	 *
	 * @param bits     ArrayList of Booleans containing the GeoHash bits
	 *
	 * @param latitude If set to <code>true</code> the latitude bits are decoded. If
	 *                 set to <code>false</code> the longitude bits are decoded.
	 *
	 * @return low, high range that the GeoHashed location falls between.
	 */
	private static float[] decodeBits(ArrayList<Boolean> bits, boolean latitude) {
		float low, high, middle;
		int offset;

		if (latitude) {
			offset = 1;
			low = -90.0f;
			high = 90.0f;
		} else {
			offset = 0;
			low = -180.0f;
			high = 180.0f;
		}

		for (int i = offset; i < bits.size(); i += 2) {
			middle = (high + low) / 2;

			if (bits.get(i)) {
				low = middle;
			} else {
				high = middle;
			}
		}

		if (latitude) {
			return new float[] { low, high };
		} else {
			return new float[] { low, high };
		}
	}

	/**
	 * Converts a GeoHash string to its binary representation.
	 *
	 * @param hash GeoHash string to convert to binary
	 *
	 * @return The GeoHash in binary form, as an ArrayList of Booleans.
	 */
	private static ArrayList<Boolean> getBits(String hash) {
		hash = hash.toLowerCase();

		/* Create an array of bits, 5 bits per character: */
		ArrayList<Boolean> bits = new ArrayList<Boolean>(hash.length() * BITS_PER_CHAR);

		/* Loop through the hash string, setting appropriate bits. */
		for (int i = 0; i < hash.length(); ++i) {
			int charValue = charLookupTable.get(hash.charAt(i));

			/* Set bit from charValue, then shift over to the next bit. */
			for (int j = 0; j < BITS_PER_CHAR; ++j, charValue <<= 1) {
				bits.add((charValue & 0x10) == 0x10);
			}
		}
		return bits;
	}

	public static Polygon buildAwtPolygon(List<Coordinates> geometry) {
		Polygon polygon = new Polygon();
		for (Coordinates coords : geometry) {
			Point<Integer> point = coordinatesToXY(coords);
			polygon.addPoint(point.X(), point.Y());
		}
		return polygon;
	}

	/**
	 * 
	 * @param polygon
	 * @param gh
	 * @param bitPrecision  geohash precision in bits instead of characters of the
	 *                      filesystem (fixed)
	 * @param intersections
	 */
	public static void getGeohashPrefixes(Polygon polygon, GeoHash gh, int bitPrecision, Set<GeoHash> intersections) {
		if (gh.getPrecision() >= bitPrecision) {
			intersections.add(gh);
		} else {
			if (polygon.contains(gh.getRectangle())) {
				intersections.add(gh);
			} else {
				GeoHash leftGH = new GeoHash(gh.getBinaryHash() + "0");
				GeoHash rightGH = new GeoHash(gh.getBinaryHash() + "1");
				if (polygon.intersects(leftGH.getRectangle()))
					getGeohashPrefixes(polygon, leftGH, bitPrecision, intersections);
				if (polygon.intersects(rightGH.getRectangle()))
					getGeohashPrefixes(polygon, rightGH, bitPrecision, intersections);
			}
		}
	}

	/**
	 * find out what the valid neighbors to a geohash are with respect to a query
	 * polygon
	 * 
	 * @author sapmitra
	 * @param queryPolygon
	 * @param precision
	 * @param cGeoString   a------b . . . . d------c
	 * @param blocks
	 * @param borderMap
	 * @return
	 */
	public static String[] checkForNeighborValidity(List<Coordinates> polygon, int precision, String cGeoString,
			Map<String, BorderingProperties> borderMap, List<String> blocks) {
		List<String> ignore = new ArrayList<String>();

		BorderingProperties bp = getBorderingGeohashHeuristic(cGeoString, precision, 0, null, null);

		String ne = bp.getNe();

		String nw = bp.getNw();

		String se = bp.getSe();

		String sw = bp.getSw();

		// System.out.println(nw);

		SpatialRange nwBounds = decodeHash(nw);
		SpatialRange neBounds = decodeHash(ne);
		SpatialRange seBounds = decodeHash(se);
		SpatialRange swBounds = decodeHash(sw);

		/* Getting North West Flank */
		Coordinates caNW = new Coordinates(nwBounds.getUpperBoundForLatitude(), nwBounds.getLowerBoundForLongitude());
		Coordinates cbNW = new Coordinates(nwBounds.getUpperBoundForLatitude(), nwBounds.getUpperBoundForLongitude());
		Coordinates ccNW = new Coordinates(nwBounds.getLowerBoundForLatitude(), nwBounds.getUpperBoundForLongitude());
		Coordinates cdNW = new Coordinates(nwBounds.getLowerBoundForLatitude(), nwBounds.getLowerBoundForLongitude());

		List<Coordinates> nwFlank = Arrays.asList(caNW, cbNW, ccNW, cdNW);

		// System.out.println(nwFlank);

		/* Getting North East Flank */
		Coordinates caNE = new Coordinates(neBounds.getUpperBoundForLatitude(), neBounds.getLowerBoundForLongitude());
		Coordinates cbNE = new Coordinates(neBounds.getUpperBoundForLatitude(), neBounds.getUpperBoundForLongitude());
		Coordinates ccNE = new Coordinates(neBounds.getLowerBoundForLatitude(), neBounds.getUpperBoundForLongitude());
		Coordinates cdNE = new Coordinates(neBounds.getLowerBoundForLatitude(), neBounds.getLowerBoundForLongitude());

		List<Coordinates> neFlank = Arrays.asList(caNE, cbNE, ccNE, cdNE);

		// System.out.println(neFlank);

		/* Getting South West Flank */
		Coordinates caSW = new Coordinates(swBounds.getUpperBoundForLatitude(), swBounds.getLowerBoundForLongitude());
		Coordinates cbSW = new Coordinates(swBounds.getUpperBoundForLatitude(), swBounds.getUpperBoundForLongitude());
		Coordinates ccSW = new Coordinates(swBounds.getLowerBoundForLatitude(), swBounds.getUpperBoundForLongitude());
		Coordinates cdSW = new Coordinates(swBounds.getLowerBoundForLatitude(), swBounds.getLowerBoundForLongitude());

		List<Coordinates> swFlank = Arrays.asList(caSW, cbSW, ccSW, cdSW);

		// System.out.println(swFlank);

		/* Getting South East Flank */
		Coordinates caSE = new Coordinates(seBounds.getUpperBoundForLatitude(), seBounds.getLowerBoundForLongitude());
		Coordinates cbSE = new Coordinates(seBounds.getUpperBoundForLatitude(), seBounds.getUpperBoundForLongitude());
		Coordinates ccSE = new Coordinates(seBounds.getLowerBoundForLatitude(), seBounds.getUpperBoundForLongitude());
		Coordinates cdSE = new Coordinates(seBounds.getLowerBoundForLatitude(), seBounds.getLowerBoundForLongitude());

		List<Coordinates> seFlank = Arrays.asList(caSE, cbSE, ccSE, cdSE);

		// System.out.println(seFlank);

		/* Getting North,South,East,West Flank */

		List<Coordinates> northFlank = Arrays.asList(caNW, cbNE, ccNE, cdNW);
		List<Coordinates> southFlank = Arrays.asList(caSW, cbSE, ccSE, cdSW);
		List<Coordinates> eastFlank = Arrays.asList(caNE, cbNE, ccSE, cdSE);
		List<Coordinates> westFlank = Arrays.asList(caNW, cbNW, ccSW, cdSW);

		boolean checkN = checkIntersection(polygon, northFlank);
		boolean checkS = checkIntersection(polygon, southFlank);
		boolean checkE = checkIntersection(polygon, eastFlank);
		boolean checkW = checkIntersection(polygon, westFlank);
		boolean checkNE = checkIntersection(polygon, neFlank);
		boolean checkNW = checkIntersection(polygon, nwFlank);
		boolean checkSE = checkIntersection(polygon, seFlank);
		boolean checkSW = checkIntersection(polygon, swFlank);

		/* Checking for North */
		if (!checkN) {

			ignore.addAll(Arrays.asList("n", "nw", "ne"));

		} else if (!checkNE || !checkNW) {

			if (!checkNE) {
				ignore.add("ne");
			}
			if (!checkNW) {
				ignore.add("nw");
			}
		}

		/* Checking for South */
		if (!checkS) {

			ignore.addAll(Arrays.asList("s", "sw", "se"));

		} else if (!checkSE || !checkSW) {

			if (!checkSE) {
				ignore.add("se");
			}
			if (!checkSW) {
				ignore.add("sw");
			}
		}

		/* Checking for East */
		if (!checkE) {

			ignore.addAll(Arrays.asList("e", "ne", "se"));

		} else if (!checkSE || !checkNE) {

			if (!checkNE) {
				ignore.add("ne");
			}
			if (!checkSE) {
				ignore.add("se");
			}
		}

		/* Checking for West */
		if (!checkW) {

			ignore.addAll(Arrays.asList("w", "nw", "sw"));

		} else if (!checkSE || !checkNW) {

			if (!checkNW) {
				ignore.add("nw");
			}
			if (!checkSW) {
				ignore.add("sw");
			}
		}

		/* Looking at bordering properties */
		/* In bordering properties, during storage, N & nw are disjoint areas */

		List<String> ignoreBorder = new ArrayList<String>(Arrays.asList("n", "s", "e", "w", "ne", "nw", "se", "sw"));

		for (String block : blocks) {

			if (ignoreBorder.size() > 0) {

				BorderingProperties bpr = borderMap.get(block);

				if (ignoreBorder.contains("n")) {
					if (bpr.getNorthEntries().size() > 0) {
						ignoreBorder.remove("n");
					}
				}
				if (ignoreBorder.contains("s")) {
					if (bpr.getSouthEntries().size() > 0) {
						ignoreBorder.remove("s");
					}
				}
				if (ignoreBorder.contains("e")) {
					if (bpr.getEastEntries().size() > 0) {
						ignoreBorder.remove("e");
					}
				}
				if (ignoreBorder.contains("w")) {
					if (bpr.getWestEntries().size() > 0) {
						ignoreBorder.remove("w");
					}
				}
				if (ignoreBorder.contains("ne")) {
					if (bpr.getNeEntries().size() > 0) {
						ignoreBorder.remove("ne");
					}
				}
				if (ignoreBorder.contains("nw")) {
					if (bpr.getNwEntries().size() > 0) {
						ignoreBorder.remove("nw");
					}
				}
				if (ignoreBorder.contains("se")) {
					if (bpr.getSeEntries().size() > 0) {
						ignoreBorder.remove("se");
					}
				}
				if (ignoreBorder.contains("sw")) {
					if (bpr.getSwEntries().size() > 0) {
						ignoreBorder.remove("sw");
					}
				}

			}

		}
		ignore.addAll(ignoreBorder);

		Set<String> ignoring = new HashSet<String>(ignore);

		List<String> valids = new ArrayList<String>(Arrays.asList("n", "s", "e", "w", "ne", "nw", "se", "sw"));

		valids.removeAll(ignoring);

		String[] validArray = new String[valids.size() + 1];
		validArray[0] = cGeoString;
		int count = 1;

		for (String s : valids) {
			validArray[count] = getNeighbour(cGeoString, s);

			count++;
		}

		return validArray;
	}

	/**
	 * 
	 * @author sapmitra
	 * @param geoHash
	 * @param precision
	 * @param timeLapse
	 * @param tp
	 * @return
	 */
	public static BorderingProperties getBorderingGeohashHeuristic(String geoHash, int precision, int timeLapse,
			TemporalProperties tp, TemporalType temporalType) {

		BorderingProperties bg = new BorderingProperties();

		if (tp != null) {

			long startTS = getStartTimeStamp(tp, temporalType);
			long endTS = getEndTimeStamp(tp, temporalType);

			/* Up is the end part of the time slab */
			bg.setUp1(endTS);
			bg.setUp2(endTS - timeLapse);

			/* Down is the beginning part of the time slab */
			bg.setDown1(startTS);
			bg.setDown2(startTS + timeLapse);
		}

		String[] northAddFat = { "b", "c", "f", "g", "u", "v", "y", "z" };
		String[] southAddFat = { "0", "1", "4", "5", "h", "j", "n", "p" };
		String[] westAddFat = { "b", "8", "2", "0" };
		String[] eastAddFat = { "z", "x", "r", "p" };

		String[] northAddSlim = { "p", "r", "x", "z" };
		String[] southAddSlim = { "0", "2", "8", "b" };
		String[] westAddSlim = { "p", "n", "j", "h", "5", "4", "1", "0" };
		String[] eastAddSlim = { "z", "y", "v", "u", "g", "f", "c", "b" };

		List<String> north = Arrays.asList(geoHash);
		List<String> south = Arrays.asList(geoHash);
		List<String> east = Arrays.asList(geoHash);
		List<String> west = Arrays.asList(geoHash);

		int currentLength = 0;
		int orientation = 0;

		if (geoHash != null && geoHash.length() > 0) {
			orientation = geoHash.length() % 2;
			currentLength = geoHash.length() + 1;

		} else {
			return null;
		}

		while (currentLength <= precision) {

			if (orientation == 0) {
				List<String> northLocal = new ArrayList<String>();
				List<String> southLocal = new ArrayList<String>();
				List<String> eastLocal = new ArrayList<String>();
				List<String> westLocal = new ArrayList<String>();

				for (String s : north) {
					for (String sapp : northAddFat) {
						northLocal.add(s + sapp);
					}
				}

				for (String s : south) {
					for (String sapp : southAddFat) {
						southLocal.add(s + sapp);
					}
				}

				for (String s : east) {
					for (String sapp : eastAddFat) {
						eastLocal.add(s + sapp);
					}
				}

				for (String s : west) {
					for (String sapp : westAddFat) {
						westLocal.add(s + sapp);
					}
				}

				north = northLocal;
				south = southLocal;
				east = eastLocal;
				west = westLocal;
				orientation++;

			} else {
				List<String> northLocal = new ArrayList<String>();
				List<String> southLocal = new ArrayList<String>();
				List<String> eastLocal = new ArrayList<String>();
				List<String> westLocal = new ArrayList<String>();

				for (String s : north) {
					for (String sapp : northAddSlim) {
						northLocal.add(s + sapp);
					}
				}

				for (String s : south) {
					for (String sapp : southAddSlim) {
						southLocal.add(s + sapp);
					}
				}

				for (String s : east) {
					for (String sapp : eastAddSlim) {
						eastLocal.add(s + sapp);
					}
				}

				for (String s : west) {
					for (String sapp : westAddSlim) {
						westLocal.add(s + sapp);
					}
				}

				north = northLocal;
				south = southLocal;
				east = eastLocal;
				west = westLocal;
				orientation++;

			}
			orientation = currentLength % 2;
			currentLength++;
		}

		bg.setN(north);
		bg.setS(south);
		bg.setE(east);
		bg.setW(west);

		bg.setNw(north.get(0));
		bg.setSw(south.get(0));
		bg.setNe(north.get(north.size() - 1));
		bg.setSe(south.get(south.size() - 1));

		return bg;
	}

	private static long getEndTimeStamp(TemporalProperties tp, TemporalType temporalType) {
		if (tp == null) {
			return -1;
		}
		Calendar calendar = Calendar.getInstance();
		calendar.setTimeZone(TemporalHash.TIMEZONE);
		calendar.setTimeInMillis(tp.getEnd());

		switch (temporalType) {
		case HOUR_OF_DAY:
			calendar.set(Calendar.MINUTE, 59);
			calendar.set(Calendar.SECOND, 59);
			calendar.set(Calendar.MILLISECOND, 999);
			break;
		case DAY_OF_MONTH:
			calendar.set(Calendar.HOUR_OF_DAY, 23);
			calendar.set(Calendar.MINUTE, 59);
			calendar.set(Calendar.SECOND, 59);
			calendar.set(Calendar.MILLISECOND, 999);
			break;
		case MONTH:
			int m = calendar.get(Calendar.MONTH) + 1;
			if (m == 1 || m == 3 || m == 5 || m == 7 || m == 8 || m == 10 || m == 12)
				calendar.set(Calendar.DAY_OF_MONTH, 31);
			else if (m == 2) {
				int yr = calendar.get(Calendar.YEAR);
				if (yr % 4 == 0)
					calendar.set(Calendar.DAY_OF_MONTH, 29);
				else
					calendar.set(Calendar.DAY_OF_MONTH, 28);
			} else
				calendar.set(Calendar.DAY_OF_MONTH, 30);

			calendar.set(Calendar.HOUR_OF_DAY, 23);
			calendar.set(Calendar.MINUTE, 59);
			calendar.set(Calendar.SECOND, 59);
			calendar.set(Calendar.MILLISECOND, 999);
			break;
		case YEAR:
			calendar.set(Calendar.MONTH, 11);
			calendar.set(Calendar.DAY_OF_MONTH, 31);
			calendar.set(Calendar.HOUR_OF_DAY, 23);
			calendar.set(Calendar.MINUTE, 59);
			calendar.set(Calendar.SECOND, 59);
			calendar.set(Calendar.MILLISECOND, 999);
			break;

		}

		long offset = TemporalHash.TIMEZONE.getRawOffset() - TimeZone.getDefault().getRawOffset();

		// Date d1 = new Date(calendar.getTimeInMillis() + offset);
		// System.out.println("CONVERTED:" + d1);

		// System.out.println(calendar.getTimeInMillis() + offset);
		return calendar.getTimeInMillis() + offset;

		// return calendar.getTimeInMillis();

	}

	private static long getStartTimeStamp(TemporalProperties tp, TemporalType temporalType) {
		if (tp == null) {
			return -1;
		}
		Calendar calendar = Calendar.getInstance();
		calendar.setTimeZone(TemporalHash.TIMEZONE);
		calendar.setTimeInMillis(tp.getStart());

		switch (temporalType) {
		case HOUR_OF_DAY:
			calendar.set(Calendar.MINUTE, 0);
			calendar.set(Calendar.SECOND, 0);
			calendar.set(Calendar.MILLISECOND, 0);
			break;
		case DAY_OF_MONTH:
			calendar.set(Calendar.HOUR_OF_DAY, 0);
			calendar.set(Calendar.MINUTE, 0);
			calendar.set(Calendar.SECOND, 0);
			calendar.set(Calendar.MILLISECOND, 0);
			break;
		case MONTH:
			calendar.set(Calendar.DAY_OF_MONTH, 1);
			calendar.set(Calendar.HOUR_OF_DAY, 0);
			calendar.set(Calendar.MINUTE, 0);
			calendar.set(Calendar.SECOND, 0);
			calendar.set(Calendar.MILLISECOND, 0);
			break;
		case YEAR:
			calendar.set(Calendar.MONTH, 0);
			calendar.set(Calendar.DAY_OF_MONTH, 1);
			calendar.set(Calendar.HOUR_OF_DAY, 0);
			calendar.set(Calendar.MINUTE, 0);
			calendar.set(Calendar.SECOND, 0);
			calendar.set(Calendar.MILLISECOND, 0);
			break;

		}

		long offset = TemporalHash.TIMEZONE.getRawOffset() - TimeZone.getDefault().getRawOffset();

		// Date d1 = new Date(calendar.getTimeInMillis() + offset);
		// System.out.println("CONVERTED:" + d1);

		// System.out.println(calendar.getTimeInMillis() + offset);
		return calendar.getTimeInMillis() + offset;
		// return calendar.getTimeInMillis();

	}

	/**
	 * @author sapmitra Getting Super Coordinates. THIS IS THE FINAL METHOD THAT
	 *         GIVES THE SUPER BOUNDS FOR A GIVEN GEOHASH
	 * @param geoHash
	 * @param precision
	 */
	public static List<Coordinates> getSuperGeohashes(String geoHash, int precision) {

		String nwNeighbor = getNeighbour(geoHash, "nw");
		String nwCornerGeohash = getCornerGeohash(nwNeighbor, precision, "se");
		Coordinates nwPoint = getCornerPoint(nwCornerGeohash, "nw");

		String neNeighbor = getNeighbour(geoHash, "ne");
		String neCornerGeohash = getCornerGeohash(neNeighbor, precision, "sw");
		Coordinates nePoint = getCornerPoint(neCornerGeohash, "ne");

		String swNeighbor = getNeighbour(geoHash, "sw");
		String swCornerGeohash = getCornerGeohash(swNeighbor, precision, "ne");
		Coordinates swPoint = getCornerPoint(swCornerGeohash, "sw");

		String seNeighbor = getNeighbour(geoHash, "se");
		String seCornerGeohash = getCornerGeohash(seNeighbor, precision, "nw");
		Coordinates sePoint = getCornerPoint(seCornerGeohash, "se");

		// System.out.println(nwCornerGeohash);
		// System.out.println(neCornerGeohash);
		// System.out.println(seCornerGeohash);
		// System.out.println(swCornerGeohash);

		List<Coordinates> bounds = new ArrayList<Coordinates>();
		bounds.add(swPoint);
		bounds.add(nwPoint);
		bounds.add(nePoint);
		bounds.add(sePoint);

		// System.out.println(bounds);

		return bounds;

	}

	public static String getCornerGeohash(String geoHash, int precision, String direction) {

		// ADD ACCOMMODATION FOR TIME

		BorderingProperties bp = getBorderingGeohashHeuristic(geoHash, precision, 0, null, null);

		if ("ne".equals(direction)) {

			return bp.getNe();

		} else if ("nw".equals(direction)) {

			return bp.getNw();

		} else if ("se".equals(direction)) {

			return bp.getSe();

		} else if ("sw".equals(direction)) {

			return bp.getSw();

		}

		return null;
	}

	/**
	 * @author sapmitra
	 * @param geoHash
	 * @param direction
	 * @return
	 */
	public static Coordinates getCornerPoint(String geoHash, String direction) {

		SpatialRange range = decodeHash(geoHash);

		Coordinates c = null;

		if ("nw".equals(direction)) {

			c = new Coordinates(range.getUpperBoundForLatitude(), range.getLowerBoundForLongitude());

		} else if ("ne".equals(direction)) {

			c = new Coordinates(range.getUpperBoundForLatitude(), range.getUpperBoundForLongitude());

		} else if ("sw".equals(direction)) {

			c = new Coordinates(range.getLowerBoundForLatitude(), range.getLowerBoundForLongitude());

		} else if ("se".equals(direction)) {

			c = new Coordinates(range.getLowerBoundForLatitude(), range.getUpperBoundForLongitude());

		}

		return c;
	}

	/**
	 * 
	 * @author sapmitra
	 * @param hashes
	 * @param validNeighbors
	 * @return
	 */
	public static String[] filterUnwantedGeohashes(String[] hashes, String[] validNeighbors) {
		// TODO Auto-generated method stub

		if (hashes == null || hashes.length == 0)
			return hashes;

		List<String> filteredHashes = new ArrayList<String>();

		for (String hash : hashes) {

			boolean valid = false;

			for (String n : validNeighbors) {
				if (checkIntersection(n, hash)) {
					valid = true;
					break;
				}
			}

			if (valid) {
				filteredHashes.add(hash);
			}

		}
		return (filteredHashes.toArray(new String[filteredHashes.size()]));
	}

	/**
	 * Get all outer flanks of a given geohash
	 * 
	 * @author sapmitra
	 * @param geoHash
	 * @param precision
	 * @return
	 * 
	 *         a-----b . . . . c-----d
	 * 
	 *         returned flanks are in the order nw,n,ne,w,e,sw,s,se
	 */
	public static List<List<Coordinates>> getAllInnerFlanks(String geoHash, int precision) {

		BorderingProperties bp = getBorderingGeohashHeuristic(geoHash, precision, 0, null, null);

		String ne = bp.getNe();

		String nw = bp.getNw();

		String se = bp.getSe();

		String sw = bp.getSw();

		// System.out.println(nw);

		SpatialRange nwBounds = decodeHash(nw);
		SpatialRange neBounds = decodeHash(ne);
		SpatialRange seBounds = decodeHash(se);
		SpatialRange swBounds = decodeHash(sw);

		Coordinates caNW = new Coordinates(nwBounds.getUpperBoundForLatitude(), nwBounds.getLowerBoundForLongitude());
		Coordinates cbNW = new Coordinates(nwBounds.getUpperBoundForLatitude(), nwBounds.getUpperBoundForLongitude());
		Coordinates ccNW = new Coordinates(nwBounds.getLowerBoundForLatitude(), nwBounds.getUpperBoundForLongitude());
		Coordinates cdNW = new Coordinates(nwBounds.getLowerBoundForLatitude(), nwBounds.getLowerBoundForLongitude());

		List<Coordinates> nwFlank = Arrays.asList(caNW, cbNW, ccNW, cdNW);

		Coordinates caNE = new Coordinates(neBounds.getUpperBoundForLatitude(), neBounds.getLowerBoundForLongitude());
		Coordinates cbNE = new Coordinates(neBounds.getUpperBoundForLatitude(), neBounds.getUpperBoundForLongitude());
		Coordinates ccNE = new Coordinates(neBounds.getLowerBoundForLatitude(), neBounds.getUpperBoundForLongitude());
		Coordinates cdNE = new Coordinates(neBounds.getLowerBoundForLatitude(), neBounds.getLowerBoundForLongitude());

		List<Coordinates> neFlank = Arrays.asList(caNE, cbNE, ccNE, cdNE);

		Coordinates caSW = new Coordinates(swBounds.getUpperBoundForLatitude(), swBounds.getLowerBoundForLongitude());
		Coordinates cbSW = new Coordinates(swBounds.getUpperBoundForLatitude(), swBounds.getUpperBoundForLongitude());
		Coordinates ccSW = new Coordinates(swBounds.getLowerBoundForLatitude(), swBounds.getUpperBoundForLongitude());
		Coordinates cdSW = new Coordinates(swBounds.getLowerBoundForLatitude(), swBounds.getLowerBoundForLongitude());

		List<Coordinates> swFlank = Arrays.asList(caSW, cbSW, ccSW, cdSW);

		Coordinates caSE = new Coordinates(seBounds.getUpperBoundForLatitude(), seBounds.getLowerBoundForLongitude());
		Coordinates cbSE = new Coordinates(seBounds.getUpperBoundForLatitude(), seBounds.getUpperBoundForLongitude());
		Coordinates ccSE = new Coordinates(seBounds.getLowerBoundForLatitude(), seBounds.getUpperBoundForLongitude());
		Coordinates cdSE = new Coordinates(seBounds.getLowerBoundForLatitude(), seBounds.getLowerBoundForLongitude());

		List<Coordinates> seFlank = Arrays.asList(caSE, cbSE, ccSE, cdSE);

		List<Coordinates> northFlank = Arrays.asList(caNW, cbNE, ccNE, cdNW);
		List<Coordinates> southFlank = Arrays.asList(caSW, cbSE, ccSE, cdSW);
		List<Coordinates> eastFlank = Arrays.asList(caNE, cbNE, ccSE, cdSE);
		List<Coordinates> westFlank = Arrays.asList(caNW, cbNW, ccSW, cdSW);

		List<Coordinates> innerFlank = Arrays.asList(ccNW, cdNE, caSE, cbSW);

		List<List<Coordinates>> flanks = Arrays.asList(nwFlank, northFlank, neFlank, westFlank, innerFlank, eastFlank,
				swFlank, southFlank, seFlank);

		return flanks;

	}

	/**
	 * finding what part of geohash2 is actually needed Only that part will be
	 * loaded for a superpolygon and feature query
	 * 
	 * @author sapmitra
	 * @param geoHash1 metadata geohash from fs1
	 * @param geoHash2 metadata geohash from fs2
	 * @param time1    metadata time from fs1
	 * @param time2    metadata time from fs2
	 * @return
	 * @throws ParseException
	 */

	public static String getOrientation(String geoHash1, String geoHash2, String time1, String time2, TemporalType t1,
			TemporalType t2) throws ParseException {
		String temporalOrientation = getTemporalOrientation(time1, time2, t1, t2);
		if (temporalOrientation.contains("ignore"))
			return "ignore-ignore";
		String spatialOrientation = getSpatialOrientationHeuristic(geoHash1, geoHash2);

		return spatialOrientation + "-" + temporalOrientation;

	}

	/**
	 * RETURNS ORIENTATION OF TIMESTRING2 WRT TIMESTRING1
	 * 
	 * @author sapmitra
	 * @param time1
	 * @param time2
	 * @param t1
	 * @param t2
	 * @return
	 * @throws ParseException
	 */
	public static String getTemporalOrientation(String time1, String time2, TemporalType t1, TemporalType t2)
			throws ParseException {

		String[] tokens = time1.split("-");
		String[] tokens2 = time2.split("-");

		long startTime1 = getStartTimeStamp(tokens[0], tokens[1], tokens[2], tokens[3], t1);
		long endTime1 = getEndTimeStamp(tokens[0], tokens[1], tokens[2], tokens[3], t1);

		long startTime2 = getStartTimeStamp(tokens2[0], tokens2[1], tokens2[2], tokens2[3], t2);
		long endTime2 = getEndTimeStamp(tokens2[0], tokens2[1], tokens2[2], tokens2[3], t2);

		// time1 encloses time2
		if (startTime2 >= startTime1 && endTime2 <= endTime1)
			return "full";
		// time2 encloses time1
		else if (startTime1 >= startTime2 && endTime1 <= endTime2)
			return "full";
		// time2 lies before time1
		else if (startTime1 == endTime2 + 1 && endTime2 < endTime1)
			return "down";
		// time2 lies after time1
		else if (endTime1 + 1 == startTime2 && startTime1 < startTime2)
			return "up";

		return "ignore";
	}

	/**
	 * RETURNS ORIENTATION OF TIMESTRING2 WRT TIMESTRING1
	 * 
	 * @author sapmitra
	 * @param time1
	 * @param time2
	 * @param t1
	 * @param t2
	 * @return
	 * @throws ParseException
	 */
	public static String getTemporalOrientationSimplified(String time1, String time2, TemporalType t1, TemporalType t2)
			throws ParseException {

		String[] tokens = time1.split("-");
		String[] tokens2 = time2.split("-");

		long startTime1 = getStartTimeStamp(tokens[0], tokens[1], tokens[2], tokens[3], t1);
		long endTime1 = getEndTimeStamp(tokens[0], tokens[1], tokens[2], tokens[3], t1);

		long startTime2 = getStartTimeStamp(tokens2[0], tokens2[1], tokens2[2], tokens2[3], t2);
		long endTime2 = getEndTimeStamp(tokens2[0], tokens2[1], tokens2[2], tokens2[3], t2);

		// time1 equals time2
		if (startTime2 == startTime1 && endTime2 == endTime1)
			return "full";
		// time 1 encloses time 2
		else if (startTime2 >= startTime1 && endTime2 <= endTime1)
			return "full-st";
		// time2 encloses time1
		else if (startTime1 >= startTime2 && endTime1 <= endTime2)
			return "full-rev";

		return "ignore";
	}

	/**
	 * CHEKING TEMPORAL INTERSECTION
	 * 
	 * @author sapmitra
	 * @param timeString
	 * @param time2
	 * @param t1
	 * @param t2
	 * @return
	 * @throws ParseException
	 */
	public static boolean checkTemporalIntersection(String timeString, long qt1, long qt2) throws ParseException {

		long startTs = getStartTimeStamp(timeString, TemporalType.HOUR_OF_DAY);

		if (startTs >= qt1 && startTs <= qt2)
			return true;

		return false;
	}

	/**
	 * 
	 * @author sapmitra
	 * @param g1
	 * @param g2
	 * @param p1
	 * @param p2
	 * @return
	 */
	public static String getSpatialOrientationHeuristic(String g1, String g2) {

		// g1 is greater than equals in size than g2
		if (g1.length() <= g2.length()) {

			boolean enc = checkEnclosure(g1, g2);

			if (enc) {
				return "full";
			} else {

				// g2 may lie somewhere on the outskirts of g1

				// g2_dash is the bigger geohash that encloses g2, with the same size as g1
				String g2_dash = g2.substring(0, g1.length());

				// Neighbors are returned in the order nw,n,ne,w,e,sw,s,se
				List<String> neighbors = new ArrayList<String>(Arrays.asList(getNeighbours(g1)));

				// g2 has to lie on the border of g1
				if (neighbors.contains(g2_dash)) {

					// What kind of neighbor is g2_dash
					int indx = neighbors.indexOf(g2_dash);

					BorderingProperties bp = getBorderingGeohashHeuristic(g2_dash, g2.length(), 0, null, null);

					return getChunkDirection(g2, indx, bp, 0);

				} else {
					return "ignore";
				}

			}

		} else if (g1.length() > g2.length()) {

			// g1 is smaller in size than g2
			boolean enc = checkEnclosure(g2, g1);

			if (enc) {
				// check inside
				BorderingProperties bp = getBorderingGeohashHeuristic(g2, g1.length(), 0, null, null);

				return "full";

			} else {

				// check outside
				String g1_dash = g1.substring(0, g2.length());
				// Neighbors are returned in the order nw,n,ne,w,e,sw,s,se
				List<String> neighbors = new ArrayList<String>(Arrays.asList(getNeighbours(g2)));

				if (neighbors.contains(g1_dash)) {

					// whar kind of neighbbor is
					int indx = neighbors.indexOf(g1_dash);

					BorderingProperties bp = getBorderingGeohashHeuristic(g1_dash, g1.length(), 0, null, null);

					return getChunkDirection(g1, indx, bp, 1);

				} else {
					return "ignore";
				}

			}

		}

		return "ignore";

	}

	/**
	 * Check which of g1 and g2 contains whom
	 * 
	 * @author sapmitra
	 * @param g1
	 * @param g2
	 * @return
	 */
	public static String getSpatialOrientationSimplified(String g1, String g2) {

		String orientation = "none";

		// THEY BOTH ARE THE SAME
		if (g1.equals(g2))
			return "full";

		boolean enc = checkEnclosure(g1, g2);

		if (enc)
			return "full-st";

		boolean enc2 = checkEnclosure(g2, g1);
		if (enc2)
			return "full-rev";

		return orientation;

	}

	/**
	 * Determines which flank of a geohash is needed
	 * 
	 * @author sapmitra
	 * @param g2
	 * @param indx
	 * @param bp
	 * @return
	 */
	private static String getChunkDirection(String g2, int indx, BorderingProperties bp, int type) {
		// if g2_dash is a nw neighbor
		if (indx == 0) {

			// Only case where g2 would matter is if g2 is the se corner
			if (bp.getSe().equals(g2)) {

				if (type == 0)
					return "se";
				else if (type == 1)
					return "nw";

			}

		}
		// if g2_dash is a n neighbor
		else if (indx == 1) {

			// Only case where g2 would matter is if g2 is the s corner
			if (bp.getS().contains(g2)) {
				if (type == 0)
					return "s";
				else if (type == 1)
					return "n";
			}
		}

		// if g2_dash is a ne neighbor
		else if (indx == 2) {

			// Only case where g2 would matter is if g2 is the s corner
			if (bp.getSw().equals(g2)) {
				if (type == 0)
					return "sw";
				else if (type == 1)
					return "ne";
			}
		}

		// if g2_dash is a w neighbor
		else if (indx == 3) {

			// Only case where g2 would matter is if g2 is the s corner
			if (bp.getE().contains(g2)) {
				if (type == 0)
					return "e";
				else if (type == 1)
					return "w";

			}
		}

		// if g2_dash is a e neighbor
		else if (indx == 4) {

			// Only case where g2 would matter is if g2 is the s corner
			if (bp.getW().contains(g2)) {
				if (type == 0)
					return "w";
				else if (type == 1)
					return "e";

			}
		}
		// if g2_dash is a sw neighbor
		else if (indx == 5) {

			// Only case where g2 would matter is if g2 is the s corner
			if (bp.getNe().equals(g2)) {

				if (type == 0)
					return "ne";
				else if (type == 1)
					return "sw";

			}
		}

		// if g2_dash is a n neighbor
		else if (indx == 6) {

			// Only case where g2 would matter is if g2 is the s corner
			if (bp.getN().contains(g2)) {

				if (type == 0)
					return "n";
				else if (type == 1)
					return "s";
			}
		}

		// if g2_dash is a ne neighbor
		else if (indx == 7) {

			// Only case where g2 would matter is if g2 is the s corner
			if (bp.getNw().equals(g2)) {

				if (type == 0)
					return "nw";
				else if (type == 1)
					return "se";
			}
		}
		return "ignore";
	}

	/**
	 * Returns the geohash id of the parent. level specifies how far up to go. THIS
	 * DOES NOT RETURN SPATIAL ID.
	 * 
	 * @author sapmitra
	 * @param geoHash
	 * @param level   - how many levels above do you want to go?
	 * @return
	 */
	public static String getSpatialParent(String geoHash, int level) {

		if (geoHash.length() <= level) {
			return null;
		}

		String parentGeohash = geoHash.substring(0, geoHash.length() - level);

		return parentGeohash;

	}

	/**
	 * Get the Ids of all the internal children. THIS DOES NOT RETURN SPATIAL ID.
	 * 
	 * @author sapmitra
	 * @param geoHash
	 * @param level
	 * @return
	 */
	public static List<String> getSpatialChildren(String geoHash) {

		int length = geoHash.length();

		List<Character> addOns = null;

		if (length % 2 == 0) {
			addOns = evenChars;
		} else {
			addOns = oddChars;
		}

		List<String> neighbors = new ArrayList<String>();

		for (Character c : addOns) {

			neighbors.add(geoHash + c);
		}

		return neighbors;

	}

	/**
	 * For Neighbors of the same size of the actual geohash THIS DOES NOT RETURN
	 * SPATIAL ID. Neighbors are returned in the order nw,n,ne,w,e,sw,s,se
	 * 
	 * @param geoHash
	 * @return
	 */

	public static SpatialBorder getSpatialNeighbours(String geoHash, int reqPrecision) {

		SpatialBorder bp = new SpatialBorder();

		if (geoHash == null || geoHash.trim().length() == 0)
			throw new IllegalArgumentException("Invalid Geohash");
		geoHash = geoHash.trim();

		int precision = geoHash.length();
		SpatialRange boundingBox = decodeHash(geoHash);
		Coordinates centroid = boundingBox.getCenterPoint();
		float widthDiff = boundingBox.getUpperBoundForLongitude() - centroid.getLongitude();
		float heightDiff = boundingBox.getUpperBoundForLatitude() - centroid.getLatitude();

		if (reqPrecision == precision) {

			bp.setNw(encode(boundingBox.getUpperBoundForLatitude() + heightDiff,
					boundingBox.getLowerBoundForLongitude() - widthDiff, precision));

			bp.addN(encode(boundingBox.getUpperBoundForLatitude() + heightDiff, centroid.getLongitude(), precision));

			bp.setNe(encode(boundingBox.getUpperBoundForLatitude() + heightDiff,
					boundingBox.getUpperBoundForLongitude() + widthDiff, precision));

			bp.addW(encode(centroid.getLatitude(), boundingBox.getLowerBoundForLongitude() - widthDiff, precision));

			bp.addE(encode(centroid.getLatitude(), boundingBox.getUpperBoundForLongitude() + widthDiff, precision));

			bp.setSw(encode(boundingBox.getLowerBoundForLatitude() - heightDiff,
					boundingBox.getLowerBoundForLongitude() - widthDiff, precision));

			bp.addS(encode(boundingBox.getLowerBoundForLatitude() - heightDiff, centroid.getLongitude(), precision));

			bp.setSe(encode(boundingBox.getLowerBoundForLatitude() - heightDiff,
					boundingBox.getUpperBoundForLongitude() + widthDiff, precision));

			return bp;

		} else {

			String nw = encode(boundingBox.getUpperBoundForLatitude() + heightDiff,
					boundingBox.getLowerBoundForLongitude() - widthDiff, precision);
			String n = encode(boundingBox.getUpperBoundForLatitude() + heightDiff, centroid.getLongitude(), precision);
			String ne = encode(boundingBox.getUpperBoundForLatitude() + heightDiff,
					boundingBox.getUpperBoundForLongitude() + widthDiff, precision);
			String w = encode(centroid.getLatitude(), boundingBox.getLowerBoundForLongitude() - widthDiff, precision);
			String e = encode(centroid.getLatitude(), boundingBox.getUpperBoundForLongitude() + widthDiff, precision);
			String sw = encode(boundingBox.getLowerBoundForLatitude() - heightDiff,
					boundingBox.getLowerBoundForLongitude() - widthDiff, precision);
			String s = encode(boundingBox.getLowerBoundForLatitude() - heightDiff, centroid.getLongitude(), precision);
			String se = encode(boundingBox.getLowerBoundForLatitude() - heightDiff,
					boundingBox.getUpperBoundForLongitude() + widthDiff, precision);

			// nw bordering geohash
			List<String> nwS = getBorderingGeohashesGivenDirection(nw, reqPrecision, "s");
			bp.setNw(nwS.get(nwS.size() - 1));

			// ne bordering geohash
			List<String> neS = getBorderingGeohashesGivenDirection(ne, reqPrecision, "s");
			bp.setNe(neS.get(0));

			// sw bordering geohash
			List<String> swS = getBorderingGeohashesGivenDirection(sw, reqPrecision, "n");
			bp.setSw(swS.get(swS.size() - 1));

			// se bordering geohash
			List<String> seS = getBorderingGeohashesGivenDirection(se, reqPrecision, "n");
			bp.setSe(seS.get(0));

			bp.setN(getBorderingGeohashesGivenDirection(n, reqPrecision, "s"));
			bp.setS(getBorderingGeohashesGivenDirection(s, reqPrecision, "n"));
			bp.setE(getBorderingGeohashesGivenDirection(e, reqPrecision, "w"));
			bp.setW(getBorderingGeohashesGivenDirection(w, reqPrecision, "e"));

			return bp;

		}

	}

	/**
	 * Returns the geohash id of the parent. level specifies how far up to go.
	 * 
	 * @author sapmitra
	 * @param geoHash
	 * @param level
	 * @return
	 */
	public static int getSpatialParentID(String geoHash, int level) {

		if (geoHash.length() <= level) {
			return -1;
		}

		String parentGeohash = geoHash.substring(0, geoHash.length() - level);

		return getGeohashID(parentGeohash);

	}

	/**
	 * Get the Ids of all the internal geohashes
	 * 
	 * @author sapmitra
	 * @param geoHash
	 * @param level
	 * @return
	 */
	public static int[] getSpatialChildrenID(String geoHash, int level) {

		SpatialRange range1 = GeoHash.decodeHash(geoHash);

		Coordinates c1 = new Coordinates(range1.getLowerBoundForLatitude(), range1.getLowerBoundForLongitude());
		Coordinates c2 = new Coordinates(range1.getUpperBoundForLatitude(), range1.getLowerBoundForLongitude());
		Coordinates c3 = new Coordinates(range1.getUpperBoundForLatitude(), range1.getUpperBoundForLongitude());
		Coordinates c4 = new Coordinates(range1.getLowerBoundForLatitude(), range1.getUpperBoundForLongitude());

		ArrayList<Coordinates> cs1 = new ArrayList<Coordinates>();
		cs1.add(c1);
		cs1.add(c2);
		cs1.add(c3);
		cs1.add(c4);

		String[] internalGeohashes = GeoHash.getIntersectingGeohashesForConvexBoundingPolygon(cs1, level);

		int[] internalIDs = new int[internalGeohashes.length];

		for (int i = 0; i < internalGeohashes.length; i++) {
			internalIDs[i] = getGeohashID(internalGeohashes[i]);
		}

		return internalIDs;

	}

	/**
	 * Get the indices of a geohash's spatial neighbors, grouped by direction N,NE
	 * and NW are disjoint sets
	 * 
	 * @author sapmitra
	 * @param geohash
	 */
	public static SpatialBorderIndices getSpatialNeighborID(String geohash, int precision) {

		SpatialBorder spatialNeighbours = getSpatialNeighbours(geohash, precision);

		SpatialBorderIndices sind = new SpatialBorderIndices();
		sind.setNe(getGeohashID(spatialNeighbours.getNe()));
		sind.setNw(getGeohashID(spatialNeighbours.getNw()));
		sind.setSe(getGeohashID(spatialNeighbours.getSe()));
		sind.setSw(getGeohashID(spatialNeighbours.getSw()));

		for (String s : spatialNeighbours.getN()) {
			sind.addN(getGeohashID(s));
		}
		for (String s : spatialNeighbours.getS()) {
			sind.addS(getGeohashID(s));
		}
		for (String s : spatialNeighbours.getE()) {
			sind.addE(getGeohashID(s));
		}
		for (String s : spatialNeighbours.getW()) {
			sind.addW(getGeohashID(s));
		}

		return sind;
	}

	/**
	 * Converting geohash into a serial number
	 * 
	 * @author sapmitra
	 * @param geohash
	 * @return
	 */
	private static int getGeohashID(String geohash) {

		List<Character> evens = Arrays.asList('b', 'c', 'f', 'g', 'u', 'v', 'y', 'z', '8', '9', 'd', 'e', 's', 't', 'w',
				'x', '2', '3', '6', '7', 'k', 'm', 'q', 'r', '0', '1', '4', '5', 'h', 'j', 'n', 'p');
		List<Character> odds = Arrays.asList('p', 'r', 'x', 'z', 'n', 'q', 'w', 'y', 'j', 'm', 't', 'v', 'h', 'k', 's',
				'u', '5', '7', 'e', 'g', '4', '6', 'd', 'f', '1', '3', '9', 'c', '0', '2', '8', 'b');

		// odd 8x4
		// even 4x8

		int resolution = geohash.length();
		char[] gArray = geohash.toCharArray();

		int rowMultiplier = 1;
		int colMultiplier = 1;

		int row_prev = 0;
		int col_prev = 0;

		for (int i = 0; i < resolution; i++) {

			char cr = gArray[i];

			if (i % 2 == 0) {

				int indx = evens.indexOf(cr);

				int r = indx / 8; // 1
				int c = indx % 8; // 2

				row_prev = row_prev * 4 + r;
				col_prev = col_prev * 8 + c;

				// if(i != resolution-1) {
				rowMultiplier *= 8;
				colMultiplier *= 4;
				// }

			} else {

				int indx = odds.indexOf(cr);

				int r = indx / 4;
				int c = indx % 4;

				row_prev = row_prev * 8 + r;
				col_prev = col_prev * 4 + c;

				// if(i != resolution-1) {
				rowMultiplier *= 4;
				colMultiplier *= 8;
				// }

			}

		}
		int finalIndex = row_prev * rowMultiplier + col_prev;

		// System.out.println(finalIndex);
		return finalIndex;

	}

	private static int getGeohashIDWithOffset(String geohash, int offset) {

		List<Character> evens = Arrays.asList('b', 'c', 'f', 'g', 'u', 'v', 'y', 'z', '8', '9', 'd', 'e', 's', 't', 'w',
				'x', '2', '3', '6', '7', 'k', 'm', 'q', 'r', '0', '1', '4', '5', 'h', 'j', 'n', 'p');
		List<Character> odds = Arrays.asList('p', 'r', 'x', 'z', 'n', 'q', 'w', 'y', 'j', 'm', 't', 'v', 'h', 'k', 's',
				'u', '5', '7', 'e', 'g', '4', '6', 'd', 'f', '1', '3', '9', 'c', '0', '2', '8', 'b');

		// odd 8x4
		// even 4x8

		int resolution = geohash.length();
		char[] gArray = geohash.toCharArray();

		int rowMultiplier = 1;
		int colMultiplier = 1;

		int row_prev = 0;
		int col_prev = 0;

		for (int i = offset; i < resolution; i++) {

			char cr = gArray[i];

			if (i % 2 == 0) {

				int indx = evens.indexOf(cr);

				int r = indx / 8; // 1
				int c = indx % 8; // 2

				row_prev = row_prev * 4 + r;
				col_prev = col_prev * 8 + c;

				// if(i != resolution-1) {
				rowMultiplier *= 8;
				colMultiplier *= 4;
				// }

			} else {

				int indx = odds.indexOf(cr);

				int r = indx / 4;
				int c = indx % 4;

				row_prev = row_prev * 8 + r;
				col_prev = col_prev * 4 + c;

				// if(i != resolution-1) {
				rowMultiplier *= 4;
				colMultiplier *= 8;
				// }

			}

		}
		int finalIndex = row_prev * rowMultiplier + col_prev;

		// System.out.println(finalIndex);
		return finalIndex;

	}

	/**
	 * Get the geohashes that lie outside a geohash's border in a particular
	 * direction
	 * 
	 * @author sapmitra
	 * @param geoHash
	 * @param precision
	 * @param direction
	 * @return
	 */
	public static List<String> getBorderingGeohashesGivenDirection(String geoHash, int precision, String direction) {

		String[] northAddFat = { "b", "c", "f", "g", "u", "v", "y", "z" };
		String[] southAddFat = { "0", "1", "4", "5", "h", "j", "n", "p" };
		String[] westAddFat = { "b", "8", "2", "0" };
		String[] eastAddFat = { "z", "x", "r", "p" };

		String[] northAddSlim = { "p", "r", "x", "z" };
		String[] southAddSlim = { "0", "2", "8", "b" };
		String[] westAddSlim = { "p", "n", "j", "h", "5", "4", "1", "0" };
		String[] eastAddSlim = { "z", "y", "v", "u", "g", "f", "c", "b" };

		List<String> north = Arrays.asList(geoHash);
		List<String> south = Arrays.asList(geoHash);
		List<String> east = Arrays.asList(geoHash);
		List<String> west = Arrays.asList(geoHash);

		int currentLength = 0;
		int orientation = 0;

		if (geoHash != null && geoHash.length() > 0) {
			orientation = geoHash.length() % 2;
			currentLength = geoHash.length() + 1;

		} else {
			return null;
		}

		while (currentLength <= precision) {

			if (orientation == 0) {
				List<String> northLocal = new ArrayList<String>();
				List<String> southLocal = new ArrayList<String>();
				List<String> eastLocal = new ArrayList<String>();
				List<String> westLocal = new ArrayList<String>();

				if (direction == "n") {
					for (String s : north) {
						for (String sapp : northAddFat) {
							northLocal.add(s + sapp);
						}
					}
				} else if (direction == "s") {

					for (String s : south) {
						for (String sapp : southAddFat) {
							southLocal.add(s + sapp);
						}
					}
				} else if (direction == "e") {

					for (String s : east) {
						for (String sapp : eastAddFat) {
							eastLocal.add(s + sapp);
						}
					}
				} else if (direction == "w") {
					for (String s : west) {
						for (String sapp : westAddFat) {
							westLocal.add(s + sapp);
						}
					}
				}
				north = northLocal;
				south = southLocal;
				east = eastLocal;
				west = westLocal;
				orientation++;

			} else {
				List<String> northLocal = new ArrayList<String>();
				List<String> southLocal = new ArrayList<String>();
				List<String> eastLocal = new ArrayList<String>();
				List<String> westLocal = new ArrayList<String>();

				if (direction == "n") {
					for (String s : north) {
						for (String sapp : northAddSlim) {
							northLocal.add(s + sapp);
						}
					}
				} else if (direction == "s") {
					for (String s : south) {
						for (String sapp : southAddSlim) {
							southLocal.add(s + sapp);
						}
					}
				} else if (direction == "e") {
					for (String s : east) {
						for (String sapp : eastAddSlim) {
							eastLocal.add(s + sapp);
						}
					}
				} else if (direction == "w") {
					for (String s : west) {
						for (String sapp : westAddSlim) {
							westLocal.add(s + sapp);
						}
					}
				}
				north = northLocal;
				south = southLocal;
				east = eastLocal;
				west = westLocal;
				orientation++;

			}
			orientation = currentLength % 2;
			currentLength++;
		}

		if (direction == "n")
			return north;
		else if (direction == "s")
			return south;
		else if (direction == "e")
			return east;
		else if (direction == "w")
			return west;
		else
			return null;
	}

	/**
	 * Takes the yyyy-mm-dd-hh string as input This can also be used to get the
	 * temporal parent if the temporal resolution is lesser than the timestring's
	 * resolution
	 * 
	 * @author sapmitra
	 * @param timeString
	 * @param temporalResolution
	 * @return
	 */

	public static int getTemporalID(String timeString, int temporalResolution) {

		String[] components = timeString.split("-");

		if (components.length < temporalResolution) {
			return -1;
		}

		if (temporalResolution == 1) {
			// year
			int yr = Integer.valueOf(components[0]);
			return yr;

		} else if (temporalResolution == 2) {
			// month
			int yr = Integer.valueOf(components[0]);
			int mm = Integer.valueOf(components[1]);
			int id = yr * 12 + mm;
			return id;

		} else if (temporalResolution == 3) {
			// day
			int yr = Integer.valueOf(components[0]);
			int mm = Integer.valueOf(components[1]);
			int dd = Integer.valueOf(components[2]);
			return daysBetweenDates(yr, mm, dd);

		} else if (temporalResolution == 4) {
			// hour
			int yr = Integer.valueOf(components[0]);
			int mm = Integer.valueOf(components[1]);
			int dd = Integer.valueOf(components[2]);
			int hh = Integer.valueOf(components[3]);
			int daysBw = daysBetweenDates(yr, mm, dd);
			return daysBw * 24 + hh;

		} else {
			return -1;
		}

	}

	/**
	 * Returns time-string one level up
	 * 
	 * @author sapmitra
	 * @param timeString
	 * @param temporalLevel - current temporal level
	 * @return
	 */
	public static String getTemporalParent(String timeString, int temporalLevel) {

		String[] components = timeString.split("-");

		if (temporalLevel <= 1) {
			return null;
		}

		String parent = components[0];

		for (int i = 1; i < temporalLevel - 1; i++) {

			parent += "-" + components[i];
		}

		for (int i = temporalLevel - 1; i < 4; i++) {

			parent += "-xx";
		}

		return parent;

	}

	/**
	 * Get temporal children for the immediate lower temporal level. RETURNS
	 * CHILDREN ONE LEVEL DOWN
	 * 
	 * @author sapmitra
	 * @param timeString
	 * @param temporalResolution This is the current temporal resolution of the
	 *                           timeString
	 * @return
	 */
	public static List<String> getTemporalChildren(String timeString, int temporalResolution) {

		List<String> children = new ArrayList<String>();

		String[] components = timeString.split("-");

		if (temporalResolution == 1) {
			// year

			for (int i = 1; i < 13; i++) {

				String child = components[0] + "-" + i + "-xx-xx";
				children.add(child);

			}

			return children;

		} else if (temporalResolution == 2) {
			// month

			int yr = Integer.valueOf(components[0]);
			int mm = Integer.valueOf(components[1]);

			Calendar mycal = new GregorianCalendar(yr, mm - 1, 1);

			// Get the number of days in that month
			int daysInMonth = mycal.getActualMaximum(Calendar.DAY_OF_MONTH);

			for (int i = 1; i <= daysInMonth; i++) {

				children.add(components[0] + "-" + components[1] + "-" + i + "-xx");

			}
			return children;

		} else if (temporalResolution == 3) {
			// day

			for (int i = 0; i < 24; i++) {

				children.add(components[0] + "-" + components[1] + "-" + components[2] + "-" + i);

			}

			return children;

		} else {
			return null;
		}

	}

	public static String getTemporalString(String yr, String mon, String day, String hr) {

		return yr + "-" + mon + "-" + day + "-" + hr;
	}

	/**
	 * Getting temporal neighbors on the same level. THIS DOES NOT RETURN TEMPORAL
	 * ID.
	 * 
	 * @author sapmitra
	 * @param dateString
	 * @param temporalResolution
	 * @return
	 */
	public static List<String> getTemporalNeighbors(String dateString, int temporalResolution) {

		List<String> neighbors = new ArrayList<String>();

		String[] components = dateString.split("-");

		// year
		if (temporalResolution == 1) {

			int yr = Integer.valueOf(components[0]);

			for (int i = 0; i < GeospatialFileSystem.TEMPORAL_SPREAD * 5; i++) {

				neighbors.add(
						getTemporalString(String.valueOf(yr - i + 1), components[1], components[2], components[3]));
				neighbors.add(
						getTemporalString(String.valueOf(yr + i + 1), components[1], components[2], components[3]));

			}

			return neighbors;

		} else if (temporalResolution == 2) {
			// month
			int yr = Integer.valueOf(components[0]);
			int mm = Integer.valueOf(components[1]);

			for (int i = 0; i < GeospatialFileSystem.TEMPORAL_SPREAD * 6; i++) {
				GregorianCalendar cal1 = new GregorianCalendar(yr, mm - 1, 1);
				cal1.add(Calendar.MONTH, -1 * (i + 1));
				neighbors.add(cal1.get(Calendar.YEAR) + "-" + (cal1.get(Calendar.MONTH) + 1) + "-xx-xx");

				GregorianCalendar cal2 = new GregorianCalendar(yr, mm - 1, 1);
				cal2.add(Calendar.MONTH, (i + 1));
				neighbors.add(cal2.get(Calendar.YEAR) + "-" + (cal2.get(Calendar.MONTH) + 1) + "-xx-xx");

			}

			return neighbors;

		} else if (temporalResolution == 3) {
			// day
			int yr = Integer.valueOf(components[0]);
			int mm = Integer.valueOf(components[1]);
			int dd = Integer.valueOf(components[2]);

			for (int i = 0; i < GeospatialFileSystem.TEMPORAL_SPREAD * 4; i++) {

				GregorianCalendar cal1 = new GregorianCalendar(yr, mm - 1, dd);
				cal1.add(Calendar.DATE, -1 * (i + 1));
				neighbors.add(cal1.get(Calendar.YEAR) + "-" + (cal1.get(Calendar.MONTH) + 1) + "-"
						+ cal1.get(Calendar.DATE) + "-xx");

				GregorianCalendar cal2 = new GregorianCalendar(yr, mm - 1, dd);
				cal2.add(Calendar.DATE, (i + 1));
				neighbors.add(cal2.get(Calendar.YEAR) + "-" + (cal2.get(Calendar.MONTH) + 1) + "-"
						+ cal2.get(Calendar.DATE) + "-xx");

			}

			return neighbors;

		} else if (temporalResolution == 4) {

			// day
			int yr = Integer.valueOf(components[0]);
			int mm = Integer.valueOf(components[1]);
			int dd = Integer.valueOf(components[2]);
			int hh = Integer.valueOf(components[3]);

			for (int i = 0; i < GeospatialFileSystem.TEMPORAL_SPREAD * 3; i++) {

				GregorianCalendar cal1 = new GregorianCalendar(yr, mm - 1, dd, hh, 0);
				cal1.add(Calendar.HOUR_OF_DAY, -1 * (i + 1));
				neighbors.add(cal1.get(Calendar.YEAR) + "-" + (cal1.get(Calendar.MONTH) + 1) + "-"
						+ cal1.get(Calendar.DATE) + "-" + cal1.get(Calendar.HOUR_OF_DAY));

				GregorianCalendar cal2 = new GregorianCalendar(yr, mm - 1, dd, hh, 0);
				cal2.add(Calendar.HOUR_OF_DAY, (i + 1));
				neighbors.add(cal2.get(Calendar.YEAR) + "-" + (cal2.get(Calendar.MONTH) + 1) + "-"
						+ cal2.get(Calendar.DATE) + "-" + cal2.get(Calendar.HOUR_OF_DAY));

			}

			return neighbors;

		} else {
			return null;
		}

	}

	public static int daysBetweenDates(int yr, int mm, int dd) {
		Calendar cal1 = new GregorianCalendar();
		Calendar cal2 = new GregorianCalendar();

		cal1.set(1970, 0, 1);
		cal2.set(yr, mm - 1, dd);

		return daysBetween(cal1.getTime(), cal2.getTime());
	}

	public static int daysBetween(Date d1, Date d2) {
		return (int) ((d2.getTime() - d1.getTime()) / (1000 * 60 * 60 * 24));
	}

	/**
	 * Getting temporal neighbors on the same level
	 * 
	 * @author sapmitra
	 * @param dateString
	 * @param temporalResolution
	 * @return
	 */
	public static int[] getTemporalNeighborsID(String dateString, int temporalResolution) {

		int[] neighbors = new int[2];
		String[] components = dateString.split("-");
		int yr = Integer.valueOf(components[0]);
		int mm = Integer.valueOf(components[1]);
		int dd = Integer.valueOf(components[2]);
		int hh = Integer.valueOf(components[3]);
		// year
		if (temporalResolution == 1) {

			neighbors[0] = yr - 1;
			neighbors[1] = yr + 1;
			return neighbors;

		} else if (temporalResolution == 2) {
			// month

			int id = yr * 12 + mm;

			neighbors[0] = id - 1;
			neighbors[1] = id + 1;
			return neighbors;

		} else if (temporalResolution == 3) {
			// day
			int id = daysBetweenDates(yr, mm, dd);
			neighbors[0] = id - 1;
			neighbors[1] = id + 1;
			return neighbors;

		} else if (temporalResolution == 4) {

			int daysBw = daysBetweenDates(yr, mm, dd);
			int id = daysBw * 24 + hh;

			neighbors[0] = id - 1;
			neighbors[1] = id + 1;
			return neighbors;

		} else {
			return null;
		}

	}

	/**
	 * RETURNS THE DESIRED KEY AS TIME$$SPACE TIME IS YEAR-MONTH-...
	 * 
	 * SUMMARY KEY SHOULD BE CALCULATED FOR EACH RECORD ONLY IF THE REQUESTED
	 * RESOLUTION IS HIGHER THAN THE RESOLUTION OF THE ACTUAL BLOCK
	 * 
	 * @author sapmitra
	 * @param lat
	 * @param lng
	 * @param time
	 * @param spatialResolution
	 * @param temporalResolution
	 * @return
	 * @throws ParseException
	 */
	public static String getSummaryKey(float lat, float lng, float time, int spatialResolution,
			int temporalResolution) {

		String geohash = encode(lat, lng, spatialResolution);

		TemporalType tt = TemporalType.YEAR;

		if (temporalResolution == 1)
			tt = TemporalType.YEAR;
		else if (temporalResolution == 2)
			tt = TemporalType.MONTH;
		else if (temporalResolution == 3)
			tt = TemporalType.DAY_OF_MONTH;
		else if (temporalResolution == 4)
			tt = TemporalType.HOUR_OF_DAY;

		String temporalString = getTimeStringFromTimestamp(time, tt);

		return temporalString + "$$" + geohash;

	}

	/**
	 * 
	 * @author sapmitra
	 * @param year
	 * @param month
	 * @param day
	 * @param hour
	 * @param temporalType
	 * @return
	 * @throws ParseException
	 */
	public static String getTimeStringFromTimestamp(float timestamp, TemporalType temporalType) {

		String temporalString = "";
		Calendar calendar = Calendar.getInstance(TemporalHash.TIMEZONE);
		calendar.setTimeZone(TemporalHash.TIMEZONE);

		long ts = (long) timestamp;
		if (Long.toString(ts).length() < 13) {
			ts = ts * 1000;
		}

		calendar.setTimeInMillis(ts);

		// calendar.setLenient(false);

		int year = calendar.get(Calendar.YEAR);
		int month = calendar.get(Calendar.MONTH) + 1;
		int day = calendar.get(Calendar.DAY_OF_MONTH);
		int hour = calendar.get(Calendar.HOUR_OF_DAY);

		switch (temporalType) {
		case HOUR_OF_DAY:
			temporalString = year + "-" + month + "-" + day + "-" + hour;
			break;
		case DAY_OF_MONTH:
			temporalString = year + "-" + month + "-" + day + "-xx";
			break;
		case MONTH:
			temporalString = year + "-" + month + "-xx-xx";
			break;
		case YEAR:
			temporalString = String.valueOf(year) + "-xx-xx-xx";
			break;

		}

		return temporalString;

	}

	/**
	 * GO UP A FEW TEMPORAL LEVELS, IF NEEDED
	 * 
	 * @author sapmitra
	 * @param timestamp
	 * @param temporalType
	 * @return
	 */
	public static String getTimeStringForLevel(String timestring, TemporalType temporalType) {

		String[] tokens = timestring.split("-");

		// calendar.setLenient(false);
		String temporalString = "";
		String year = tokens[0];
		String month = tokens[1];
		String day = tokens[2];
		String hour = tokens[3];

		switch (temporalType) {
		case HOUR_OF_DAY:
			temporalString = year + "-" + month + "-" + day + "-" + hour;
			break;
		case DAY_OF_MONTH:
			temporalString = year + "-" + month + "-" + day + "-xx";
			break;
		case MONTH:
			temporalString = year + "-" + month + "-xx-xx";
			break;
		case YEAR:
			temporalString = String.valueOf(year) + "-xx-xx-xx";
			break;

		}

		return temporalString;

	}

	public static void main3(String arg[]) {

		System.out.println(getSpatialOrientationSimplified("9zh0", "9z5b1"));
		System.out.println(getSpatialOrientationSimplified("9zh0", "9zh0"));
		System.out.println(getSpatialOrientationSimplified("9zh0", "9zh01"));
		System.out.println(getSpatialOrientationSimplified("9zh0", "9zh"));
		// List<String> borderingGeohashesForDirection =
		// getBorderingGeohashesGivenDirection("u9", 4, "s");

		// System.out.println(borderingGeohashesForDirection);

		// SpatialBorder spatialNeighbours = getSpatialNeighbours("s", 3);

		// System.out.println(spatialNeighbours);
		// getSpatialParent("u9hv",2);
		// int[] spatialChildren = getSpatialChildren("u9", 4);
		// daysBetweenDates(1990, 7, 7);

		// System.out.println(getSummaryKey(40.58f, -105.08f, 1537990007000f, 7, 4));

		// System.out.println(GeoHash.hashToLong("00p0"));

		// System.out.println(getTemporalParent("2018-01-11-xx", 3));

		System.out.println(getEndTimeStamp("2017-2-xx-xx", TemporalType.MONTH));
		System.out.println(getStartTimeStamp("2017-2-xx-xx", TemporalType.MONTH));

		long tmp = 1544227021l;

		System.out.println(Long.toString(tmp).length());

		tmp = 1544227021000l;
		System.out.println(Long.toString(tmp).length());

	}

	/**
	 * 
	 * @author sapmitra
	 * @param geoHash
	 * @param reqPrecision - This tells the number blocks of size geohashlength-1 to
	 *                     extend
	 * @return
	 */
	public static List<String> getSpatialNeighboursSimplified(String geoHash, int reqPrecision) {

		List<String> spatialNeighbors = new ArrayList<String>();

		int numStepsSide = 0;
		int numStepsUp = 0;

		if (geoHash.length() % 2 == 0) {
			numStepsSide = 4 * reqPrecision;
			numStepsUp = 2 * reqPrecision;
		} else {
			numStepsUp = 4 * reqPrecision;
			numStepsSide = 2 * reqPrecision;
		}

		List<String> startingLineup = new ArrayList<String>();
		startingLineup.add(geoHash);

		for (int i = 0; i < numStepsUp; i++) {

			startingLineup.add(getNeighbour(geoHash, "n"));

			startingLineup.add(getNeighbour(geoHash, "s"));

			spatialNeighbors.add(getNeighbour(geoHash, "n"));

			spatialNeighbors.add(getNeighbour(geoHash, "s"));
		}

		for (String s : startingLineup) {

			for (int i = 0; i < numStepsSide; i++) {

				spatialNeighbors.add(getNeighbour(s, "e"));

				spatialNeighbors.add(getNeighbour(s, "w"));

			}
		}

		return spatialNeighbors;

	}

	/**
	 * GIVEN A GEOHASH, FIND THE GEOHASH DIAGONALLY OPPOSITE TO IT.
	 * 
	 * @author sapmitra
	 * @param geoHash
	 */
	public static String getAntipodeGeohash_Old(String geoHash) {

		SpatialRange decodeHash = decodeHash(geoHash);

		double centralLat = (decodeHash.getLowerBoundForLatitude() + decodeHash.getUpperBoundForLatitude()) / 2d;
		double centralLon = (decodeHash.getLowerBoundForLongitude() + decodeHash.getUpperBoundForLongitude()) / 2d;

		if (centralLon > 0)
			centralLon = centralLon - 180;
		else
			centralLon = 180 + centralLon;

		String newGeoHash = encode((float) (centralLat * -1), (float) centralLon, geoHash.length());

		// System.out.println(newGeoHash);

		return newGeoHash;

	}

	/**
	 * GIVEN A GEOHASH, FIND THE GEOHASH DIAGONALLY OPPOSITE TO IT.
	 * 
	 * @author sapmitra
	 * @param geoHash
	 */
	public static String getAntipodeGeohash(String geoHash) {

		// {upperLat, lowerLat, upperLon, lowerLon, centralLat, centralLon};
		double[] partitionLatLons = StandardDHTPartitioner.getCentralLatLons();

		SpatialRange decodeHash = decodeHash(geoHash);

		// CENTRAL LAT LON OF THE GIVEN GEOHASH
		double geohashCentralLat = (decodeHash.getLowerBoundForLatitude() + decodeHash.getUpperBoundForLatitude()) / 2d;
		double geohashCentralLon = (decodeHash.getLowerBoundForLongitude() + decodeHash.getUpperBoundForLongitude())
				/ 2d;

		double diffLat = geohashCentralLat - partitionLatLons[4];

		double newLat = partitionLatLons[4] - diffLat;

		double diffLon = geohashCentralLon - partitionLatLons[5];

		double newLon = partitionLatLons[5] - diffLon;

		String newGeoHash = encode((float) newLat, (float) newLon, geoHash.length());

		// System.out.println(newGeoHash);

		return newGeoHash;

	}

	public static void main4(String arg[]) {

		// System.out.println(GeoHash.getAntipodeGeohash("c7"));

		float upLat = 40.6532236764f;
		float lowLat = 40.6500305066f;

		float upLon = -104.99372356f;
		float lowLon = -104.995688942f;

		Coordinates c1 = new Coordinates(lowLat, lowLon);
		Coordinates c2 = new Coordinates(upLat, lowLon);
		Coordinates c3 = new Coordinates(upLat, upLon);
		Coordinates c4 = new Coordinates(lowLat, upLon);

		ArrayList<Coordinates> cs1 = new ArrayList<Coordinates>();
		cs1.add(c1);
		cs1.add(c2);
		cs1.add(c3);
		cs1.add(c4);

		for (String s : getIntersectingGeohashesForConvexBoundingPolygon(cs1, 8)) {

			System.out.print("\"" + s + "\",");

		}

	}

	/**
	 * 1 IN 'TRIALS' CHANCE OF RETURNING TRUE
	 * 
	 * @author sapmitra
	 * @param trials
	 * @return
	 */
	public static boolean getChance(int trials) {

		int random = ThreadLocalRandom.current().nextInt(0, trials);

		if (random == 0) {
			return true;
		}

		return false;

	}

	/**
	 * 'CHANCE' IN 10 CHANCE OF RETURNING TRUE
	 * 
	 * @author sapmitra
	 * @param chance
	 * @return
	 */
	public static boolean getProb(int chance) {

		int random = 0;

		// Get a number between 1 and 10;
		random = ThreadLocalRandom.current().nextInt(1, 11);

		if (random <= chance) {
			return true;
		}

		return false;

	}

	public static int getRandom(int trials) {

		int random = ThreadLocalRandom.current().nextInt(0, trials);

		return random;

	}

	public static List<String> getInternalGeohashes(String geohash) {

		List<String> childrenGeohashes = new ArrayList<>();

		for (char c : charMap) {
			childrenGeohashes.add(geohash + c);
		}

		return childrenGeohashes;
	}

	public static List<String> getInternalGeohashes_test(String geohash, int precision) {

		List<String> allGeoHashes = new ArrayList<String>();
		allGeoHashes.add(geohash);

		for (int i = geohash.length(); i < precision; i++) {

			List<String> currentGeohashes = new ArrayList<String>();

			for (String geoHash : allGeoHashes) {

				SpatialRange range1 = GeoHash.decodeHash(geoHash);

				Coordinates c1 = new Coordinates(range1.getLowerBoundForLatitude(), range1.getLowerBoundForLongitude());
				Coordinates c2 = new Coordinates(range1.getUpperBoundForLatitude(), range1.getLowerBoundForLongitude());
				Coordinates c3 = new Coordinates(range1.getUpperBoundForLatitude(), range1.getUpperBoundForLongitude());
				Coordinates c4 = new Coordinates(range1.getLowerBoundForLatitude(), range1.getUpperBoundForLongitude());

				ArrayList<Coordinates> cs1 = new ArrayList<Coordinates>();
				cs1.add(c1);
				cs1.add(c2);
				cs1.add(c3);
				cs1.add(c4);

				currentGeohashes.addAll(GeoHash.getInternalGeohashes(geoHash));

			}
			allGeoHashes = currentGeohashes;

		}
		// Collections.sort(allGeoHashes);
		// String[] returnArray = allGeoHashes.toArray(new String[allGeoHashes.size()]);
		return allGeoHashes;
	}

	public static List<String> getInternalGeohashes(String geohash, int precision) {

		List<String> allGeoHashes = new ArrayList<String>();
		allGeoHashes.add(geohash);

		for (int i = geohash.length(); i < precision; i++) {

			List<String> currentGeohashes = new ArrayList<String>();

			for (String geoHash : allGeoHashes) {

				SpatialRange range1 = GeoHash.decodeHash(geoHash);

				Coordinates c1 = new Coordinates(range1.getLowerBoundForLatitude(), range1.getLowerBoundForLongitude());
				Coordinates c2 = new Coordinates(range1.getUpperBoundForLatitude(), range1.getLowerBoundForLongitude());
				Coordinates c3 = new Coordinates(range1.getUpperBoundForLatitude(), range1.getUpperBoundForLongitude());
				Coordinates c4 = new Coordinates(range1.getLowerBoundForLatitude(), range1.getUpperBoundForLongitude());

				ArrayList<Coordinates> cs1 = new ArrayList<Coordinates>();
				cs1.add(c1);
				cs1.add(c2);
				cs1.add(c3);
				cs1.add(c4);

				currentGeohashes.addAll(GeoHash.getInternalGeohashes(geoHash));

			}
			allGeoHashes = currentGeohashes;

		}
		Collections.sort(allGeoHashes);
		// String[] returnArray = allGeoHashes.toArray(new String[allGeoHashes.size()]);
		return allGeoHashes;
	}

	public static String getCoordinateBounds(List<String> geohashes) throws IOException {

		String fullStr = "[";
		String nameStr = "[";

		List<String> rectangles = new ArrayList<String>();

		int i = 0;
		for (String s : geohashes) {

			SpatialRange range1 = GeoHash.decodeHash(s);

			String latlong = "[" + range1.getLowerBoundForLongitude() + "," + range1.getLowerBoundForLatitude() + ","
					+ range1.getUpperBoundForLongitude() + "," + range1.getUpperBoundForLatitude() + "]";

			if (i == 0) {
				fullStr += latlong;

				nameStr += "\"" + s + "\"";
			} else {
				fullStr += "," + latlong;

				nameStr += ",\"" + s + "\"";
			}
			i++;
		}

		// System.out.println(fullStr+"]");
		// System.out.println(nameStr+"]");

		/*
		 * FileWriter fileWriter = new
		 * FileWriter("/s/chopin/b/grad/sapmitra/Desktop/tempOutput/rikiC.txt", false);
		 * PrintWriter printWriter = new PrintWriter(fileWriter);
		 * printWriter.println(fullStr+"]"); printWriter.close();
		 * 
		 * 
		 * fileWriter = new
		 * FileWriter("/s/chopin/b/grad/sapmitra/Desktop/tempOutput/rikiN.txt", false);
		 * printWriter = new PrintWriter(fileWriter); printWriter.println(nameStr+"]");
		 * printWriter.close();
		 */

		System.out.println("COMPLETE...");
		return fullStr + "]" + "$$" + nameStr + "]";

	}

	public static void mainX(String arg[]) throws IOException {

		String allCoords = "[";
		String allNames = "[";

		int i = 0;
		for (String s : getInternalGeohashes("9xjqb")) {
			String op = getCoordinateBounds(getInternalGeohashes(s, 7));

			String cs = op.split("\\$\\$")[0];
			String ns = op.split("\\$\\$")[1];

			if (i == 0) {
				allCoords += cs;

				allNames += ns;
			} else {
				allCoords += "," + cs;

				allNames += "," + ns + "";
			}
			i++;
		}

		FileWriter fileWriter = new FileWriter("/s/chopin/b/grad/sapmitra/Desktop/tempOutput/rikiC.txt", false);
		PrintWriter printWriter = new PrintWriter(fileWriter);
		printWriter.println(allCoords + "]");
		printWriter.close();

		fileWriter = new FileWriter("/s/chopin/b/grad/sapmitra/Desktop/tempOutput/rikiN.txt", false);
		printWriter = new PrintWriter(fileWriter);
		printWriter.println(allNames + "]");
		printWriter.close();
	}

	public static int getIndexWithinGeohash(String geohash, int startLength) {

		String remainder = geohash.substring(startLength);

		// System.out.println(remainder);
		int num = getGeohashIDWithOffset(geohash, startLength);

		return num;

	}

	public static void main_revert(String arg[]) {
		//System.out.println(getInternalGeohashes_test("96", 3).size());
		//-111.96569161279297 -111.96481223242188 33.06369681494141 33.065075920654294
		Coordinates c11 = new Coordinates(33.065075920654294f, -111.96569161279297f);
		Coordinates c21 = new Coordinates(33.065075920654294f, -111.96481223242188f);
		Coordinates c31 = new Coordinates(33.06369681494141f, -111.96481223242188f);
		Coordinates c41 = new Coordinates(33.06369681494141f, -111.96569161279297f);
		List<Coordinates> cl1 = new ArrayList<Coordinates>();
		cl1.add(c11);
		cl1.add(c21);
		cl1.add(c31);
		cl1.add(c41);
		
		String[] intersectingGeohashes = getIntersectingGeohashes(cl1, 8);
		Arrays.sort(intersectingGeohashes);
		
		System.out.println(intersectingGeohashes.length);
		for(String s: getIntersectingGeohashes(cl1, 8))
			System.out.print(s+"\",\"");
	}
	
	
	public static void main1(String arg[]) {
		//(33.064443019694764, -111.964975382), (33.0644430196937, -111.96499170149933), (33.06442054969379, -111.96499146649519), (33.06442054969485, -111.964975147), (33.064443019694764, -111.964975382)
		Coordinates c11 = new Coordinates(33.064443019694764f, -111.964975382f);
		Coordinates c21 = new Coordinates(33.0644430196937f, -111.96499170149933f);
		Coordinates c31 = new Coordinates(33.06442054969379f, -111.96499146649519f);
		Coordinates c41 = new Coordinates(33.06442054969485f, -111.964975147f);
		List<Coordinates> cl1 = new ArrayList<Coordinates>();
		cl1.add(c11);
		cl1.add(c21);
		cl1.add(c31);
		cl1.add(c41);
		
		String[] intersectingGeohashes = getIntersectingGeohashes(cl1, 11);
		Arrays.sort(intersectingGeohashes);
		
		System.out.println(intersectingGeohashes.length);
		for(String s: getIntersectingGeohashes(cl1, 11))
			System.out.print(s+"\",\"");
	}
	
	public static void main(String arg[]) {
		
		
		float lat2 = 33.06507110595703f;
		float lat1 = 33.064727783203125f;
		//33.060944, -111.970672
		float lon2 = -111.96510314941406f;
		float lon1 = -111.96544647216797f;
		
		
		Coordinates c1 = new Coordinates(lat2, lon2);
		Coordinates c2 = new Coordinates(lat2, lon1);
		Coordinates c3 = new Coordinates(lat1, lon1);
		Coordinates c4 = new Coordinates(lat1, lon2);
		
		List<Coordinates> cl = new ArrayList<Coordinates>();
		cl.add(c1); cl.add(c2); cl.add(c3); cl.add(c4); 
		
		String[] intersectingGeohashes = getIntersectingGeohashes(cl, 11);
		System.out.println(intersectingGeohashes.length);
		for(String s: getIntersectingGeohashes(cl, 11))
			System.out.print(s+"\",\"");
	}

	public static void main_earthengine(String arg[]) {

		// String geohash = "9b";

		// getIndexWithinGeohash(geohash,1);

		// INSERTING NEW FIELD

		List<String> result = new ArrayList<String>();

		try (Stream<Path> walk = Files.walk(Paths.get("/s/chopin/b/grad/sapmitra/Documents/SENTINEL"))) {

			result = walk.map(x -> x.toString()).filter(f -> f.endsWith(".csv")).collect(Collectors.toList());

		} catch (IOException e) {
			e.printStackTrace();
		}

		System.out.println(result);

		int i = 0;
		for (String filePath : result) {

			if (i % 10 == 0)
				System.out.println(i + "=====" + filePath);

			i++;
			int labelIndex = 6;
			if (filePath.contains("earthengineLS"))
				labelIndex = 5;

			// String filePath =
			// "/s/chopin/b/grad/sapmitra/Documents/SENTINEL/earthengine/9xjqb0.csv";

			String newFilepath = filePath.replace("SENTINEL", "SENTINEL_UP");

			File resultsDir1 = new File(newFilepath);
			File resultsDir = new File(resultsDir1.getParent());
			if (!resultsDir.exists())
				resultsDir.mkdirs();

			String[] fTokens = filePath.split("/");

			String baseHash = fTokens[fTokens.length - 1].split("\\.")[0];

			BufferedReader reader;
			try {

				reader = new BufferedReader(new FileReader(filePath));

				String fileData = "";

				String line = reader.readLine();
				while (line != null) {

					if (line.startsWith("system")) {

						fileData += line;

					} else if (line.trim().length() > 0) {

						String[] tokens = line.split(",");

						String geohash = tokens[labelIndex];

						int pixel_indx = getIndexWithinGeohash(geohash, baseHash.length());

						String newLine = tokens[0] + "," + tokens[1] + "," + tokens[2] + "," + tokens[3] + ","
								+ tokens[4] + "," + tokens[5] + "," + tokens[6] + "," + pixel_indx;

						if (labelIndex == 5)
							newLine = tokens[0] + "," + tokens[1] + "," + tokens[2] + "," + tokens[3] + "," + tokens[4]
									+ "," + tokens[5] + "," + pixel_indx;

						fileData += "\n" + newLine;
					}
					// System.out.println(line);
					// read next line
					line = reader.readLine();
				}
				reader.close();

				FileWriter fileWriter = new FileWriter(newFilepath, false);
				PrintWriter printWriter = new PrintWriter(fileWriter);
				printWriter.print(fileData);
				printWriter.close();

			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	}

}
